/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
grammar RubanetraSystemConfiguration;

/**
 * This file defines the EBNF grammar of the default system configuration
 * using the ANTLR v4 syntax (see http://www.antlr.org/wiki/display/ANTLR4/ANTLR+4+Documentation).
 * The general settings '*_configuration_file' may be used
 * to indicate the location of the corresponding configuration files, however
 * the main configuration file path containing the 'general' section
 * must be known by the system a priori at a fixed location (usually the classpath) or it has to be passed as
 * command line parameter during runtime.
 * The recommended setup is to use two configuration files, i.e.
 * one file containing all static configuration blocks ('general', 'protocols')
 * and another file containing the dynamic runtime configuration ('input', 'output').
 * The system configuration itself may be split into four different files according
 * to the main settings categories 'general', 'protocols', 'input/output' (a coarser split into a single, two or three
 * file(s) is possible as well).
 *
 * The 'system' node serves as grammar entry point and consists of a mandatory 'general' 
 * configuration section, followed by the remaining configurations optionally within the
 * same file. However, this does not implicate that the protocol/library/runtime configurations 
 * are optional in general.
 **/ 
system : (generalConfig|inputConfig|outputConfig|protocolConfig)* ;

/**
 * Although the general configuration contains a predefined number of settings, 
 * future extensions may lead to a different set of settings, therefore no semantic 
 * restrictions will be laid upon the parser at the top level nodes.  
 **/  
generalConfig  : 'general' '{' (generalSetting)* '}' ';' ;
protocolConfig : 'protocols' '{' (protocolConfigurationFileSetting | (protocolSetting)*) '}' ';' ;
inputConfig    : 'input' '{' (inputConfigurationFileSetting | (inputSetting)*) '}' ';' ;
outputConfig   : 'output' '{' (outputConfigurationFileSetting | (outputSetting)*) '}' ';' ;

/**
 * ##################################
 * # General Setting Specification  #
 * ##################################
 **/
generalSetting :  ( 'drools_base_model_name'         '=' droolsBaseModelName       // optional, the name of the knowledge base to use (defined in 'kmodule.xml')
                  | 'drools_session_model_name'      '=' droolsSessionModelName    // optional, the name of the session to use (defined in 'kmodule.xml')
                  | 'drools_configuration_directory' '=' droolsConfigurationFile   // required, path of a separate file containing the Drools knowledge base configuration ('kmodule.xml')
                  ) ';' ;
droolsBaseModelName       : STRING ;
droolsSessionModelName    : STRING ;
droolsConfigurationFile   : filePath ;
// End of General Setting Specification

/**
 * ###################################
 * # Protocol Setting Specification  #
 * ###################################
 **/
protocolConfigurationFileSetting : 'protocol_configuration_file' '=' filePath ;

protocolSetting : (protocolHeader (portSpecification|protocolBinding|transportLayerMappingSetting)*) ;
protocolHeader  : '[' protocolId ']' ':' ;     // every protocol parser configuration section has an unique identifier

portSpecification : 'port' '=' ( ALL           // matches all network ports [0;2^16-1]
                               | NONE          // matches no ports (i.e. another way to disable a protocol)
                               | multiplePorts // matches either one or several single ports or port ranges
                               ) ';' ;
multiplePorts    : portRange (',' portRange)* ;
portRange        : singlePort ('-' singlePort)? ;
protocolBinding  : 'protocol_binding'  '=' protocolId BIND_OP protocolId ';';
transportLayerMappingSetting : 'transport_layer_mapping_strategy' '=' transportLayerMappingStrategy ';' ;
transportLayerMappingStrategy : (DESTINATION_PORT) ;
// End of Protocol Setting Specification

/**
 * ###########################################
 * # Dynamic Runtime Setting Specification   #
 * ###########################################
 * The following settings will be used to adjust the concrete runtime behaviour of the system,
 * i.e. a set of one or multiple PCAP input files, a set of predefined protocols (see Protocol Setting Specification),
 * a set of rules to enable.
 **/
inputConfigurationFileSetting : 'input_configuration_file'    '=' filePath ';' ;   // optional, path of a separate file containing the I/O configuration
inputSetting : ( 'input_file'   '=' inputSpecification // required, list of network capture input files
               | 'input_format' '=' inputFormat        // required, currently only PCAP (all format versions supported by the utilized libpcap library)
               | 'bpf_filter'   '=' bpfFilterString    // optional, String representing a Berkeley Packet Filter (BPF) expression - a syntax check will not be performed
               | 'bpf_netmask'  '=' bpfNetmask         // optional, integer representing a Berkeley Packet Filter (BPF) netmask
               | 'bpf_optimize' '=' bpfOptimize        // optional, boolean representing a Berkeley Packet Filter (BPF) optimization flag
               | 'sort_by_first_timestamp' '=' sortByFirstTimestamp // optional, boolean flag indicating whether or not to sort according to the first timestamp of the PCAP entry
               ) ';' ;

outputConfigurationFileSetting : 'output_configuration_file'   '=' filePath ';';   // optional, path of a separate file containing the I/O configuration
outputSetting : ( 'output_file'   '=' outputSpecification // required, content output stream
                | 'output_format' '=' outputFormat        // plaso parser or general csv format
                ) ';' ;

inputSpecification   : filePath (',' filePath)* ;
inputFormat          : (PCAP) ;
bpfFilterString      : STRING ;
bpfNetmask           : INT ;
bpfOptimize          : (TRUE|FALSE) ;
sortByFirstTimestamp : (TRUE|FALSE) ;
outputSpecification  : (STDOUT|filePath) ;
outputFormat         : (PLASO|CSV|NOP) ;
// End of Runtime Setting Specification
        
regexFlag : // see docs.oracle.com/javase/7/docs/api/java/util/regex/Pattern.html#UNIX_LINES
            CASE_INSENSITIVE | MULTILINE | DOTALL | UNICODE_CASE | CANON_EQ
          | UNIX_LINES | LITERAL | UNICODE_CHARACTER_CLASS | COMMENTS | NONE ;
ipv6 :  // loosely based on RFC 5954, see http://tools.ietf.org/html/rfc5954, semantic check (int range) will be performed after parsing
     ( INT ':' INT ':' INT ':' INT ':' INT ':' INT ':' optIPv6Trailer
     |                                                                                '::' INT ':' INT ':' INT ':' INT ':' INT ':' optIPv6Trailer
     | INT?                                                                           '::' INT ':' INT ':' INT ':' INT ':' optIPv6Trailer
     | ((INT ':' )? INT)?                                                             '::' INT ':' INT ':' INT ':' optIPv6Trailer
     | ((INT ':' )? (INT ':' )? INT)?                                                 '::' INT ':' INT ':' optIPv6Trailer
     | ((INT ':' )? (INT ':' )? (INT ':' )? INT)?                                     '::' INT ':' optIPv6Trailer
     | ((INT ':' )? (INT ':' )? (INT ':' )? (INT ':' )? INT)?                         '::' optIPv6Trailer
     | ((INT ':' )? (INT ':' )? (INT ':' )? (INT ':' )? (INT ':' )? INT)?             '::' INT
     | ((INT ':' )? (INT ':' )? (INT ':' )? (INT ':' )? (INT ':' )? (INT ':' )? INT)? '::'
     )
     ;

optIPv6Trailer : ( INT ':' INT) | ipv4 ;
ipv4 : INT '.' INT '.' INT '.' INT ; // Match an IPv4 address, e.g. 127.0.0.1

protocolId : ID ;
singlePort : INT ;
filePath   : STRING ;

DEBUG    : [Dd][Ee][Bb][Uu][Gg] ;
WARNING  : [Ww][Aa][Rr][Nn][Ii][Nn][Gg] ;
INFO     : [Ii][Nn][Ff][Oo] ;
DISABLED : [Dd][Ii][Ss][Aa][Bb][Ll][Ee][Dd] ;
PLASO    : [Pp][Ll][Aa][Ss][Oo] ;
PCAP     : [Pp][Cc][Aa][Pp] ;
DEFAULT  : [Dd][Ee][Ff][Aa][Uu][Ll][Tt] ;
DESTINATION_PORT: [Dd][Ee][Ss][Tt][Ii][Nn][Aa][Tt][Ii][Oo][Nn]'_'[Pp][Oo][Rr][Tt] ;
CSV      : [Cc][Ss][Vv] ;
CUSTOM   : [Cc][Uu][Ss][Tt][Oo][Mm] ;
TRUE     : ([Tt][Rr][Uu][Ee]) ;
FALSE    : ([Ff][Aa][Ll][Ss][Ee]) ;
BIND_OP  : ([Tt][Oo]) | '->' ;
ALL      : ([Aa][Ll][Ll]) | ([Aa][Nn][Yy]) ;
NONE     : [Nn][Oo][Nn][Ee] ;
NOP      : [Nn][Oo][Pp] ;
STDOUT   : ([Ss][Tt][Dd][Oo][Uu][Tt]) ;

/**
 * Possible Java regular expression flags
 * See docs.oracle.com/javase/7/docs/api/java/util/regex/Pattern.html#UNIX_LINES
 **/
CASE_INSENSITIVE        : [Cc][Aa][Ss][Ee][_][Ii][Nn][Ss][Ee][Nn][Ss][Ii][Tt][Ii][Vv][Ee] ;
MULTILINE               : [Mm][Uu][Ll][Tt][Ii][Ll][Ii][Nn][Ee];
DOTALL                  : [Dd][Oo][Tt][Aa][Ll][Ll] ;
UNICODE_CASE            : [Uu][Nn][Ii][Cc][Oo][Dd][Ee][_][Cc][Aa][Ss][Ee] ;
CANON_EQ                : [Cc][Aa][Nn][Oo][Nn][_][Ee][Qq] ;
UNIX_LINES              : [Uu][Nn][Ii][Xx][_][Ll][Ii][Nn][Ee][Ss] ;
LITERAL                 : [Ll][Ii][Tt][Ee][Rr][Aa][Ll] ;
UNICODE_CHARACTER_CLASS : [Uu][Nn][Ii][Cc][Oo][Dd][Ee][_][Cc][Hh][Aa][Rr][Aa][Cc][Tt][Ee][Rr][_][Cc][Ll][Aa][Ss][Ss] ;
COMMENTS                : [Cc][Oo][Mm][Mm][Ee][Nn][Tt][Ss] ;


fragment DIGIT : [0-9] ;
ID             : ([a-zA-Z]+ DIGIT*)+ ; // match lower-case and upper-case identifiers
INT            : DIGIT+ ; // match natural, positive numbers including 0


STRING       : '"' ('\\"'|.)*? '"' ; // match any characters between double quotes, including escaped '"'
LINE_COMMENT : ('//'|'#') .*? '\r'? '\n' -> skip ; // match line comments
COMMENT      : '/*' .*? '*/' -> skip ; // match "/*" multiline comment "*/"
WS           : [ \t\r\n]+ -> skip ; // skip spaces, tabs, newlines
