/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.config;

import at.jku.fim.rubanetra.config.model.RubanetraSystemConfiguration;
import at.jku.fim.rubanetra.config.model.impl.RubanetraSystemConfigurationBaseListenerImpl;
import at.jku.fim.rubanetra.config.model.impl.RubanetraSystemConfigurationBuilder;
import at.jku.fim.rubanetra.drools.DroolsKrakenProtocolHandler;
import org.antlr.v4.runtime.tree.ParseTreeWalker;
import org.apache.commons.cli.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;
import java.io.OutputStream;
import java.net.URISyntaxException;
import java.net.URL;
import java.nio.file.Paths;
import java.util.Enumeration;
import java.util.LinkedList;
import java.util.jar.Attributes;
import java.util.jar.Manifest;

/**
 * This class serves as the projects main entry point, i.e. it contains a main-method and provides
 * a basic command line interface. However, it may also serve as basic configuration initializer for use by other
 * classes.
 * <p>
 * The main method of this class relies on the fact that it has been executed and loaded from within a runnable
 * jar archive containing a manifest file consisting of basic project information.
 * By default, i.e. without providing any command line arguments, the main configuration will be looked up at the
 * following locations:
 * <ul>
 * <li>File named "rubanetra.conf" in the classpath and as a relative path</li>
 * <li>"./conf/rubanetra.conf" in the classpath and as relative path</li>
 * <li>Value of environment variable "RUBANETRA_SYSTEM_CONF_FILE" interpreted as file path.</li>
 * </ul>
 * <p>
 * If the file could not be located successfully an exception will be thrown.
 * Alternatively, the main configuration file path may be provided as command line argument using
 * the switch '-c' or '--configuration-file'.
 * <p>
 * During the initialization procedure the ANTLR configuration parser will be invoked. Any encountered syntax errors
 * will be printed to the standard error stream, however, unless any critical errors during the setup process
 * occurred the application will continue running.
 * <p>
 * After completing the initialization procedure the control over the program flow will be returned
 * to the caller in the main method, who in turn is able to retrieve the parsed and validated @{RubanetraSystemConfiguration}.
 * By default, invoking the start-method
 * of this class will initialize a default @{DroolsKrakenProtocolHandler} which effectively passes all PCAP-entries
 * to the Kraken protocol decoding pipeline and ultimately to the Drools rule engine.
 *
 * @see at.jku.fim.rubanetra.drools.DroolsKrakenProtocolHandler
 * @see at.jku.fim.rubanetra.config.model.impl.RubanetraSystemConfigurationBuilder
 * @see at.jku.fim.rubanetra.config.model.RubanetraSystemConfiguration
 */
public class ConfigurationController {
    public static final String defaultMainConfigName = "rubanetra.conf";
    public static final String defaultMainConfigPath = "./conf/" + defaultMainConfigName;
    public static final String defaultMainConfigEnvVar = "RUBANETRA_SYSTEM_CONF_FILE";
    private static final Logger log = LoggerFactory.getLogger(ConfigurationController.class);
    private static final Option helpOption = new Option("h", "help", false, "print the option overview message");
    private static final Option versionOption = new Option("v", "version", false, "print the version information");
    private static final Option mainConfigFileOption = new Option("c", "configuration-file", true, "path to the main configuration file");
    private static final Option inputFiles = new Option("i", "input-files", true, String.format("pcap input files, delimited by '%s' (complements config file)", File.pathSeparatorChar));
    private static final Option outputFile = new Option("o", "output-file", true, "path to the output file (contents will be overridden)");
    private static final Option outputType = new Option("t", "output-type", true, "output type/format of the generated content (PLASO, CSV, JSON, XML or NOP)");

    private final RubanetraSystemConfigurationBuilder systemConfigBuilder;
    private RubanetraSystemConfiguration systemConfiguration;
    private final LinkedList<String> inputFilePaths = new LinkedList<>();
    private String outputFilePath;
    private String outputFormat;

    /**
     * Creates a new instance of itself and a {@link at.jku.fim.rubanetra.config.model.impl.RubanetraSystemConfigurationBuilder}
     */
    public ConfigurationController() {
        systemConfigBuilder = RubanetraSystemConfigurationBuilder.create();
    }

    /**
     * @return the default command line options, i.e. help (h), version (v), configuration-file (c), input-files (i),
     * output-file (o), output-type (t)
     */
    public static Options createDefaultOptions() {
        inputFiles.setArgs(Option.UNLIMITED_VALUES);
        inputFiles.setValueSeparator(File.pathSeparatorChar);
        Options options = new Options();
        options.addOption(helpOption);
        options.addOption(versionOption);
        options.addOption(mainConfigFileOption);
        options.addOption(inputFiles);
        options.addOption(outputFile);
        options.addOption(outputType);
        return options;
    }

    /**
     * Creates a {@link at.jku.fim.rubanetra.config.ConfigurationController} using default options
     * {@link #createDefaultOptions()} and a {@link org.apache.commons.cli.GnuParser} to parse the command line
     * arguments.
     * After a successful parsing process, {@link ConfigurationController#start()}
     * will be invoked.
     *
     * @param args command line arguments as defined by {@link #createDefaultOptions()}
     */
    public static void main(String[] args) {
        ConfigurationController controller = new ConfigurationController();
        Options defaultOptions = ConfigurationController.createDefaultOptions();
        CommandLineParser cliParser = new GnuParser();

        CommandLine cli;
        try {
            cli = cliParser.parse(defaultOptions, args);
        } catch (ParseException e) {
            log.error("Unable to process parsed arguments: ", e);
            printHelp(defaultOptions);
            return;
        }

        if (cli.hasOption(ConfigurationController.helpOption.getOpt())) {
            printHelp(defaultOptions);
            return;
        } else if (cli.hasOption(ConfigurationController.versionOption.getOpt())) {
            printVersion();
            return;
        }

        if (cli.hasOption(ConfigurationController.inputFiles.getOpt())) {
            String[] inputFileValues = cli.getOptionValues(ConfigurationController.inputFiles.getOpt());
            controller.addInputFiles(inputFileValues);
        }

        if (cli.hasOption(ConfigurationController.outputFile.getOpt())) {
            String outputFileValue = cli.getOptionValue(ConfigurationController.outputFile.getOpt());
            controller.setOutputFile(outputFileValue);
        }

        if (cli.hasOption(ConfigurationController.outputType.getOpt())) {
            String outputType = cli.getOptionValue(ConfigurationController.outputType.getOpt());
            controller.setOutputFormat(outputType);
        }

        boolean initialized;

        String mainConfigurationPath = cli.getOptionValue(ConfigurationController.mainConfigFileOption.getOpt());
        if (mainConfigurationPath == null) {
            // config was not provided via command line arg
            initialized = controller.initialize();
        } else {
            initialized = controller.initialize(mainConfigurationPath);
        }

        if (!initialized || controller.getRubanetraSystemConfiguration() == null) {
            log.error("Unable to initialize the system: configuration file does either not exist or it is not readable or it is not valid.");
            return;
        }

        log.info("System configuration has been initialized successfully, starting the rule engine and input file parsing process.");

        controller.start();
    }

    private static void printVersion() {
        Package classPackage = ConfigurationController.class.getPackage();
        StringBuilder strBuilder = new StringBuilder();
        Attributes manifestAttributes = getMainAttributes();
        if (manifestAttributes != null) {
            strBuilder.append(String.format("%s %s%n", classPackage.getImplementationTitle(), classPackage.getImplementationVersion()));
            String licShort = manifestAttributes.getValue("License-Short-Name");
            String licLong = manifestAttributes.getValue("License-Long-Name");
            String licHeaderLine = manifestAttributes.getValue("License-Short-Header");
            String licUrl = manifestAttributes.getValue("License-Url");
            String licInceptionYear = manifestAttributes.getValue("License-Inception-Year");
            String licRecentYears = manifestAttributes.getValue("License-Recent-Years");
            String copyrightOwner = manifestAttributes.getValue("Copyright-Owner");
            strBuilder.append(String.format("Copyright (C) %s,%s %s%n", licInceptionYear, licRecentYears, copyrightOwner));
            strBuilder.append(String.format("License %s: %s <%s>%n", licShort, licLong, licUrl));
            String[] headerSplit = licHeaderLine.split("\\.\\s", 2);
            String headerML = headerSplit.length == 2 ?
                    String.format("%s.%n%s%n", headerSplit[0], headerSplit[1]) : String.format("%s%n", licHeaderLine);
            strBuilder.append(headerML);
        } else {
            strBuilder.append("Unknown version.%n");
        }
        System.out.println(strBuilder.toString());
    }

    private static void printHelp(Options defaultOptions) {
        Package classPackage = ConfigurationController.class.getPackage();
        HelpFormatter formatter = new HelpFormatter();
        Attributes manifestAttributes = getMainAttributes();

        String cmdLineSyntax;
        try {
            cmdLineSyntax = Paths.get(ConfigurationController.class.getProtectionDomain().getCodeSource().getLocation().toURI()).getFileName().toString();
        } catch (URISyntaxException e) {
            cmdLineSyntax = String.format("./rubanetra-%s.jar", classPackage.getImplementationVersion());
        }
        formatter.printHelp(String.format("java -jar %s <args>", cmdLineSyntax), defaultOptions);

        if (manifestAttributes != null) {
            StringBuilder strBuilder = new StringBuilder();
            strBuilder.append(String.format("%nReport bugs to: <%s> or <stefan_rubanetra@swerk.priv.at>%n",
                    manifestAttributes.getValue("Issue-Management")));
            strBuilder.append(String.format("Project home page: <%s>%n", manifestAttributes.getValue("Project-Home")));
            System.out.println(strBuilder.toString());
        }
    }

    /**
     * @return the main attributes of the invoked JAR's manifest, containing at least "Implementation-Title",
     * "Implementation-Version" and "Implementation-Vendor" attributes or null, if the manifest was not found
     */
    public static Attributes getMainAttributes() {
        Package classPackage = ConfigurationController.class.getPackage();
        try {
            Enumeration<URL> resourceEnum = ClassLoader.getSystemResources("META-INF/MANIFEST.MF");
            while (resourceEnum.hasMoreElements()) {
                try {
                    URL nextManifestUrl = resourceEnum.nextElement();
                    Manifest manifest = new Manifest(nextManifestUrl.openStream());
                    Attributes manifestAttributes = manifest.getMainAttributes();
                    String mainClassVal = manifestAttributes.getValue("Main-Class");
                    String implementationTitleVal = manifestAttributes.getValue("Implementation-Title");
                    String implementationVersionVal = manifestAttributes.getValue("Implementation-Version");
                    String implementationVendorVal = manifestAttributes.getValue("Implementation-Vendor");
                    if (mainClassVal != null && mainClassVal.equals(ConfigurationController.class.getCanonicalName())
                            && classPackage.getImplementationTitle().equals(implementationTitleVal)
                            && classPackage.getImplementationVendor().equals(implementationVendorVal)
                            && classPackage.getImplementationVersion().equals(implementationVersionVal)) {
                        return manifestAttributes;
                    }
                } catch (IOException e) {
                    log.warn("Unable to open manifest, skipping...", e);
                }
            }
        } catch (IOException e) {
            log.warn("Unable to find any manifest", e);
        }
        return null;
    }

    /**
     * This method initializes the system invoking a parser for the main configuration file
     * denoted by mainConfigFile. It tries to canonicalize all encountered file paths and  will fail if it is unable
     * to create a canonical representation.
     * It relies on {@link at.jku.fim.rubanetra.config.RubanetraSystemConfigurationParser} and
     * {@link at.jku.fim.rubanetra.config.model.impl.RubanetraSystemConfigurationBaseListenerImpl} for parsing
     * functionality and builds the final {@link at.jku.fim.rubanetra.config.model.RubanetraSystemConfiguration}
     * object which can be obtained iff this method returned true via {@link #getRubanetraSystemConfiguration()}.
     *
     * @param mainConfigFile the main configuration file, which may not be null
     * @return true, iff a main configuration instance was created (one may obtain this reference by calling
     * {@link #getRubanetraSystemConfiguration()}.
     * @throws java.lang.NullPointerException     if the main config file is a null pointer
     * @throws java.lang.IllegalArgumentException if the main config file is not valid
     */
    public boolean initialize(File mainConfigFile) {
        if (mainConfigFile == null) {
            throw new NullPointerException("The main configuration file must not be a null pointer");
        }
        try {
            mainConfigFile = mainConfigFile.getCanonicalFile();
        } catch (IOException e) {
            log.error("Unable to canonicalize file {}", mainConfigFile);
            return false;
        }
        if (!checkMainConfigFile(mainConfigFile)) {
            return false;
        }
        log.info("Trying to load main configuration from file {}", mainConfigFile);

        RubanetraSystemConfigurationParser systemConfigParser = RubanetraSystemConfigurationBaseListenerImpl.createConfigurationFileParser(mainConfigFile);
        RubanetraSystemConfigurationBaseListenerImpl systemConfigBaseListener = new RubanetraSystemConfigurationBaseListenerImpl(systemConfigBuilder);

        RubanetraSystemConfigurationParser.SystemContext systemContext = systemConfigParser.system();

        if (systemContext == null) {
            throw new IllegalArgumentException("Invalid configuration file.");
        }
        String errorMsg = "Exactly one %s configuration section required, %d have been found";
        int generalConfigurations = systemContext.generalConfig().size();
        int inputConfigurations = systemContext.inputConfig().size();
        int outputConfigurations = systemContext.outputConfig().size();
        int protocolConfigurations = systemContext.protocolConfig().size();
        if (generalConfigurations != 1)
            throw new IllegalArgumentException(String.format(errorMsg, "general", generalConfigurations));
        if (inputConfigurations != 1)
            throw new IllegalArgumentException(String.format(errorMsg, "input", inputConfigurations));
        if (outputConfigurations != 1)
            throw new IllegalArgumentException(String.format(errorMsg, "output", outputConfigurations));
        if (protocolConfigurations != 1)
            throw new IllegalArgumentException(String.format(errorMsg, "protocol", protocolConfigurations));

        /*
         The general section should be parsed first, as it could contain vital information,
         followed by the I/O configuration. Due to potentially existing file path entries in those sections and the
         possibility of a redundant path specification via command line arguments, these settings must be adjusted
         accordingly, i.e. either complemented or replaced altogether before the Pcap handler can be built, since the
         protocol configuration depends on an already existing Pcap handler in order to setup listeners.
         */
        ParseTreeWalker.DEFAULT.walk(systemConfigBaseListener, systemContext.generalConfig(0));
        ParseTreeWalker.DEFAULT.walk(systemConfigBaseListener, systemContext.inputConfig(0));
        ParseTreeWalker.DEFAULT.walk(systemConfigBaseListener, systemContext.outputConfig(0));
        complementOrReplaceExistingConfig();
        systemConfigBuilder.createPcapHandler();
        ParseTreeWalker.DEFAULT.walk(systemConfigBaseListener, systemContext.protocolConfig(0));

        systemConfiguration = systemConfigBuilder.build();

        return true;
    }

    /**
     * Tries to either complement (input files) or replace (output file/type) the existing configuration parsed from
     * the main config files by the provided command line arguments.
     */
    private void complementOrReplaceExistingConfig() {
        if (this.inputFilePaths.size() > 0) {

            if (!systemConfigBuilder.getInputFiles().isEmpty()) {
                log.warn("The configuration file contains at least one reference to an input file, complementing with" +
                        " specified command line input file(s)");
            }
            for (String inputFilePath : this.inputFilePaths) {
                systemConfigBuilder.addInputPath(inputFilePath);
            }
        }

        if (this.outputFilePath != null) {
            final OutputStream oldOutputStream = systemConfigBuilder.getOutputStream();
            if (oldOutputStream != null) {
                log.warn("The configuration file contains at least one reference to an output stream ... replacing it");
                if (oldOutputStream != System.out && oldOutputStream != System.err) {
                    try {
                        oldOutputStream.close();
                    } catch (IOException e) {
                        log.warn("IOException while trying to close the replaced output stream", e);
                    }
                }
            }
            if (outputFilePath.equalsIgnoreCase("stdout") || outputFilePath.equals("-")) {
                systemConfigBuilder.setOutputStream(System.out);
            } else {
                systemConfigBuilder.setOutputFile(outputFilePath);
            }
        }

        if (this.outputFormat != null) {
            if (systemConfigBuilder.getOutputFormat() != null) {
                log.warn("The configuration file contains at least one reference to an output format ... replacing it");
            }
            systemConfigBuilder.setOutputFormat(outputFormat);
        }
    }

    /**
     * Convenience method for {@link #initialize(java.io.File)}.
     * It creates a file based on the given path.
     *
     * @param mainConfigurationPath the path of the main configuration file
     * @return true, iff the configuration could be parsed (one may obtain a reference to the configuration object by
     * calling {@link #getRubanetraSystemConfiguration()}.
     */
    public boolean initialize(String mainConfigurationPath) {
        File systemConfigurationFile = createFileFromResourceString(mainConfigurationPath);
        if (systemConfigurationFile == null) {
            systemConfigurationFile = getCanonicalFile(mainConfigurationPath);
        }

        return initialize(systemConfigurationFile);
    }

    /**
     * This method initializes the system invoking a parser for the main configuration file
     * denoted by (in processing order):
     * <ul>
     * <li>"rubanetra.conf" in the classpath and working directory</li>
     * <li>"./conf/rubanetra.conf" in the classpath and working directory</li>
     * <li>path denoted by environment variable FNA_SYSTEM_CONF_FILE in the class path and working directory</li>
     * </ul>
     * It tries to canonicalize all encountered file paths and  will fail if it is unable
     * to create a canonical representation.
     * It relies on {@link at.jku.fim.rubanetra.config.RubanetraSystemConfigurationParser} and
     * {@link at.jku.fim.rubanetra.config.model.impl.RubanetraSystemConfigurationBaseListenerImpl} for parsing
     * functionality and builds the final {@link at.jku.fim.rubanetra.config.model.RubanetraSystemConfiguration}
     * object which can be obtained iff this method returned true via {@link #getRubanetraSystemConfiguration()}.
     *
     * @return true, iff a main configuration instance was created (one may obtain this reference by calling
     * {@link #getRubanetraSystemConfiguration()}.
     * @throws java.lang.NullPointerException     if the main config file is a null pointer
     * @throws java.lang.IllegalArgumentException if the main config file is not valid
     */
    public boolean initialize() {
        boolean initialized;
        log.info("Trying to look up {} in the classpath and the working directory", defaultMainConfigName);
        // try the default name
        initialized = initialize(defaultMainConfigName);

        if (!initialized) {
            log.info("Trying to look up {} in the classpath and the working directory", defaultMainConfigPath);
            // try the default full path
            initialized = initialize(defaultMainConfigPath);
        }
        if (!initialized) {
            log.info("Trying to look up path denoted by environment variable {} in the classpath and the working directory",
                    defaultMainConfigEnvVar);
            // try the default env variable
            initialized = initializeFromEnvVar(defaultMainConfigEnvVar);
        }

        return initialized;
    }

    /**
     * Tries to look up the value of the environment variable denoted by envVariableName, and interprets it as
     * main configuration path ({@link #initialize(java.io.File)} will be called)
     *
     * @param envVariableName the name of the environment variable to look up (the value should represent the main
     *                        configuration file path)
     * @return true, iff a main configuration instance was created (one may obtain this reference by calling
     * {@link #getRubanetraSystemConfiguration()} afterwards.
     */
    public boolean initializeFromEnvVar(String envVariableName) {
        // try environment variable
        log.info("Trying to load main configuration path from environment variable {}", envVariableName);
        File systemConfigurationFile = createFileFromEnvVariable(envVariableName);

        return initialize(systemConfigurationFile);
    }

    /**
     * Tries to represent the path denoted by mainConfigurationPath canonically.
     *
     * @param mainConfigurationPath the file to canonicalize.
     * @return the file denoted by mainConfigurationPath using a canonical representation or null, if not possible
     */
    public File getCanonicalFile(String mainConfigurationPath) {
        File systemConfigurationFile;
        try {
            systemConfigurationFile = new File(mainConfigurationPath).getCanonicalFile();
        } catch (IOException e) {
            log.warn("Unable to canonicalize the configuration file path");
            return null;
        }
        return systemConfigurationFile;
    }

    private File createFileFromEnvVariable(String optionalMainConfigEnvVar) {
        String mainConfigPath;
        try {
            mainConfigPath = System.getenv(optionalMainConfigEnvVar);
        } catch (SecurityException e) {
            log.warn("Unable to access environment variable", e);
            return null;
        }
        if (mainConfigPath == null) {
            log.warn("Unable to load configuration file via environment variable.");
        } else {
            File fileFromResourceString = createFileFromResourceString(mainConfigPath);
            return fileFromResourceString == null ? getCanonicalFile(mainConfigPath) : fileFromResourceString;
        }

        return null;
    }

    /**
     * Tries to create a file object by looking up filePath in the current classloader's classpath.
     *
     * @param filePath the path of the file in the current classpath
     * @return a file object for filePath, or null if not found
     */
    public File createFileFromResourceString(String filePath) {
        log.info("Trying to load main configuration as classpath resource: {}", filePath);
        URL fileUrl = ClassLoader.getSystemResource(filePath);
        if (fileUrl == null) {
            log.warn("Unable to find configuration file via class path resource loader: {}", filePath);
            return null;
        }
        File file;
        try {
            file = new File(fileUrl.toURI());
        } catch (URISyntaxException e) {
            file = new File(filePath);
        }

        return file;
    }

    /**
     * Performs a simple file metadata validation (is it readable, is it a file),
     * however, it does not verify the contents of the file itself.
     *
     * @param file the file to check
     * @return true, iff the file is not a directory and is readable and is not null
     */
    public boolean checkMainConfigFile(File file) {
        if (file != null && !file.isDirectory() && file.canRead()) {
            return true;
        }
        log.warn("Cannot read configuration file {}", file);
        return false;
    }

    /**
     * Adds files containing network captures to the configuration builder
     * {@link at.jku.fim.rubanetra.config.model.impl.RubanetraSystemConfigurationBuilder}.
     * <p>
     * This method may not be invoked after the initialization step has been completed, i.e. this method must be
     * invoked before {@link #initialize(java.io.File)}.
     *
     * @param filePaths a number of file paths containing network captures in the specified
     *                  {@link at.jku.fim.rubanetra.config.model.RubanetraSystemConfiguration.InputFormat}
     */
    public void addInputFiles(String... filePaths) {
        checkUninitialized();
        if (filePaths == null) {
            log.error("Invalid command line input file specification.");
            throw new IllegalArgumentException();
        }
        for (String filePath : filePaths) {
            if (filePath == null || filePath.isEmpty()) {
                log.warn("Input file path passed from command line appears to be null/empty - ignoring");
            } else {
                inputFilePaths.addLast(filePath);
            }
        }
    }

    /**
     * Sets the default output file for the derived and parsed data using
     * {@link at.jku.fim.rubanetra.config.model.RubanetraSystemConfiguration.OutputFormat}
     * <p>
     * This method may not be invoked after the initialization step has been completed, i.e. this method must be
     * invoked before {@link #initialize(java.io.File)}.
     *
     * @param outputFilePath the path of the default output file
     */
    public void setOutputFile(String outputFilePath) {
        checkUninitialized();
        if (outputFilePath == null || outputFilePath.isEmpty()) {
            log.error("Output file path passed from command line appears to be null/empty");
            throw new IllegalArgumentException();
        }

        this.outputFilePath = outputFilePath;
    }

    /**
     * Sets the {@link at.jku.fim.rubanetra.config.model.RubanetraSystemConfiguration.OutputFormat} to apply
     * as serialization variant.
     * <p>
     * This method may not be invoked after the initialization step has been completed, i.e. this method must be
     * invoked before {@link #initialize(java.io.File)}.
     *
     * @param outputFormat the String representation of a {@link at.jku.fim.rubanetra.config.model.RubanetraSystemConfiguration.OutputFormat}
     */
    public void setOutputFormat(String outputFormat) {
        checkUninitialized();
        if (outputFormat == null || outputFormat.isEmpty()) {
            log.error("Output format path passed from command line appears to be null/empty");
            throw new IllegalArgumentException();
        }
        this.outputFormat = outputFormat;
    }

    private void checkUninitialized() {
        if (this.systemConfiguration != null) {
            log.error("This system configuration has already been initialized, unable to add additional input files");
            throw new IllegalStateException();
        }
    }

    /**
     * @return this method will always return null, unless one of the {@link #initialize(java.io.File)} methods was invoked
     * successfully, then it will return the constructed {@link at.jku.fim.rubanetra.config.model.RubanetraSystemConfiguration}.
     */
    public RubanetraSystemConfiguration getRubanetraSystemConfiguration() {
        return systemConfiguration;
    }

    /**
     * This method will do nothing, unless one of the {@link #initialize(java.io.File)} methods has been invoked
     * successfully.
     * <p>
     * If a valid {@link at.jku.fim.rubanetra.config.model.RubanetraSystemConfiguration} is encountered,
     * a {@link at.jku.fim.rubanetra.drools.DroolsKrakenProtocolHandler} will be constructed and
     * {@link at.jku.fim.rubanetra.pcap.PcapHandler#readNextPcapEntry(at.jku.fim.rubanetra.pcap.PcapActivityListener)}
     * will be called as long as it returns true (using an infinite loop), i.e. until all network packets have been read
     * from the input files. Afterwards, it will try to close all opened input and output streams, that is it will call
     * {@link at.jku.fim.rubanetra.drools.DroolsKrakenProtocolHandler#close()}
     * and {@link at.jku.fim.rubanetra.pcap.PcapHandler#close()}.
     * Ultimately the existing reference to the {@link at.jku.fim.rubanetra.config.model.RubanetraSystemConfiguration}
     * will be erased, therefore allowing for a subsequent initialization call while reusing the same object.
     */
    public void start() {
        if (this.systemConfiguration == null || systemConfiguration.getPcapHandler() == null) {
            log.error("System configuration has not been initialized completely.");
            return;
        }
        DroolsKrakenProtocolHandler droolsKrakenProtocolHandler = new DroolsKrakenProtocolHandler(systemConfiguration);
        while (true) {
            try {
                final boolean processingSuccessful = (systemConfiguration.getPcapHandler().readNextPcapEntry(droolsKrakenProtocolHandler));
                if (!processingSuccessful) break;
            } catch (Exception e) {
                log.debug("Catch all exception block was entered", e);
            }
        }

        try {
            droolsKrakenProtocolHandler.close();
        } catch (IOException e) {
            log.warn("Exception occurred while trying to close the Drools/Kraken Protocol Handler:", e);
        } finally {
            this.systemConfiguration.getPcapHandler().close();
            this.systemConfiguration = null;
        }
    }
}
