/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.config.model;


import at.jku.fim.rubanetra.pcap.PcapActivityListener;
import at.jku.fim.rubanetra.pcap.PcapHandler;
import at.jku.fim.rubanetra.protocol.KrakenBaseProtocol;
import at.jku.fim.rubanetra.protocol.mapper.KrakenTransportLayerMappingFactory;

import java.util.Map;

/**
 * This interface specifies methods that may be used to represent the configuration of a single application layer
 * protocol of the Kraken-library. A {@link at.jku.fim.rubanetra.config.model.impl.KrakenProtocolConfigurationBuilderImpl}
 * should be used to obtain an implementation of this interface.
 * <p>
 * Due to the current design of the Kraken-PCAP-library all application layer protocols
 * rely on either a UDP ({@link org.krakenapps.pcap.decoder.udp.UdpProtocolMapper}) or a TCP
 * ({@link org.krakenapps.pcap.decoder.tcp.TcpProtocolMapper}) protocol mapper.
 * One of the design goals of this framework is the abstraction of this mechanism to enable customizations that do
 * not require modifications of existing application layer parsers (e.g. HTTP over a custom transport layer protocol
 * or multiple ports).
 * A first step to achieve this goal is the introduction of a generic
 * {@link at.jku.fim.rubanetra.protocol.mapper.TransportLayerMappingStrategy}.
 * However, currently only a port/service based approach has been implemented. Since Kraken apparently does not allow
 * the use of multiple ports per application layer parser (since the TCP/UDP-Mapper would override the bindings with
 * the default parser), a custom {@link at.jku.fim.rubanetra.config.model.PortSpecification} approach
 * was implemented on an application layer basis, i.e. a single application layer protocol may be bound to a
 * Kraken-application layer parser for a single, multiple, all or no port(s).
 * However, as soon as JNetPcap provides suitable application layer parsers for the currently supported protocols,
 * this mechanism should be replaced by JNetPcap-bindings which is more elegant and convenient.
 * This interface is implemented by
 * {@link at.jku.fim.rubanetra.config.model.impl.KrakenProtocolConfigurationBuilderImpl.KrakenProtocolConfigurationImpl}.
 */
public interface KrakenProtocolConfiguration {
    /**
     * The underlying source of all PCAP-packets for this application layer protocol. A listener will be attached
     * by default to listen for new {@link at.jku.fim.rubanetra.pcap.PcapActivity}-objects.
     * All parsed PCAP-packets will be either parsed by a Kraken-Application layer parser for this protocol or
     * discarded by the transport layer mapping strategy and/or port-specification.
     *
     * @return the {@link at.jku.fim.rubanetra.pcap.PcapHandler} to use as input stream for
     * {@link at.jku.fim.rubanetra.pcap.PcapActivity}-objects.
     * @see #getTransportLayerMappingStrategy()
     * @see #getPortSpecification()
     */
    public abstract PcapHandler<PcapActivityListener> getPcapHandler();

    /**
     * Represents a primitive protocol to Kraken-protocol decoder mapping mechanism.
     * A network protocol within this framework is always identified by a
     * {@link at.jku.fim.rubanetra.config.model.ProtocolId}. Therefore a mapping of a protocol to
     * an appropriate Kraken-protocol decoder/parser is needed. All
     * {@link at.jku.fim.rubanetra.protocol.KrakenBaseProtocol} implementations provide a method to
     * retrieve a suitable identifier
     * {@link at.jku.fim.rubanetra.protocol.KrakenBaseProtocol#getProtocolId()}.
     * These bindings may be configured by using a
     * {@link at.jku.fim.rubanetra.config.model.impl.KrakenProtocolConfigurationBuilderImpl}.
     *
     * @return the immutable copy of all application layer protocol bindings
     */
    public abstract Map<ProtocolId, KrakenBaseProtocol> getBoundProtocols();

    /**
     * The transport layer mapping strategy to use for this Kraken-application layer protocol decoder.
     *
     * @return the transport layer protocol mapping strategy identifier which may be used to obtain an instance of
     * a {@link at.jku.fim.rubanetra.protocol.mapper.TransportLayerMappingStrategy} by using
     * {@link at.jku.fim.rubanetra.protocol.mapper.KrakenTransportLayerMappingFactory}.
     */
    public abstract String getTransportLayerMappingStrategy();

    /**
     * The specification of a number of {@link at.jku.fim.rubanetra.config.model.PortRange}s or
     * {@link at.jku.fim.rubanetra.config.model.Port}s which filters packets at the transport layer
     * for this application layer protocol. A simpler alternative to this method is the use of
     * {@link at.jku.fim.rubanetra.config.model.PortSpecification.UniversalPortSpecification} and
     * by restricting all transport layer packets globally via a BP-Filter,
     * see {@link at.jku.fim.rubanetra.pcap.PcapHandler}.
     *
     * @return the {@link at.jku.fim.rubanetra.config.model.PortSpecification} as a transport layer
     * filter for this application layer protocol only
     * @see at.jku.fim.rubanetra.config.model.PortSpecification.UniversalPortSpecification
     */
    public abstract PortSpecification getPortSpecification();

    /**
     * This {@link at.jku.fim.rubanetra.protocol.mapper.KrakenTransportLayerMappingFactory} may be used
     * by the {@link at.jku.fim.rubanetra.protocol.KrakenBaseProtocol} parsers
     * to obtain instances of the configured {@link #getTransportLayerMappingStrategy()}.
     *
     * @return a factory to obtain instances of the transport layer mapping strategy
     */
    public abstract KrakenTransportLayerMappingFactory getTransportLayerMappingFactory();
}
