/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.config.model;

import java.util.HashMap;
import java.util.Map;

/**
 * A transport layer TCP/UDP port number.
 */
public class Port implements Comparable<Port> {
    /**
     * the lower bound of the available port number pool
     */
    public static final int LOWER_BOUND = 0;
    /**
     * the upper bound of the available port number pool
     */
    public static final int UPPER_BOUND = 65535;

    private static final Map<Integer, Port> numberPortMap = new HashMap<>();

    private int portNumber;


    /**
     * Private constructor, use {@link #create(int)} instead
     *
     * @throws java.lang.IllegalArgumentException if the port-number is not within the allowed range
     */
    private Port(int portNumber) {
        checkPortNumber(portNumber);
        this.portNumber = portNumber;
    }

    /**
     * Create a new {@link at.jku.fim.rubanetra.config.model.Port}
     *
     * @param portNumber the port number to use
     * @return a new {@link at.jku.fim.rubanetra.config.model.Port} or
     * an instance of an already instantiated port.
     * @throws java.lang.IllegalArgumentException if the port-number is not within the allowed range
     */
    public static Port create(int portNumber) {
        if (!numberPortMap.containsKey(portNumber)) {
            numberPortMap.put(portNumber, new Port(portNumber));
        }
        return numberPortMap.get(portNumber);
    }

    /**
     * This method is equivalent to calling {@link #create(int)}.
     * Create a new {@link at.jku.fim.rubanetra.config.model.Port}
     *
     * @param portNumber the port number to use
     * @return a new {@link at.jku.fim.rubanetra.config.model.Port} or
     * an instance of an already instantiated port.
     * @throws java.lang.IllegalArgumentException if the port-number is not within the allowed range
     */
    public static Port getPort(int portNumber) {
        return create(portNumber);
    }

    private void checkPortNumber(int portNumber) {
        if (portNumber < LOWER_BOUND || portNumber > UPPER_BOUND) {
            throw new IllegalArgumentException(String.format("Port %d is out of range", portNumber));
        }
    }

    /**
     * @return the represented port number
     */
    public int getPortNumber() {
        return portNumber;
    }

    @Override
    public int compareTo(Port otherPort) {
        return Integer.compare(getPortNumber(), otherPort.getPortNumber());
    }

    @Override
    public String toString() {
        return String.valueOf(portNumber);
    }

}
