/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.config.model;

/**
 * A port range is defined by an interval that is bound by two ports, i.e.
 * [lowerPort;upperPort], both bounds are inclusive.
 */
public class PortRange {
    private Port lowerPort;
    private Port higherPort;

    /**
     * Create a new port range, which represents all possible ports within the interval [lowerPort;upperPort], inclusive
     *
     * @param lowerPort  the lower bound of the range, must be less than higherPort.
     * @param higherPort the upper bound of the range, must be greater than lowerPort.
     * @throws java.lang.IllegalArgumentException if invalid arguments are passed
     */
    public PortRange(Port lowerPort, Port higherPort) {
        this.lowerPort = lowerPort;
        this.higherPort = higherPort;
        checkPortRange(lowerPort, higherPort);
    }

    private void checkPortRange(Port lowerPort, Port higherPort) {
        if (lowerPort == null || higherPort == null) {
            throw new IllegalArgumentException();
        }
        if (lowerPort.compareTo(higherPort) >= 0) {
            throw new IllegalArgumentException();
        }
    }

    /**
     * Checks whether the given port is within this range, i.e. port in [lowerPort;higherPort]
     * @param port to check
     * @return true, iff if port is in this range
     * @throws java.lang.IllegalArgumentException if the port is null
     */
    public boolean isWithinRange(Port port) {
        if (port == null) {
            throw new IllegalArgumentException();
        }

        return port.compareTo(getLowerPort()) >= 0 && port.compareTo(getHigherPort()) <= 0;
    }

    /**
     * @return the upper bound of this range
     */
    public Port getHigherPort() {
        return higherPort;
    }

    /**
     * Sets the upper bound of this range
     * @param higherPort the upper bound of the range, must be greater than lowerPort.
     */
    public void setHigherPort(Port higherPort) {
        checkPortRange(lowerPort, higherPort);
        this.higherPort = higherPort;
    }

    /**
     * @return the lower bound of this range
     */
    public Port getLowerPort() {
        return lowerPort;
    }

    /**
     * Sets the lower bound of the range
     * @param lowerPort the lower bound of the range, must be less than higherPort.
     */
    public void setLowerPort(Port lowerPort) {
        checkPortRange(lowerPort, higherPort);
        this.lowerPort = lowerPort;
    }
}
