/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.config.model;


import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 * This class constitutes a higher level specification of the mapping between an application layer protocol and
 * the associated transport layer (TCP/UDP) port numbers and/or port ranges.
 * It is utilized for a number of Kraken application layer protocol parser to transport layer mapping abstractions, e.g.
 * {@link at.jku.fim.rubanetra.protocol.mapper.TransportLayerMappingStrategy}, and configured by
 * an instance of {@link at.jku.fim.rubanetra.config.model.KrakenProtocolConfigurationBuilder} resulting
 * in a {@link at.jku.fim.rubanetra.config.model.KrakenProtocolConfiguration} object.
 * Note, that this implementation currently utilizes the Java 8 - Stream API for performing {@link at.jku.fim.rubanetra.config.model.PortRange}
 * bound checking via predicates, see {@link #contains(Port)}.
 * Additionally, the subclass {@link at.jku.fim.rubanetra.config.model.PortSpecification.UniversalPortSpecification}
 * was defined for convenience reasons.
 * This class is not thread-safe.
 */
public class PortSpecification {
    private final Map<Integer, Port> portMap;
    private final Set<PortRange> portRangeSet;

    public PortSpecification() {
        this.portMap = new HashMap<>();
        this.portRangeSet = new HashSet<>();
    }

    /**
     * Adds a single port to this specification.
     * A previously defined port with the same port number will be overwritten silently.
     *
     * @param singlePort the port to add to this specification
     * @throws java.lang.IllegalArgumentException if singlePort == null
     */
    public void add(Port singlePort) {
        if (singlePort == null) {
            throw new IllegalArgumentException();
        }
        this.portMap.put(singlePort.getPortNumber(), singlePort);
    }

    /**
     * Adds a new port-range to this specification.
     * A {@link java.util.HashSet} is used internally to store the port ranges.
     *
     * @param portRange the port range to add to this specification
     * @throws java.lang.IllegalArgumentException if the portRange is null
     */
    public void add(PortRange portRange) {
        if (portRange == null) {
            throw new IllegalArgumentException();
        }

        this.portRangeSet.add(portRange);
    }

    /**
     * This is a convenience method for {@link #contains(Port)}.
     * A new {@link at.jku.fim.rubanetra.config.model.Port} is constructed for the passed
     * integer argument.
     *
     * @param destinationPort the port number to check against this specification
     * @return false, if the given port was null or if it does not conform to the bounds of this specification,
     * true, iff the given port is backed by either a port-range or a single port number.
     * @throws java.lang.IllegalArgumentException if the port number is invalid
     * @see #contains(Port)
     */
    public boolean contains(int destinationPort) {
        Port p = Port.getPort(destinationPort);
        return contains(p);
    }

    /**
     * Performs a bound check according to the underlying specification.
     * A port adheres to the specification iff it was specified as a single {@link at.jku.fim.rubanetra.config.model.Port}
     * or any specified {@link at.jku.fim.rubanetra.config.model.PortRange} returns true by calling
     * {@link at.jku.fim.rubanetra.config.model.PortRange#isWithinRange(Port)}.
     * A Java 8 predicate check via the new stream-API is currently used for this check.
     *
     * @param destinationPort the port to check against this specification
     * @return false, if the given port was null or if it does not conform to the bounds of this specification,
     * true, iff the given port is backed by either a port-range or a single port number.
     */
    public boolean contains(Port destinationPort) {
        if (destinationPort == null) {
            return false;
        }
        return this.portMap.containsKey(destinationPort.getPortNumber())
                || portRangeSet.stream().anyMatch(portRange -> portRange.isWithinRange(destinationPort));
    }

    /**
     * A port specification that returns either always true or always false for any given port number.
     */
    public static class UniversalPortSpecification extends PortSpecification {
        public static final PortSpecification ACCEPT_ALL = new UniversalPortSpecification(true);
        public static final PortSpecification REJECT_ALL = new UniversalPortSpecification(false);
        private final boolean acceptPort;

        /**
         * Only two meaningful implementations available - either return true or false,
         * therefore a public constructor was deemed to be unnecessary.
         */
        private UniversalPortSpecification(boolean accept) {
            super();
            this.acceptPort = accept;
        }

        @Override
        public boolean contains(int destinationPort) {
            Port p = Port.getPort(destinationPort);
            return contains(p);
        }

        @Override
        public boolean contains(Port destinationPort) {
            return destinationPort != null && acceptPort;
        }
    }

}
