/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.config.model;


import at.jku.fim.rubanetra.output.OutputWriterStrategy;
import at.jku.fim.rubanetra.pcap.PcapFileHandler;
import org.kie.api.runtime.KieSession;

import java.io.File;
import java.io.OutputStream;
import java.util.Set;

/**
 * The generic specification of the system configuration itself.
 * This may be considered to constitute many of the essential settings, i.e. how and where to write the derived data, the Pcap
 * packet input stream, the input files, the Drools knowledge session.
 * Use a {@link at.jku.fim.rubanetra.config.model.impl.RubanetraSystemConfigurationBuilder} to obtain an implementation
 * instance programmatically.
 */
public interface RubanetraSystemConfiguration {

    /**
     * The main output stream that is used to write all derived information to.
     * The format of this data is defined by the {@link at.jku.fim.rubanetra.config.model.RubanetraSystemConfiguration.OutputFormat},
     * see{@link #getOutputFormat()}.
     * The point of time that defines when the data is actually written to this stream, is controlled by the
     * {@link at.jku.fim.rubanetra.output.OutputWriterStrategy}, see {@link #getOutputWriterStrategy()},
     * and the Drools knowledge session (i.e. the Rule engine and custom rules).
     *
     * @return the main output stream of the derived information
     */
    public OutputStream getOutputStream();

    /**
     * The format that is used to transform the derived data before it is written to the output-stream.
     *
     * @return the format to use for data transformation
     * @see #getOutputStream()
     */
    public OutputFormat getOutputFormat();

    /**
     * The writer strategy that is used to transform (using {@link #getOutputFormat()}) and finally write (using
     * {@link #getOutputStream()}) the derived data (delivered from {@link #getKieSession()}).
     *
     * @return the defined writer strategy
     * @see #getOutputStream()
     * @see #getOutputFormat()
     */
    public OutputWriterStrategy getOutputWriterStrategy();

    /**
     * The set of all defined protocol configurations, each representing the complete configuration of a Kraken protocol
     * parser/decoder, containing protocol mappings (ETHERNET->IPv4->...->...).
     *
     * @return the immutable set of defined Kraken protocol parser configurations
     */
    public Set<KrakenProtocolConfiguration> getProtocolSettings();

    /**
     * The Pcap packet input stream handler that is used to parse and react to Pcap-information
     * from the set of defined input files.
     *
     * @return the pcap input stream handler
     */
    public PcapFileHandler<?> getPcapHandler();

    /**
     * The set of input files possibly adhering to a PCAP-specification.
     *
     * @return the set of pcap input files used for parsing pcap-packets
     */
    public Set<File> getInputFiles();

    /**
     * The format of all input files (currently always PCAP)
     *
     * @return the common input format of all files
     */
    public InputFormat getInputFormat();

    /**
     * The Drools knowledge session reference, used to pass new facts to the rule engine
     * decoded by the protocol parsers.
     *
     * @return the Drools knowledge session reference
     */
    public KieSession getKieSession();

    /**
     * The enumeration of possible output formats, used for transforming derived data before it is written to the output
     * stream.
     *
     * @see #getOutputFormat()
     */
    public enum OutputFormat {
        /**
         * PLASO may be considered synonymous to XML,
         */
        PLASO,
        CSV,
        JSON,
        XML,
        /**
         * no-op, i.e. data will not be written at all
         */
        NOP, CALLBACK
    }

    /**
     * The enumeration of possible input formats, used for defining the appropriate input file handler without
     * guessing
     *
     * @see #getInputFormat()
     */
    public enum InputFormat {
        PCAP
    }


}
