/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.config.model.impl;

import at.jku.fim.rubanetra.config.model.KrakenProtocolConfiguration;
import at.jku.fim.rubanetra.config.model.KrakenProtocolConfigurationBuilder;
import at.jku.fim.rubanetra.config.model.PortSpecification;
import at.jku.fim.rubanetra.config.model.ProtocolId;
import at.jku.fim.rubanetra.pcap.PcapActivityListener;
import at.jku.fim.rubanetra.pcap.PcapHandler;
import at.jku.fim.rubanetra.protocol.BaseProtocol;
import at.jku.fim.rubanetra.protocol.KrakenBaseProtocol;
import at.jku.fim.rubanetra.protocol.mapper.KrakenTransportLayerMappingFactory;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.*;
import java.util.stream.Collectors;

/**
 * The default implementation of the {@link at.jku.fim.rubanetra.config.model.KrakenProtocolConfigurationBuilder}.
 * This class may be used to configure protocol parsers from the Kraken-library.
 */
public class KrakenProtocolConfigurationBuilderImpl implements KrakenProtocolConfigurationBuilder {
    private final Logger log = LoggerFactory.getLogger(getClass());
    private final ProtocolClassRegistry protocolRegistry;
    // a single protocol handler may forward data to multiple receivers, e.g. consider eth->ipv4, eth->ipv6
    private final Map<ProtocolId, Set<ProtocolId>> currentProtocolBindingMap;
    private PortSpecification currentProtocolPortSpec;
    private KrakenTransportLayerMappingFactory currentTransportLayerMappingFactory;
    private String currentTransportLayerStrategy;
    private PcapHandler<PcapActivityListener> currentPcapHandler;
    // the currently set protocol identifier for the protocol-config being built (arbitrary value, should be unique)
    private String currentProtocolIdentifier;

    /**
     * Creates a new builder instance using a default protocol lookup registry by calling
     * {@link ProtocolClassRegistry#createDefaultKrakenProtocolRegistry()}.
     */
    public KrakenProtocolConfigurationBuilderImpl() {
        this(ProtocolClassRegistry.createDefaultKrakenProtocolRegistry());
    }

    /**
     * Creates a new builder instance using a custom protocol lookup registry.
     * Be aware that all classes specified in the registry must define a default constructor.
     *
     * @param protocolRegistry the custom protocol registry to use for {@link at.jku.fim.rubanetra.config.model.ProtocolId}
     *                         lookup
     */
    public KrakenProtocolConfigurationBuilderImpl(ProtocolClassRegistry protocolRegistry) {
        if (protocolRegistry == null) {
            log.error("The protocol registry must not be null.");
            throw new IllegalArgumentException();
        }
        this.protocolRegistry = protocolRegistry;
        this.currentProtocolBindingMap = new HashMap<>();

        resetCurrentProtocolConfiguration();

        log.info("Decoder available for protocols: {}",
                protocolRegistry.getRegisteredProtocolIdSet().stream().map(ProtocolId::getName).collect(Collectors.toList()));
    }

    /*
     * (non-Javadoc)
     *
     * @see at.jku.fim.rubanetra.config.KrakenProtocolConfigurationBuilder#
     * resetCurrentProtocolConfiguration()
     */
    @Override
    public void resetCurrentProtocolConfiguration() {
        this.currentProtocolIdentifier = null;
        this.currentPcapHandler = null;
        this.currentProtocolPortSpec = new PortSpecification();
        this.currentProtocolBindingMap.clear();
    }

    /*
     * (non-Javadoc)
     *
     * @see at.jku.fim.rubanetra.config.KrakenProtocolConfigurationBuilder#bindProtocol
     *
     */
    @Override
    public KrakenProtocolConfigurationBuilder bindProtocol(String protocolBindFrom, String protocolBindTo) {
        if (!existsProtocolId(protocolBindFrom) || !existsProtocolId(protocolBindTo)) {
            throw new IllegalArgumentException(String.format("Unable to bind protocol %s to %s (unkown protocol(s))", protocolBindFrom, protocolBindTo));
        }
        ProtocolId bindFrom = getProtocolIdFromWildId(protocolBindFrom);
        ProtocolId bindTo = getProtocolIdFromWildId(protocolBindTo);
        log.debug("Adding binding from {} to {}", bindFrom, bindTo);

        if (bindFrom == bindTo) {
            // or should it be allowed?
            throw new IllegalArgumentException(String.format("It is not allowed to bind a protocol handler to itself (%s)", bindFrom));
        }

        if (!currentProtocolBindingMap.containsKey(bindFrom)) {
            currentProtocolBindingMap.put(bindFrom, new HashSet<>());
        }
        currentProtocolBindingMap.get(bindFrom).add(bindTo);

        return this;
    }

    private ProtocolId getProtocolIdFromWildId(String wildProtocolId) {
        if (wildProtocolId == null || wildProtocolId.isEmpty()) {
            throw new IllegalArgumentException("Unknown protocol or no protocol is mapped to a protocol id");
        }
        // This reduction does not handle multiple registrations to a similar protocolId-String well (it picks the first match found one)
        return protocolRegistry.getRegisteredProtocolIdSet().stream()
                .filter(protocolId -> protocolId.getProtocolId().equalsIgnoreCase(wildProtocolId)).findFirst().get();
    }

    private boolean existsProtocolId(String wildProtocolId) {
        return wildProtocolId != null
                && protocolRegistry.getRegisteredProtocolIdSet().stream()
                .anyMatch(protocolId -> protocolId.getProtocolId().equalsIgnoreCase(wildProtocolId));
    }

    /*
     * (non-Javadoc)
     *
     * @see at.jku.fim.rubanetra.config.KrakenProtocolConfigurationBuilder#
     * setPortSpecification
     * (at.jku.fim.rubanetra.config.PortSpecification)
     */
    @Override
    public KrakenProtocolConfigurationBuilder setPortSpecification(PortSpecification transportProtocolPortSpec) {
        this.currentProtocolPortSpec = transportProtocolPortSpec;

        return this;
    }

    @Override
    public KrakenProtocolConfigurationBuilder setTransportProtocolMappingStrategy(String strategy) {
        if (strategy == null) {
            throw new IllegalArgumentException("Invalid transport layer mapping strategy");
        }
        currentTransportLayerStrategy = strategy;
        return this;
    }

    @Override
    public KrakenProtocolConfigurationBuilder setTransportProtocolMappingFactory(KrakenTransportLayerMappingFactory mappingFactory) {
        if (mappingFactory == null) {
            throw new IllegalArgumentException("Invalid transport layer mapping factory");
        }
        currentTransportLayerMappingFactory = mappingFactory;
        return this;
    }

    @Override
    public KrakenProtocolConfigurationBuilder setPcapHandler(PcapHandler<PcapActivityListener> pcapHandler) {
        this.currentPcapHandler = pcapHandler;
        return this;
    }

    @Override
    public KrakenProtocolConfiguration buildProtocolConfiguration() {
        KrakenProtocolConfiguration setting = new KrakenProtocolConfigurationImpl();
        resetCurrentProtocolConfiguration();
        return setting;
    }

    /**
     * @return the currently set protocol identifier (may be null)
     */
    public String getCurrentProtocolIdentifier() {
        return currentProtocolIdentifier;
    }

    @Override
    public KrakenProtocolConfigurationBuilder setCurrentProtocolIdentifier(String currentProtocolIdentifier) {
        if (this.currentProtocolIdentifier != null) {
            // build is in progress
            throw new IllegalStateException(String.format("Protocol configuration for %s is incomplete.", this.currentProtocolIdentifier));
        }
        this.currentProtocolIdentifier = currentProtocolIdentifier;
        return this;
    }

    /**
     * This inner class represents a final protocol configuration object which will be used to initialize Kraken protocol
     * handlers.
     * There are no setter methods, since all fields are now considered to be immutable and must not change during the
     * lifetime of the object.
     * Instances of this class may be obtained by using and configuring a
     * {@link at.jku.fim.rubanetra.config.model.impl.KrakenProtocolConfigurationBuilderImpl}.
     * This implementation relies on the fact that all registered protocol handlers (even custom handlers) define
     * a default constructor. All handlers may utilize information of this configuration by overriding the
     * {@link at.jku.fim.rubanetra.protocol.KrakenBaseProtocol#initialize(at.jku.fim.rubanetra.config.model.KrakenProtocolConfiguration)}
     * method, which will be called as soon as possible after construction.
     * Further note, that this implementation tries to reduce the number of protocol handler instances. The current
     * mechanism first takes an unbound protocol id of the binding map, constructs an instance and saves this reference
     * for the provided protocol id.
     * <p>
     * Constraints:
     * The call of the initialization method is in general unordered. However, it is always ensured that
     * for a given binding 'from'->'to', the protocol handler for 'from' is constructed and initialized before 'to'.
     * Double bindings 'from'->'to' + 'to'->'from' _should_ be possible, but must be handled carefully (untested).
     * All bindings are performed after initialization by calling 'from'.bind('to'). All handlers are responsible
     * to take care of the concrete binding and must handle unsupported bindings appropriately.
     *
     * @see at.jku.fim.rubanetra.protocol.BaseProtocol#bind(at.jku.fim.rubanetra.protocol.BaseProtocol)
     */
    public class KrakenProtocolConfigurationImpl implements KrakenProtocolConfiguration {
        private final PortSpecification transportLayerPortSpec;
        private final String transportLayerMappingStrategy;
        private final KrakenTransportLayerMappingFactory transportLayerMappingFactory;
        private final PcapHandler<PcapActivityListener> pcapHandler;
        private final Map<ProtocolId, KrakenBaseProtocol> registeredProtocols;

        /**
         * keep private
         */
        private KrakenProtocolConfigurationImpl() {
            this.pcapHandler = currentPcapHandler;
            this.transportLayerPortSpec = currentProtocolPortSpec;
            this.transportLayerMappingStrategy = currentTransportLayerStrategy;
            this.transportLayerMappingFactory = currentTransportLayerMappingFactory;

            this.registeredProtocols = new HashMap<>();

            check();
            log.debug("Bindings to process: {}", currentProtocolBindingMap);

            for (ProtocolId protocolId : currentProtocolBindingMap.keySet()) {
                KrakenBaseProtocol baseProtocolFrom;
                if (registeredProtocols.containsKey(protocolId)) {
                    // instance was already constructed and initialized
                    baseProtocolFrom = registeredProtocols.get(protocolId);
                } else {
                    // construct and initialize a new instance
                    baseProtocolFrom = createBaseProtocolInstance(protocolId);
                    baseProtocolFrom.initialize(this);
                    // remember this instance
                    registeredProtocols.put(protocolId, baseProtocolFrom);
                }
                log.debug("[BIND_FROM] ProtocolId: {}, Class: {}", protocolId, baseProtocolFrom.getClass());

                // bind this protocol to ...
                for (ProtocolId bindToId : currentProtocolBindingMap.get(protocolId)) {
                    KrakenBaseProtocol baseProtocolTo;
                    if (registeredProtocols.containsKey(bindToId)) {
                        // instance was already constructed and initialized
                        baseProtocolTo = registeredProtocols.get(bindToId);
                    } else {
                        // construct and initialize a new instance
                        baseProtocolTo = createBaseProtocolInstance(bindToId);
                        baseProtocolTo.initialize(this);
                        // remember this instance
                        registeredProtocols.put(bindToId, baseProtocolTo);
                    }

                    log.debug("[BIND_TO] ProtocolId: {}, Class: {}", bindToId, baseProtocolTo.getClass());
                    // perform the actual binding
                    baseProtocolFrom.bind(baseProtocolTo);
                }
            }
        }

        //ENHANCEMENT: do not rely on the fact that there is always a default constructor
        private KrakenBaseProtocol createBaseProtocolInstance(ProtocolId protocolId) {
            IllegalStateException unknownProtocol = new IllegalStateException(String.format("Unable to instantiate a BaseProtocol class from protocol identifier: %s", protocolId));
            Class<? extends BaseProtocol> protocolClass = protocolRegistry.getRegisteredProtocolClass(protocolId);
            if (protocolClass == null) {
                throw unknownProtocol;
            }
            if (KrakenBaseProtocol.class.isAssignableFrom(protocolClass)) {
                try {
                    return protocolClass.asSubclass(KrakenBaseProtocol.class).newInstance();
                } catch (InstantiationException | IllegalAccessException e) {
                    log.error("Unable to instantiate protocol handler {} - does it provide a default constructor?", protocolClass, e);
                    throw unknownProtocol;
                }
            } else {
                log.error("This Protocol Configuration Builder supports KrakenBaseProtocol sub-classes only.");
                throw new IllegalArgumentException("This Protocol Configuration Builder supports KrakenBaseProtocol sub-classes only.");
            }
        }

        /**
         * @return an immutable map consisting of all protocol id to handler bindings
         */
        public Map<ProtocolId, KrakenBaseProtocol> getBoundProtocols() {
            return Collections.unmodifiableMap(registeredProtocols);
        }

        private void check() {
            if (transportLayerMappingFactory == null || transportLayerMappingStrategy == null || pcapHandler == null || transportLayerPortSpec == null) {
                log.error("Factory {}, strategy {}, pcapHandler {}, portSpec {}", transportLayerMappingFactory, transportLayerMappingStrategy, pcapHandler, transportLayerPortSpec);
                throw new IllegalStateException("Unable to build a KrakenProtocolConfiguration.");
            }
        }

        @Override
        public String getTransportLayerMappingStrategy() {
            return this.transportLayerMappingStrategy;
        }

        @Override
        public PortSpecification getPortSpecification() {
            return this.transportLayerPortSpec;
        }

        @Override
        public PcapHandler<PcapActivityListener> getPcapHandler() {
            return this.pcapHandler;
        }

        @Override
        public KrakenTransportLayerMappingFactory getTransportLayerMappingFactory() {
            return transportLayerMappingFactory;
        }
    }

}
