/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.config.model.impl;

import at.jku.fim.rubanetra.config.model.RubanetraSystemConfiguration;
import at.jku.fim.rubanetra.output.OutputWriterStrategy;
import at.jku.fim.rubanetra.output.impl.CsvOutputWriterStrategy;
import at.jku.fim.rubanetra.output.impl.JsonOutputWriterStrategy;
import at.jku.fim.rubanetra.output.impl.NopOutputWriterStrategy;
import at.jku.fim.rubanetra.output.impl.XmlOutputWriterStrategy;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.OutputStream;

/**
 * A simple factory constructing {@link at.jku.fim.rubanetra.output.OutputWriterStrategy} objects
 * for a known {@link at.jku.fim.rubanetra.config.model.RubanetraSystemConfiguration.OutputFormat}
 */
public class OutputStrategyFactory {
    private static final Logger log = LoggerFactory.getLogger(OutputStrategyFactory.class);
    private static final NopOutputWriterStrategy nopOutputWriterStrategy = new NopOutputWriterStrategy();

    /**
     * Returns an object capable of transforming data into the provided {@link at.jku.fim.rubanetra.config.model.RubanetraSystemConfiguration.OutputFormat}
     *
     * @param outputFormat the format to use for transforming data before writing to out
     * @param out          the output stream to write to
     * @return if any provided argument is null, a {@link at.jku.fim.rubanetra.output.impl.NopOutputWriterStrategy},
     * or a capable strategy object if an appropriate class is known
     */
    public static OutputWriterStrategy createOutputWriterStrategy(RubanetraSystemConfiguration.OutputFormat outputFormat, OutputStream out) {
        if (outputFormat == null || out == null) {
            log.warn("Unable to create a suitable OutputWriterStrategy - null argument was passed.");
            return nopOutputWriterStrategy;
        }

        switch (outputFormat) {
            case PLASO:
            case XML:
                return new XmlOutputWriterStrategy(out);
            case JSON:
                return new JsonOutputWriterStrategy(out);
            case CSV:
                try {
                    return new CsvOutputWriterStrategy(out);
                } catch (IOException e) {
                    log.warn("Unable to create a suitable OutputWriterStrategy for format {}, using NOP instead", outputFormat);
                    return nopOutputWriterStrategy;
                }
            case NOP:
                return nopOutputWriterStrategy;
            default:
                log.warn("Unable to create a suitable OutputWriterStrategy for format {}, using NOP instead", outputFormat);
                return nopOutputWriterStrategy;
        }
    }
}
