/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.config.model.impl;

import at.jku.fim.rubanetra.config.model.ProtocolId;
import at.jku.fim.rubanetra.protocol.BaseProtocol;
import at.jku.fim.rubanetra.protocol.KrakenBaseProtocol;
import at.jku.fim.rubanetra.protocol.handler.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.*;

/**
 * A representation of a simple protocol identifier ({@link at.jku.fim.rubanetra.config.model.ProtocolId})
 * to an arbitrary {@link at.jku.fim.rubanetra.protocol.BaseProtocol} class.
 * It is expected that all registered classes provide a default constructor.
 *
 * @see at.jku.fim.rubanetra.config.model.impl.KrakenProtocolConfigurationBuilderImpl
 */
public class ProtocolClassRegistry {
    private final Logger log = LoggerFactory.getLogger(getClass());
    private final Map<ProtocolId, Class<? extends BaseProtocol>> protocolRegistry;
    private final Map<String, ProtocolId> protocolIdMap;

    /**
     * create a new instance
     */
    public ProtocolClassRegistry() {
        protocolIdMap = new HashMap<>();
        protocolRegistry = new HashMap<>();
    }

    /**
     * Register a protocol identifier to a protocol handler class, which must provide a default constructor
     *
     * @param protocolId           a protocol identifier, which should be advertised by the protocolHandlerClazz as well
     * @param protocolHandlerClazz a class that is able to parse the protocol identified by the protocol id and
     *                             provides a default constructor
     * @throws java.lang.IllegalArgumentException if any argument is null or the protocol id was already registered
     */
    public void registerHandler(ProtocolId protocolId, Class<? extends BaseProtocol> protocolHandlerClazz) {
        if (protocolId == null || protocolId.getProtocolId() == null || protocolHandlerClazz == null) {
            log.error("The protocol id and the protocol handler must not be null.");
            throw new IllegalArgumentException();
        }
        if (protocolRegistry.containsKey(protocolId) || protocolIdMap.containsKey(protocolId.getProtocolId())) {
            log.error("A protocol handler for the given protocol id {} was already registered before.", protocolId);
            throw new IllegalArgumentException();
        }
        log.debug("Registering {} to class {}.", protocolId, protocolHandlerClazz);
        protocolIdMap.put(protocolId.getProtocolId(), protocolId);
        protocolRegistry.put(protocolId, protocolHandlerClazz);
    }

    /**
     * Delete a registry record, does nothing if the protocol id was not registered
     * @param protocolId the identifier for the handler to delete from the registry
     */
    public void unregisterHandler(ProtocolId protocolId) {
        if(protocolId!=null) {
            protocolRegistry.remove(protocolId);
            protocolIdMap.remove(protocolId.getProtocolId());
        }
    }

    /**
     * Return the registered class for a given protocol identifier
     *
     * @param id the identifier that was registered to this registry
     * @return a class that should (unchecked) provide a default constructor and should be able to parse
     * data adhering to the protocol identified by the protocol id
     */
    public Class<? extends BaseProtocol> getRegisteredProtocolClass(ProtocolId id) {
        return protocolRegistry.get(id);
    }

    /**
     * @return all currently registered protocol identifiers
     */
    public Set<ProtocolId> getRegisteredProtocolIdSet() {
        return Collections.unmodifiableSet(protocolRegistry.keySet());
    }

    /**
     * @return all currently registered protocol handler classes (duplicates may be possible)
     */
    public Collection<Class<? extends BaseProtocol>> getRegisteredProtocolClassList() {
        return Collections.unmodifiableCollection(protocolRegistry.values());
    }

    private static ProtocolClassRegistry defaultKrakenProtocolRegistry;

    /**
     * @return a default registry containing all originally known protocol ids and handlers.
     * It will never contain custom protocol handlers unless they are added explicitly afterwards.
     */
    public static ProtocolClassRegistry createDefaultKrakenProtocolRegistry() {
        if (defaultKrakenProtocolRegistry != null) {
            return defaultKrakenProtocolRegistry;
        }
        defaultKrakenProtocolRegistry = new ProtocolClassRegistry();

        // it would be nice to implement some kind of class auto detection mechanism (Reflection maybe?)...
        KrakenBaseProtocol[] krakenProtocolHandlers = {
                new KrakenArpHandler(),
                new KrakenEthernetProtocolHandler(),
                new KrakenIpv4Handler(),
                new KrakenIpv6Handler(),
                new KrakenIcmpv4Handler(),
                new KrakenIcmpv6Handler(),
                new KrakenTcpHandler(),
                new KrakenUdpHandler(),
                new KrakenHttpHandler(),
                new KrakenDnsProtocolHandler(),
                new KrakenMsnHandler(),
                new KrakenSmtpHandler(),
                new KrakenDhcpHandler(),
                new KrakenFtpHandler(),
                new KrakenNetbiosHandler(),
                new KrakenPop3Handler(),
                new KrakenSnmpHandler(),
                new KrakenTelnetHandler()
        };

        for (KrakenBaseProtocol p : krakenProtocolHandlers) {
            defaultKrakenProtocolRegistry.registerHandler(p.getProtocolId(), p.getClass());
        }

        return defaultKrakenProtocolRegistry;
    }
}
