/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.config.model.impl;

import at.jku.fim.rubanetra.config.RubanetraSystemConfigurationBaseListener;
import at.jku.fim.rubanetra.config.RubanetraSystemConfigurationLexer;
import at.jku.fim.rubanetra.config.RubanetraSystemConfigurationParser;
import at.jku.fim.rubanetra.config.RubanetraSystemConfigurationParser.*;
import at.jku.fim.rubanetra.config.model.Port;
import at.jku.fim.rubanetra.config.model.PortRange;
import at.jku.fim.rubanetra.config.model.PortSpecification;
import at.jku.fim.rubanetra.protocol.mapper.KrakenTransportLayerMappingFactory;
import org.antlr.v4.runtime.*;
import org.antlr.v4.runtime.misc.NotNull;
import org.antlr.v4.runtime.tree.ParseTree;
import org.antlr.v4.runtime.tree.ParseTreeWalker;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;
import java.util.List;

/**
 * This class represents the implementation of the listener interface that was created previously by the ANTLRv4 parser.
 * One of the first step of the system is to parse its own configuration. This implementation may be used to parse
 * all main configuration parts, i.e. input/output, protocol and Drools specific settings can be handled by this
 * class.
 * <p>
 * A {@link at.jku.fim.rubanetra.config.model.impl.RubanetraSystemConfigurationBuilder} is mainly utilized,
 * including a {@link at.jku.fim.rubanetra.config.model.impl.KrakenProtocolConfigurationBuilderImpl}
 * to perform the construction of the main configuration that may be used to further initialize the system.
 * In order to allow further customizations this listener does not invoke the final call to
 * {@link RubanetraSystemConfigurationBuilder#create()}, therefore it is the objects instantiator's obligation to ensure
 * completeness of the configuration and the construction
 * of the {@link at.jku.fim.rubanetra.config.model.RubanetraSystemConfiguration}.
 * <p>
 * An appropriate ANTLR parser object can be created by calling the static method {@link #createConfigurationFileParser(java.io.File)}
 * by passing a valid a configuration file.
 * The parser should then be pointed to the root node, i.e.:
 * <p>
 * {@code RubanetraSystemConfigurationParser systemConfigParser = RubanetraSystemConfigurationBaseListenerImpl.createConfigurationFileParser(mainConfigFile);}
 * {@code RubanetraSystemConfigurationParser.SystemContext systemContext = systemConfigParser.system();}
 * <p>
 * And finally a {@link org.antlr.v4.runtime.tree.ParseTreeWalker} may then be used to traverse the parse tree of the
 * general configuration:
 * <p>
 * {@code ParseTreeWalker.DEFAULT.walk(systemConfigBaseListener, systemContext.generalConfig(0));}
 */
public class RubanetraSystemConfigurationBaseListenerImpl extends RubanetraSystemConfigurationBaseListener {
    private final Logger log = LoggerFactory.getLogger(getClass());
    private final RubanetraSystemConfigurationBuilder systemConfigBuilder;
    private final KrakenProtocolConfigurationBuilderImpl protocolSettingBuilder;

    /**
     * Creates a new listener instance, which may be used to react to parsing event generated by a
     * {@link at.jku.fim.rubanetra.config.RubanetraSystemConfigurationParser}.
     *
     * @param systemConfigBuilder an instance of a builder which will be used to setup the system configuration,
     *                            may not be null
     * @throws java.lang.IllegalArgumentException if parser errors are encountered
     */
    public RubanetraSystemConfigurationBaseListenerImpl(RubanetraSystemConfigurationBuilder systemConfigBuilder) {
        if (systemConfigBuilder == null) {
            log.error("The passed system configuration may not be null");
            throw new IllegalArgumentException();
        }
        this.systemConfigBuilder = systemConfigBuilder;
        protocolSettingBuilder = new KrakenProtocolConfigurationBuilderImpl();
    }

    public static RubanetraSystemConfigurationParser createConfigurationFileParser(File configurationFile) {
        final Logger log = LoggerFactory.getLogger(RubanetraSystemConfigurationParser.class);
        ANTLRFileStream fileStream;
        try {
            log.debug("Trying to create an ANTLRFileStream for file {}", configurationFile);
            fileStream = new ANTLRFileStream(configurationFile.getCanonicalPath());
        } catch (IOException e) {
            log.error("Invalid configuration file: {}", configurationFile, e);
            throw new IllegalArgumentException(e);
        }
        log.debug("Creating a system configuration lexer.");
        RubanetraSystemConfigurationLexer systemConfigLexer = new RubanetraSystemConfigurationLexer(fileStream);
        systemConfigLexer.removeErrorListeners();
        CommonTokenStream configTokens = new CommonTokenStream(systemConfigLexer);
        log.debug("Creating a system configuration parser.");
        RubanetraSystemConfigurationParser systemConfigurationParser = new RubanetraSystemConfigurationParser(configTokens);
        systemConfigurationParser.removeErrorListeners();
        systemConfigurationParser.addErrorListener(new BaseErrorListener() {
            @Override
            public void syntaxError(Recognizer<?, ?> recognizer, Object offendingSymbol, int line, int charPositionInLine, String msg, RecognitionException e) {
                String sourceName = recognizer.getInputStream().getSourceName();
                if (!sourceName.isEmpty()) {
                    sourceName = String.format("%s: ", sourceName);
                }

                log.error("Configuration syntax error: {} line {}:{} {}", sourceName, line, charPositionInLine, msg);
            }
        });

        return systemConfigurationParser;
    }

    @Override
    public void enterSystem(@NotNull SystemContext ctx) {
        log.debug("Entered system node");
    }

    @Override
    public void exitSystem(@NotNull SystemContext ctx) {
        // we could call systemConfigBuilder.build() here, instead the user is expected to call it
        log.debug("Exiting system node");
    }

    /**
     * This method will be called by a {@link org.antlr.v4.runtime.tree.ParseTreeWalker} only if the
     * protocol configuration section was specified within the main system
     * configuration file and not in a separate file.
     */
    @Override
    public void enterProtocolConfig(@NotNull ProtocolConfigContext ctx) {
        // Use another listener to handle the complex protocol configuration
        // settings
        //ParseTreeWalker.DEFAULT.walk(this, ctx);
        log.debug("Entered protocol config node");
    }

    @Override
    public void enterProtocolSetting(@NotNull ProtocolSettingContext ctx) {
        log.debug("Entered protocol setting node");
        ProtocolHeaderContext protocolHeaderContext = ctx.protocolHeader();
        List<PortSpecificationContext> portSpecificationContext = ctx.portSpecification();
        List<ProtocolBindingContext> protocolBindingContexts = ctx.protocolBinding();
        List<TransportLayerMappingSettingContext> mappingStrategyContexts = ctx.transportLayerMappingSetting();

        if (protocolHeaderContext == null) {
            throw new IllegalArgumentException("Missing protocol identifier.");
        }
        String protocolIdentifier = protocolHeaderContext.protocolId().getText();
        protocolSettingBuilder.setCurrentProtocolIdentifier(protocolIdentifier);

        if (portSpecificationContext.isEmpty()) {
            protocolSettingBuilder.setPortSpecification(PortSpecification.UniversalPortSpecification.ACCEPT_ALL);
        } else if (portSpecificationContext.size() > 1) {
            throw new IllegalArgumentException(String.format("Multiple port specifications encountered, but only one is permitted (protocol setting: %s)", protocolIdentifier));

        } else {
            portSpecificationContext.get(0).enterRule(this);
        }

        for (ProtocolBindingContext bindingContext : protocolBindingContexts) {
            List<ProtocolIdContext> protocolIdContexts = bindingContext.protocolId();
            if (protocolIdContexts.size() != 2)
                throw new IllegalArgumentException(String.format("Invalid protocol binding encountered (protocol setting: %s, binding: %s)", protocolIdentifier, protocolIdContexts));
            protocolSettingBuilder.bindProtocol(bindingContext.protocolId(0).ID().getText(), bindingContext.protocolId(1).ID().getText());
        }

        protocolSettingBuilder.setTransportProtocolMappingFactory(systemConfigBuilder.getTransportLayerMappingFactory());
        if (mappingStrategyContexts.isEmpty()) {
            protocolSettingBuilder.setTransportProtocolMappingStrategy(KrakenTransportLayerMappingFactory.DESTINATION_PORT_STRATEGY);
        } else if (mappingStrategyContexts.size() > 1) {
            throw new IllegalArgumentException(String.format("Multiple transport layer mapping strategies encountered (protocol setting: %s, binding: %s)", protocolIdentifier, mappingStrategyContexts));
        } else {
            protocolSettingBuilder.setTransportProtocolMappingStrategy(mappingStrategyContexts.get(0).transportLayerMappingStrategy().getText().toUpperCase());
        }

        protocolSettingBuilder.setPcapHandler(systemConfigBuilder.getPcapHandler());
        systemConfigBuilder.addProtocolSetting(protocolSettingBuilder.buildProtocolConfiguration());
        protocolSettingBuilder.resetCurrentProtocolConfiguration();
    }

    @Override
    public void enterPortSpecification(@NotNull PortSpecificationContext ctx) {
        PortSpecification portSpec;
        if (ctx.ALL() != null) {
            portSpec = PortSpecification.UniversalPortSpecification.ACCEPT_ALL;
        } else if (ctx.NONE() != null) {
            portSpec = PortSpecification.UniversalPortSpecification.REJECT_ALL;
        } else {
            portSpec = new PortSpecification();
        }

        MultiplePortsContext multiplePortsContext = ctx.multiplePorts();
        if (multiplePortsContext != null) {
            for (PortRangeContext portRangeContext : multiplePortsContext.portRange()) {
                List<SinglePortContext> singlePortContexts = portRangeContext.singlePort();
                switch (singlePortContexts.size()) {
                    case 1:
                        portSpec.add(parsePortNumber(singlePortContexts.get(0)));
                        break;
                    case 2:
                        portSpec.add(parsePortRange(singlePortContexts.get(0), singlePortContexts.get(1)));
                        break;
                    default:
                        throw new IllegalArgumentException("Invalid port specification encountered");
                }
            }

        }

        protocolSettingBuilder.setPortSpecification(portSpec);
    }

    private PortRange parsePortRange(SinglePortContext portContextA, SinglePortContext portContextB) {
        Port lowerPort, upperPort;
        lowerPort = parsePortNumber(portContextA);
        upperPort = parsePortNumber(portContextB);
        if (lowerPort.compareTo(upperPort) > 0) {
            Port tmpPort = lowerPort;
            lowerPort = upperPort;
            upperPort = tmpPort;
        }

        return new PortRange(lowerPort, upperPort);
    }

    private Port parsePortNumber(SinglePortContext singlePortContext) {
        return Port.create(Integer.parseUnsignedInt(singlePortContext.INT().getText()));
    }

    /**
     * **********************************************************************
     * General configuration options will be parsed by the following methods.
     * Semantic checks will be performed by the
     * {@link at.jku.fim.rubanetra.config.model.impl.RubanetraSystemConfigurationBuilder} only.
     * **********************************************************************
     */

    @Override
    public void enterBpfFilterString(@NotNull BpfFilterStringContext ctx) {
        String bpfFilter = unquoteString(ctx.STRING().getText());
        // syntax check will be performed by the native BPF compiler
        systemConfigBuilder.setBpfFilter(bpfFilter);
    }

    @Override
    public void enterBpfNetmask(@NotNull BpfNetmaskContext ctx) {
        int bpfNetmask = Integer.parseUnsignedInt(ctx.INT().getText());
        systemConfigBuilder.setBpfNetmask(bpfNetmask);
    }

    @Override
    public void enterBpfOptimize(@NotNull BpfOptimizeContext ctx) {
        boolean bpfOptimize = false;
        if (ctx.TRUE() != null) {
            bpfOptimize = true;
        }
        systemConfigBuilder.setBpfOptimize(bpfOptimize);
    }

    @Override
    public void enterDroolsBaseModelName(@NotNull DroolsBaseModelNameContext ctx) {
        String droolsBaseModelName = unquoteString(ctx.getText());
        systemConfigBuilder.setDroolsBaseModelName(droolsBaseModelName);
    }

    @Override
    public void enterDroolsSessionModelName(@NotNull DroolsSessionModelNameContext ctx) {
        String droolsSessionModelName = unquoteString(ctx.getText());
        systemConfigBuilder.setDroolsSessionModelName(droolsSessionModelName);
    }

    /**
     * This method will be called by a {@link org.antlr.v4.runtime.tree.ParseTreeWalker} only if the
     * optional protocol configuration file setting was specified by the user in
     * the main system configuration file. It will be assumed that there will
     * not follow a separate protocol configuration section within the main
     * configuration file, i.e. the
     * {@link #enterProtocolConfig(at.jku.fim.rubanetra.config.RubanetraSystemConfigurationParser.ProtocolConfigContext)} will not be called by
     * the {@link org.antlr.v4.runtime.tree.ParseTreeWalker} (otherwise the user has specified two
     * separate protocol configuration sections, one within the main system
     * configuration and one in a different file, which is not allowed).
     */
    @Override
    public void enterProtocolConfigurationFileSetting(@NotNull ProtocolConfigurationFileSettingContext ctx) {
        String protocolConfigurationString = unquoteString(ctx.filePath().getText());
        // check file path syntax, i.e. an InvalidPathException may be thrown
        systemConfigBuilder.setProtocolConfigurationPath(protocolConfigurationString);

        RubanetraSystemConfigurationParser protocolConfigParser = createConfigurationFileParser(systemConfigBuilder.getProtocolConfigurationFile());
        ParseTree parseTree = protocolConfigParser.protocolConfig(); // begin parsing at the protocol configuration node

        if (parseTree == null) {
            throw new IllegalArgumentException("The specified protocol configuration file does not contain the protocol configuration section");
        }

        ParseTreeWalker.DEFAULT.walk(this, parseTree);
    }

    @Override
    public void enterDroolsConfigurationFile(@NotNull DroolsConfigurationFileContext ctx) {
        String droolsConfigurationString = unquoteString(ctx.filePath().getText());
        systemConfigBuilder.setDroolsConfigurationPath(droolsConfigurationString);

        // we do not need to parse the drools configuration file, due to the already included Drools Parser (also based on ANTLR)
    }

    @Override
    public void enterInputConfigurationFileSetting(@NotNull InputConfigurationFileSettingContext ctx) {
        String inputConfigurationString = unquoteString(ctx.filePath().getText());
        systemConfigBuilder.setInputConfigurationPath(inputConfigurationString);

        RubanetraSystemConfigurationParser inputConfigurationParser = createConfigurationFileParser(systemConfigBuilder.getInputConfigurationFile());
        ParseTree inputConfigurationParseTree = inputConfigurationParser.inputConfig();

        if (inputConfigurationParseTree == null) {
            throw new IllegalArgumentException("The specified input configuration file does not contain the input configuration section");
        }

        ParseTreeWalker.DEFAULT.walk(this, inputConfigurationParseTree);
    }

    @Override
    public void enterOutputConfigurationFileSetting(@NotNull OutputConfigurationFileSettingContext ctx) {
        String outputConfigurationString = unquoteString(ctx.filePath().getText());
        systemConfigBuilder.setOutputConfigurationPath(outputConfigurationString);

        RubanetraSystemConfigurationParser outputConfigurationParser = createConfigurationFileParser(systemConfigBuilder.getOutputConfigurationFile());
        ParseTree outputConfigurationParseTree = outputConfigurationParser.outputConfig();

        if (outputConfigurationParseTree == null) {
            throw new IllegalArgumentException("The specified output configuration file does not contain the output configuration section");
        }

        ParseTreeWalker.DEFAULT.walk(this, outputConfigurationParseTree);
    }

    /**
     * **********************************************************************
     * Dynamic I/O configuration options will be parsed by the following methods.
     * Semantic checks will be performed by the
     * {@link at.jku.fim.rubanetra.config.model.impl.RubanetraSystemConfigurationBuilder} only.
     * **********************************************************************
     */

    @Override
    public void exitInputConfig(@NotNull InputConfigContext ctx) {
//        systemConfigBuilder.createPcapHandler();
    }

    @Override
    public void enterInputSpecification(@NotNull InputSpecificationContext ctx) {
        List<FilePathContext> filePathContexts = ctx.filePath();
        if (filePathContexts == null || filePathContexts.isEmpty()) {
            throw new IllegalArgumentException("At least one input file must be specified.");
        }

        for (FilePathContext f : filePathContexts) {
            systemConfigBuilder.addInputPath(unquoteString(f.getText()));
        }
    }

    @Override
    public void enterInputFormat(@NotNull InputFormatContext ctx) {
        systemConfigBuilder.setInputFormat(ctx.getText());
    }

    @Override
    public void enterSortByFirstTimestamp(@NotNull SortByFirstTimestampContext ctx) {
            systemConfigBuilder.setSortByFirstTimestamp(ctx.TRUE()!=null);
    }

    @Override
    public void enterOutputSpecification(@NotNull OutputSpecificationContext ctx) {
        if (ctx.STDOUT() != null) {
            systemConfigBuilder.setOutputStream(System.out);
        } else {
            systemConfigBuilder.setOutputFile(unquoteString(ctx.filePath().getText()));
        }
    }

    @Override
    public void enterOutputFormat(@NotNull OutputFormatContext ctx) {
        systemConfigBuilder.setOutputFormat(ctx.getText());
    }

    private static String unquoteString(String quotedString) {
        if (quotedString == null || quotedString.length() < 2) {
            return quotedString;
        }
        if (quotedString.startsWith("\"") && quotedString.endsWith("\"")) {
            return quotedString.substring(1, quotedString.length() - 1);
        }

        return quotedString;
    }
}
