/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.config.model.impl;

import at.jku.fim.rubanetra.config.model.KrakenProtocolConfiguration;
import at.jku.fim.rubanetra.config.model.RubanetraSystemConfiguration;
import at.jku.fim.rubanetra.config.model.RubanetraSystemConfiguration.OutputFormat;
import at.jku.fim.rubanetra.drools.DroolsKnowledgeBaseFactory;
import at.jku.fim.rubanetra.output.OutputWriterStrategy;
import at.jku.fim.rubanetra.pcap.PcapActivityListener;
import at.jku.fim.rubanetra.pcap.PcapFileHandler;
import at.jku.fim.rubanetra.pcap.PcapFilter;
import at.jku.fim.rubanetra.pcap.PcapHandler;
import at.jku.fim.rubanetra.protocol.mapper.KrakenTransportLayerMappingFactory;
import org.kie.api.runtime.KieContainer;
import org.kie.api.runtime.KieSession;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.*;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.Set;

import static at.jku.fim.rubanetra.config.model.RubanetraSystemConfiguration.InputFormat;

/**
 * This class may be used to setup the essential configuration of the Drools rule engine, the JNetPcap file parser
 * including optional Berkeley Packet Filter support, the Kraken Pcap and network protocol parser, and finally
 * general input and output options.
 * The builder produces a {@link at.jku.fim.rubanetra.config.model.RubanetraSystemConfiguration} implementation,
 * that may be used to create a {@link at.jku.fim.rubanetra.drools.DroolsKrakenProtocolHandler}
 * which in turn is capable of managing the interaction of the Drools rule engine and the Kraken protocol decoder.
 *
 * @see at.jku.fim.rubanetra.drools.DroolsKrakenProtocolHandler
 * @see at.jku.fim.rubanetra.config.ConfigurationController
 * @see at.jku.fim.rubanetra.config.model.impl.RubanetraSystemConfigurationBuilder.RubanetraSystemConfigurationImpl
 * @see at.jku.fim.rubanetra.config.model.impl.RubanetraSystemConfigurationBaseListenerImpl
 */
public class RubanetraSystemConfigurationBuilder {
    private final Logger log = LoggerFactory.getLogger(getClass());

    /**
     * The system configuration may be split up into several files (optional)
     */
    private File protocolConfigurationFile;
    private File droolsConfigurationFile;
    private File inputConfigurationFile;
    private File outputConfigurationFile;

    /**
     * The set of files containing the captured network packets
     */
    private final Set<File> inputFiles;

    /**
     * The general format of the files containing the network traffic dumps (usually Pcap)
     */
    private InputFormat inputFormat;
    /**
     * The output stream to write to, this stream should be closed by calling
     * {@link at.jku.fim.rubanetra.output.OutputWriterStrategy#closeWriter()}
     */
    private OutputStream outputStream;
    /**
     * The general format of the produced output, defines the transform to apply before the dumping process
     */
    private OutputFormat outputFormat;
    /**
     * The strategy defining the final data representation and performs the actual writing task
     */
    private OutputWriterStrategy outputWriterStrategy;
    /**
     * The Pcap data input stream, which will parse the inputFiles and compile/apply the BPF filter
     */
    private PcapFileHandler<PcapActivityListener> pcapHandler;

    private String bpfFilter;
    private boolean bpfOptimize = false;
    private int bpfNetmask = 0;
    /**
     * Kraken library specific settings
     */
    private final KrakenTransportLayerMappingFactory transportLayerMappingFactory;

    private final Set<KrakenProtocolConfiguration> protocolSettings;
    /**
     * Drools rule engine specific settings
     */
    private String droolsBaseModelName;

    private String droolsSessionModelName;
    private boolean sortByFirstTimestamp;
    private RubanetraSystemConfigurationBuilder() {
        // keep private to prevent public instantiation
        this.protocolSettings = new HashSet<>();
        this.inputFiles = new LinkedHashSet<>();
        this.transportLayerMappingFactory = KrakenTransportLayerMappingFactory.createDefaultKrakenTransportLayerMappingFactory();
    }

    /**
     * @return a new instance of the {@link at.jku.fim.rubanetra.config.model.impl.RubanetraSystemConfigurationBuilder}
     */
    public static RubanetraSystemConfigurationBuilder create() {
        return new RubanetraSystemConfigurationBuilder();
    }

    /**
     * @return the currently configured Pcap handler or null
     */
    public PcapFileHandler<PcapActivityListener> getPcapHandler() {
        return pcapHandler;
    }

    /**
     * This should be the last method called, i.e. after all the settings have been defined. This method will finalize
     * the configuration, perform several validations and checks, create a {@link at.jku.fim.rubanetra.pcap.PcapFileHandler}
     * (if it was not defined already) and return an immutable implementation of the system configuration itself.
     * <p>
     * This method will throw exceptions if invalid settings are encountered.
     *
     * @return an immutable, finalized configuration instance, which may be passed to a {@link at.jku.fim.rubanetra.drools.DroolsKrakenProtocolHandler}
     * @see at.jku.fim.rubanetra.config.model.impl.RubanetraSystemConfigurationBuilder.RubanetraSystemConfigurationImpl
     * @see at.jku.fim.rubanetra.drools.DroolsKrakenProtocolHandler
     */
    public synchronized RubanetraSystemConfiguration build() {
        if (pcapHandler == null) {
            createPcapHandler();
        }
        return new RubanetraSystemConfigurationImpl(this);
    }

    /**
     * Creates and sets a {@link at.jku.fim.rubanetra.pcap.PcapHandler} using all already defined
     * input files, see {@link #addInputPath(String)}. Unless called before invoking {@link #build()}, this method
     * will be called during the final build automatically.
     *
     * @return a PcapHandler using already defined input files
     * @throws java.lang.IllegalStateException if this method was called more than once
     */
    public PcapHandler<PcapActivityListener> createPcapHandler() {
        if (this.pcapHandler != null) {
            throw new IllegalStateException("A PcapHandler has already been built.");
        }

        final File[] pcapFiles = inputFiles.toArray(new File[inputFiles.size()]);
        if (this.bpfFilter != null) {
            PcapFilter bpfFilter = new PcapFilter(this.bpfFilter, this.bpfOptimize, this.bpfNetmask);
            this.pcapHandler = new PcapFileHandler<>(bpfFilter, sortByFirstTimestamp, pcapFiles);
        } else {
            this.pcapHandler = new PcapFileHandler<>(sortByFirstTimestamp, pcapFiles);
        }

        return this.pcapHandler;
    }

    /**
     * Sets a Berkeley Packet Filter expression.
     * Some examples are available at http://www.tcpdump.org/manpages/pcap-filter.7.html
     * See also Steven McCanne and  Van Jacobson paper
     * 'The BSD packet filter: A New Architecture for User-level Packet Capture',
     * in USENIX Winter 1993 Conference Proceedings
     * <p>
     * This setting is optional.
     *
     * @param bpfFilter a String constituting a Berkeley Packet Filter which will be compiled by JNetPcap/libpcap
     * @return the builder instance
     */
    public RubanetraSystemConfigurationBuilder setBpfFilter(String bpfFilter) {
        if (this.pcapHandler != null) {
            throw new IllegalStateException("A PcapHandler has already been built.");
        }
        this.bpfFilter = bpfFilter;
        return this;
    }

    /**
     * Whether or not to try to optimize the given BPF expression, default: false
     * <p>
     * This setting is optional.
     *
     * @param bpfOptimize true, iff the compiler should try to optimize the expression
     * @return the builder instance
     */
    public RubanetraSystemConfigurationBuilder setBpfOptimize(boolean bpfOptimize) {
        if (this.pcapHandler != null) {
            throw new IllegalStateException("A PcapHandler has already been built.");
        }
        this.bpfOptimize = bpfOptimize;
        return this;
    }

    /**
     * Sets the BPF netmask, default: 0
     * <p>
     * This setting is optional.
     *
     * @param bpfNetmask the BPF netmask to use
     * @return the builder instance
     */
    public RubanetraSystemConfigurationBuilder setBpfNetmask(int bpfNetmask) {
        if (this.pcapHandler != null) {
            throw new IllegalStateException("A PcapHandler has already been built.");
        }
        this.bpfNetmask = bpfNetmask;
        return this;
    }

    /**
     * Every Drools knowledge base is identified by a name, this name must be specified only if the default
     * knowledge base as specified in the Drools configuration file ('kmodule.xml') should not be used.
     * See http://drools.jboss.org/documentation.
     * <p>
     * This setting is optional.
     *
     * @param droolsBaseModelName the Drools knowledge base to use for fact insertion and deduction
     * @return the builder instance
     */
    public RubanetraSystemConfigurationBuilder setDroolsBaseModelName(String droolsBaseModelName) {
        this.droolsBaseModelName = droolsBaseModelName;
        return this;
    }

    /**
     * Every drools knowledge base may define several {@link org.kie.api.runtime.KieSession}s,
     * this parameter must be specified if the default KieSession as specified in the Drools configuration
     * file ('kmodule.xml') should not be used. See http://drools.jboss.org/documentation.
     * <p>
     * This setting is optional.
     *
     * @param droolsSessionModelName the Drools session to use for fact insertion and deduction
     * @return the builder instance
     */
    public RubanetraSystemConfigurationBuilder setDroolsSessionModelName(String droolsSessionModelName) {
        this.droolsSessionModelName = droolsSessionModelName;
        return this;
    }

    /**
     * The path of the file containing the protocol configuration. The file must exist and be readable.
     * <p>
     * This setting is mandatory.
     *
     * @param protocolConfigurationPath the path of the file containing the protocol configuration
     * @return the builder instance
     */
    public RubanetraSystemConfigurationBuilder setProtocolConfigurationPath(String protocolConfigurationPath) {
        protocolConfigurationFile = canonicalizeFilePath(protocolConfigurationPath);
        checkFileReadable(protocolConfigurationFile);
        return this;
    }

    /**
     * The path of the directory containing the drools configuration directory. The directory must contain a 'META-INF'
     * directory and the file 'kmodule.xml', i.e. 'META-INF/kmodule.xml', as well as 'META-INF/pom.properties' because
     * Drools apparently expects the knowledge base to reside inside a JAR archive. The creation of a JAR archive for
     * every change to the rule base is not possible for this
     * systems use-case, due to the design goal that changes to the knowledge base are both
     * convenient and easy to perform from a users perspective. The alternative to this approach is to programmatically
     * define the Drools knowledge base which is currently not feasible.
     * <p>
     * ENHANCEMENT: This will probably change in the future,
     * as soon as a new way to define the knowledge base is found.
     * <p>
     * This setting is mandatory.
     *
     * @param droolsConfigurationString the path of the directory representing the Drools configuration structure
     * @return the builder instance
     */
    public RubanetraSystemConfigurationBuilder setDroolsConfigurationPath(String droolsConfigurationString) {
        this.droolsConfigurationFile = canonicalizeFilePath(droolsConfigurationString);
        checkFileReadable(this.droolsConfigurationFile);
        if (!droolsConfigurationFile.isDirectory()) {
            throw new IllegalArgumentException("The drools configuration file must represent a directory (containing ./META-INF/kmodule.xml).");
        }
        return this;
    }

    /**
     * Set the path of the file containing the input configuration. The file must exist and be readable.
     * <p>
     * This setting is mandatory.
     *
     * @param inputConfigurationString the path of the file containing the input configuration
     * @return the builder instance
     */
    public RubanetraSystemConfigurationBuilder setInputConfigurationPath(String inputConfigurationString) {
        this.inputConfigurationFile = canonicalizeFilePath(inputConfigurationString);
        checkFileReadable(this.inputConfigurationFile);
        return this;
    }

    /**
     * Set the path of the file containing the output configuration, this file must exist and be readable.
     * <p>
     * This setting is mandatory.
     *
     * @param outputConfigurationString the path of the file containing the output configuration
     * @return the builder instance
     */
    public RubanetraSystemConfigurationBuilder setOutputConfigurationPath(String outputConfigurationString) {
        outputConfigurationFile = canonicalizeFilePath(outputConfigurationString);
        checkFileReadable(outputConfigurationFile);
        return this;
    }

    /**
     * The main output stream for writing the derived and parsed information to. A {@link at.jku.fim.rubanetra.output.OutputWriterStrategy}
     * performs the actual writing process and transformation.
     * <p>
     * This setting is mandatory
     *
     * @param out the output stream to use, may not be null
     * @return the builder instance
     */
    public RubanetraSystemConfigurationBuilder setOutputStream(OutputStream out) {
        if (out == null) {
            throw new IllegalArgumentException("Invalid output stream");
        }
        this.outputStream = out;
        return this;
    }

    /**
     * A convenience method for {@link #setOutputStream(java.io.OutputStream)}, i.e. it takes a file path
     * and sets the output stream setting.
     * The specified file must be writable.
     * <p>
     * This setting is optional and replaces the obligation to call {@link #setOutputStream(java.io.OutputStream)}
     *
     * @param outputPath the path depicting a file to write to
     */
    public void setOutputFile(String outputPath) {
        File outputFile = canonicalizeFilePath(outputPath);

        if (outputFile.isDirectory()) {
            throw new IllegalArgumentException();
        }

        if(outputFile.exists()) {
            log.warn("Output file {} exists, it will be overwritten.", outputFile.toString());
        }

        try {
            setOutputStream(new FileOutputStream(outputFile));
        } catch (FileNotFoundException e) {
            throw new IllegalArgumentException("Invalid output file, is it writable?");
        }
    }

    /**
     * The {@link at.jku.fim.rubanetra.config.model.RubanetraSystemConfiguration.OutputFormat} to use
     * for transforming the written output before it is written.
     * <p>
     * This setting is mandatory.
     *
     * @param parsedOutputFormat the output format to use as serialization representation
     * @return the builder instance
     */
    public RubanetraSystemConfigurationBuilder setOutputFormat(String parsedOutputFormat) {
        if (parsedOutputFormat == null) {
            throw new IllegalArgumentException("Invalid output format.");
        } else {
            OutputFormat outputFormat = OutputFormat.valueOf(trimAndCapitalize(parsedOutputFormat));
            if (outputFormat == null) {
                throw new IllegalArgumentException(String.format("Unknown output format, %s)", parsedOutputFormat));
            } else {
                this.outputFormat = outputFormat;
            }
        }

        return this;
    }

    /**
     * Sets the strategy used to control the serialization process.
     * <p>
     * This setting is mandatory.
     *
     * @param outputWriter the strategy to use
     * @return the builder instance
     */
    public RubanetraSystemConfigurationBuilder setOutputWriterStrategy(OutputWriterStrategy outputWriter) {
        if (outputWriter == null) {
            throw new IllegalArgumentException("OutputWriter must not be null.");
        }

        this.outputWriterStrategy = outputWriter;
        return this;
    }

    private String trimAndCapitalize(String aString) {
        return aString.trim().toUpperCase();
    }

    /**
     * Adds a file-path of a readable file containing data adhering to the specified {@link at.jku.fim.rubanetra.config.model.RubanetraSystemConfiguration.InputFormat}
     * to the list of files that are to be parsed.
     * Please note that a input file handler {@link #createPcapHandler()} must not be created before adding all
     * required input files.
     * <p>
     * This setting is optional.
     *
     * @param inputPath a readable file containing data adhering to the configured input format (unchecked)
     * @return the builder instance
     */
    public RubanetraSystemConfigurationBuilder addInputPath(String inputPath) {
        if (this.pcapHandler != null) {
            throw new IllegalStateException("A PcapHandler has already been built.");
        }
        File inputFile = canonicalizeFilePath(inputPath);
        checkFileReadable(inputFile);
        if (inputFiles.contains(inputFile)) {
            throw new IllegalArgumentException(String.format("Duplicate input file: %s", inputFile));
        }
        this.inputFiles.add(inputFile);
        return this;
    }

    /**
     * Sets the flag that specifies whether or not the Pcap file handler should sort the input files according
     * to the chronological order of their first PCAP entry.
     * @param sortByFirstTimestamp true, iff the first timestamp of each file entry should be used for specifying
     *                             the order of the input files.
     * @return the builder instance
     */
    public RubanetraSystemConfigurationBuilder setSortByFirstTimestamp(boolean sortByFirstTimestamp) {
        if (this.pcapHandler != null) {
            throw new IllegalStateException("A PcapHandler has already been built.");
        }
        this.sortByFirstTimestamp = sortByFirstTimestamp;

        return this;
    }

    /**
     * Sets the format/type of the input files to parse. Only a single common type is allowed, i.e. format mixing
     * is currently not implemented.
     * <p>
     * This setting is mandatory.
     *
     * @param parsedInputFormat the {@link at.jku.fim.rubanetra.config.model.RubanetraSystemConfiguration.InputFormat}
     *                          of the input files to parse
     * @return the builder instance
     */
    public RubanetraSystemConfigurationBuilder setInputFormat(String parsedInputFormat) {
        if (parsedInputFormat == null) {
            throw new IllegalArgumentException("Invalid input format.");
        } else {
            InputFormat inputFormat = InputFormat.valueOf(trimAndCapitalize(parsedInputFormat));
            if (inputFormat == null) {
                throw new IllegalArgumentException(String.format("Unknown input format, %s)", parsedInputFormat));
            } else {
                this.inputFormat = inputFormat;
            }
        }

        return this;
    }

    private void checkFileReadable(File file) {
        if (!file.canRead()) {
            log.warn(String.format("The specified file (%s) is not readable.", file));
        }
    }

    /**
     * Adds a concrete protocol configuration to the system's configuration.
     * It is assumed that the object has finished all required protocol handler bindings and initialization steps.
     * <p>
     * This setting is optional.
     *
     * @param setting the protocol configuration to add as a system setting
     * @return the builder instance
     */
    public RubanetraSystemConfigurationBuilder addProtocolSetting(KrakenProtocolConfiguration setting) {
        this.protocolSettings.add(setting);
        return this;
    }

    /**
     * @return the currently configured protocol configuration file
     */
    public File getProtocolConfigurationFile() {
        return protocolConfigurationFile;
    }

    /**
     * @return the currently configured directory containing the Drools configuration structure
     */
    public File getDroolsConfigurationFile() {
        return droolsConfigurationFile;
    }

    /**
     * @return the currently configured input configuration file
     */
    public File getInputConfigurationFile() {
        return inputConfigurationFile;
    }

    /**
     * @return the currently configured output configuration file
     */
    public File getOutputConfigurationFile() {
        return outputConfigurationFile;
    }

    /**
     * @return the currently configured transport layer mapping factory
     */
    public KrakenTransportLayerMappingFactory getTransportLayerMappingFactory() {
        return transportLayerMappingFactory;
    }

    private File canonicalizeFilePath(String filePath) {
        if (filePath == null) {
            throw new IllegalArgumentException();
        }
        log.debug("Trying to locate and canonicalize file path {}", filePath);
        File canonicalFile;

        URL resourceUrl = ClassLoader.getSystemResource(filePath);
        try {
            if (resourceUrl == null) {
                log.debug("Could not locate {} in the classpath, interpreting as real path", filePath);
                canonicalFile = new File(filePath).getCanonicalFile();
            } else {
                log.debug("Located {} in the classpath", filePath);
                canonicalFile = new File(resourceUrl.toURI()).getCanonicalFile();
            }
        } catch (IOException | URISyntaxException e) {
            throw new IllegalArgumentException(String.format("The specified path (%s) cannot be canonicalized.", filePath));
        }

        return canonicalFile;
    }

    /**
     * This class represents the actual immutable system configuration
     */
    public Set<File> getInputFiles() {
        return inputFiles;
    }

    public InputFormat getInputFormat() {
        return inputFormat;
    }

    public OutputStream getOutputStream() {
        return outputStream;
    }

    public OutputFormat getOutputFormat() {
        return outputFormat;
    }

    public String getBpfFilter() {
        return bpfFilter;
    }

    public boolean isBpfOptimize() {
        return bpfOptimize;
    }

    public int getBpfNetmask() {
        return bpfNetmask;
    }

    public String getDroolsBaseModelName() {
        return droolsBaseModelName;
    }

    public String getDroolsSessionModelName() {
        return droolsSessionModelName;
    }

    public boolean isSortByFirstTimestamp() {
        return sortByFirstTimestamp;
    }

    private final class RubanetraSystemConfigurationImpl implements RubanetraSystemConfiguration {
        private final Set<KrakenProtocolConfiguration> protocolSettings;
        private final PcapFileHandler<?> pcapHandler;
        private final Set<File> inputFiles;
        private final InputFormat inputFormat;
        private final OutputFormat outputFormat;
        private final OutputStream outputStream;
        private final OutputWriterStrategy outputWriterStrategy;
        private final KieSession kieSession;

        /**
         * only the surrounding builder should create instances of this class
         */
        private RubanetraSystemConfigurationImpl(RubanetraSystemConfigurationBuilder builder) {
            this.protocolSettings = Collections.unmodifiableSet(builder.protocolSettings);
            this.inputFiles = Collections.unmodifiableSet(builder.inputFiles);
            this.pcapHandler = builder.pcapHandler;
            this.inputFormat = builder.inputFormat;
            this.outputStream = builder.outputStream;
            this.outputFormat = builder.outputFormat;
            this.outputWriterStrategy = builder.outputWriterStrategy != null ? builder.outputWriterStrategy : OutputStrategyFactory.createOutputWriterStrategy(outputFormat, outputStream);
            KieContainer kContainer = DroolsKnowledgeBaseFactory.createKieContainer(builder.droolsConfigurationFile);
            this.kieSession = DroolsKnowledgeBaseFactory.createKieSession(kContainer, builder.droolsBaseModelName, builder.droolsSessionModelName);
        }

        @Override
        public OutputStream getOutputStream() {
            return outputStream;
        }

        @Override
        public OutputFormat getOutputFormat() {
            return outputFormat;
        }

        @Override
        public Set<KrakenProtocolConfiguration> getProtocolSettings() {
            return Collections.unmodifiableSet(protocolSettings);
        }

        @Override
        public PcapFileHandler<?> getPcapHandler() {
            return pcapHandler;
        }

        @Override
        public KieSession getKieSession() {
            return kieSession;
        }

        @Override
        public Set<File> getInputFiles() {
            return inputFiles;
        }

        @Override
        public InputFormat getInputFormat() {
            return inputFormat;
        }

        @Override
        public OutputWriterStrategy getOutputWriterStrategy() {
            return outputWriterStrategy;
        }
    }

}
