/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.drools;


import org.kie.api.KieServices;
import org.kie.api.builder.KieFileSystem;
import org.kie.api.builder.KieModule;
import org.kie.api.builder.KieRepository;
import org.kie.api.builder.model.KieBaseModel;
import org.kie.api.builder.model.KieModuleModel;
import org.kie.api.builder.model.KieSessionModel;
import org.kie.api.builder.model.KieSessionModel.KieSessionType;
import org.kie.api.conf.EqualityBehaviorOption;
import org.kie.api.conf.EventProcessingOption;
import org.kie.api.io.Resource;
import org.kie.api.runtime.KieContainer;
import org.kie.api.runtime.KieSession;
import org.kie.api.runtime.conf.ClockTypeOption;

import java.io.File;

/**
 * This factory may be used to create instances of a {@link org.kie.api.runtime.KieSession} and/or
 * {@link org.kie.api.runtime.KieContainer}.
 * Several default values were defined to correspond to the default options used by this framework, e.g.
 * a {@link org.kie.api.builder.model.KieSessionModel.KieSessionType#STATEFUL} session, a pseudo {@link org.kie.api.runtime.conf.ClockTypeOption}
 * and the event stream options are currently used for this purpose.
 *
 * For extended documentation of the used classes and default values, please refer to the Drools documentation at
 * http://drools.jboss.org/documentation
 */
public class DroolsKnowledgeBaseFactory {
    /**
     * The default session is a stateful one
     */
    public static final KieSessionType DEFAULT_KNOWLEDGE_SESSION_TYPE = KieSessionType.STATEFUL;
    /**
     * The default type of the rule engines session clock is a pseudo clock, see
     * https://docs.jboss.org/drools/release/6.0.1.Final/drools-docs/html/DroolsComplexEventProcessingChapter.html#d0e10068
     */
    public static final ClockTypeOption DEFAULT_CLOCK_TYPE = ClockTypeOption.get("pseudo");
    /**
     * The default processing mode is stream mode, since reasoning over time based events should be enabled
     */
    public static final EventProcessingOption DEFAULT_EVENT_PROCESSING = EventProcessingOption.STREAM;
    public static final EqualityBehaviorOption DEFAULT_EQUALS_BEHAVIOUR = EqualityBehaviorOption.EQUALITY;

    /**
     * Creates a {@link org.kie.api.runtime.KieContainer} for all the knowledge bases found in the provided
     * directory. If the provided argument is null, a default classpath container will be returned.
     * <p>
     * ENHANCEMENT: This method expects to find a knowledge base descriptor file, 'kmodule.xml' in the provided
     * directory's sub-directory 'META-INF' among a maven project reference file ('pom.properties' and 'pom.xml', for
     * Drools {@link org.kie.api.builder.ReleaseId}).
     * This is inconvenient and a better way should exist to define it programmatically?
     *
     * @param knowledgeBaseDir the directory containing the the Drools knowledge base. When null is passed, the default
     *                         classpath container will be returned.
     * @return The container for all knowledge bases found in the provided directory.
     */
    public static KieContainer createKieContainer(File knowledgeBaseDir) {
        KieServices kServices = KieServices.Factory.get();
        if (knowledgeBaseDir == null) {
            return kServices.getKieClasspathContainer();
        }
        Resource fileSystemResource = kServices.getResources().newFileSystemResource(knowledgeBaseDir);
        KieRepository kieRepository = kServices.getRepository();
        KieModule kModule = kieRepository.addKieModule(fileSystemResource);

        return kServices.newKieContainer(kModule.getReleaseId());
    }


    /**
     * Creates a KieSession using the provided base and session name, as well as a number of arbitrary Drools rule files.
     * By default, this session will be stateful, use a pseudo clock and perform reasoning in stream mode and
     * use equality as object comparison mechanism.
     *
     * @param baseModelName    the name of the knowledge base
     * @param sessionModelName the name of the {@link org.kie.api.runtime.KieSession}
     * @param ruleFiles        a number of Drools rule files
     * @return a {@link org.kie.api.runtime.KieSession} backed by the provided rule files using default options
     */
    public static KieSession createKieSession(String baseModelName, String sessionModelName, File... ruleFiles) {
        if (baseModelName == null || sessionModelName == null) {
            throw new IllegalArgumentException();
        }
        KieServices kServices = KieServices.Factory.get();
        KieModuleModel kModuleModel = kServices.newKieModuleModel();
        KieBaseModel kBaseModel = kModuleModel.newKieBaseModel(baseModelName);
        kBaseModel.setDefault(true).setEqualsBehavior(DEFAULT_EQUALS_BEHAVIOUR)
                .setEventProcessingMode(DEFAULT_EVENT_PROCESSING);

        KieSessionModel kSession = kBaseModel.newKieSessionModel(sessionModelName);
        kSession.setDefault(true).setType(DEFAULT_KNOWLEDGE_SESSION_TYPE)
                .setClockType(DEFAULT_CLOCK_TYPE);

        KieFileSystem kFileSystem = kServices.newKieFileSystem();
        kFileSystem.writeKModuleXML(kModuleModel.toXML());
        for (File ruleFile : ruleFiles) {
            kFileSystem.write(kServices.getResources().newFileSystemResource(ruleFile));
        }

        kServices.newKieBuilder(kFileSystem).buildAll();
        KieContainer kContainer = kServices.newKieContainer(kServices.getRepository()
                .getDefaultReleaseId());

        return kContainer.newKieSession();
    }

    /**
     * Creates the default {@link org.kie.api.runtime.KieSession} without changing any default values for the classpath
     * container.
     *
     * @return a default {@link org.kie.api.runtime.KieSession} using the classpath container (no settings are applied)
     */
    public static KieSession createDefaultKieSession() {
        KieServices ks = KieServices.Factory.get();
        KieContainer kContainer = ks.getKieClasspathContainer();
        return kContainer.newKieSession();
    }

    /**
     * Creates a {@link org.kie.api.runtime.KieSession} for the classpath container without
     * changing any default options, using the knowledge base and session name exactly as they are provided.
     *
     * @param kBaseName    the knowledge base name
     * @param kSessionName the session name
     * @return a default kie session instance for the knowledge base
     */
    public static KieSession createKieSession(String kBaseName, String kSessionName) {
        KieServices ks = KieServices.Factory.get();
        KieContainer kContainer = ks.getKieClasspathContainer();
        return createKieSession(kContainer, kBaseName, kSessionName);
    }

    /**
     * Creates a new {@link org.kie.api.runtime.KieSession} without changing any default options, using the
     * provided container, knowledge base and session name.
     *
     * @param kContainer   the container to use
     * @param kBaseName    the knowledge base name
     * @param kSessionName the session name
     * @return a default kie session instance for the knowledge base in the provided container, or null if the container
     * was null
     */
    public static KieSession createKieSession(KieContainer kContainer, String kBaseName, String kSessionName) {
        if (kContainer == null) {
            return null;
        }
        KieSession kSession;
        if (kSessionName == null || kSessionName.isEmpty()) {
            kSession = kContainer.newKieSession();
        } else {
            kSession = kContainer.newKieSession(kSessionName);
        }

        return kSession;
    }


}
