/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.drools;


import at.jku.fim.rubanetra.config.model.KrakenApplicationProtocolId;
import at.jku.fim.rubanetra.config.model.KrakenProtocolConfiguration;
import at.jku.fim.rubanetra.config.model.ProtocolId;
import at.jku.fim.rubanetra.config.model.RubanetraSystemConfiguration;
import at.jku.fim.rubanetra.output.OutputWriterStrategy;
import at.jku.fim.rubanetra.pcap.PcapActivity;
import at.jku.fim.rubanetra.pcap.PcapActivityListener;
import at.jku.fim.rubanetra.protocol.KrakenBaseProtocol;
import at.jku.fim.rubanetra.protocol.activity.Activity;
import at.jku.fim.rubanetra.protocol.activity.DroolsBaseActivity;
import at.jku.fim.rubanetra.protocol.activity.ReplaceableActivity;
import org.drools.core.time.SessionPseudoClock;
import org.kie.api.event.rule.ObjectDeletedEvent;
import org.kie.api.event.rule.ObjectInsertedEvent;
import org.kie.api.event.rule.ObjectUpdatedEvent;
import org.kie.api.event.rule.RuleRuntimeEventListener;
import org.kie.api.runtime.KieSession;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.Closeable;
import java.io.IOException;
import java.time.Duration;
import java.time.Instant;
import java.util.*;
import java.util.concurrent.TimeUnit;

/**
 * This class may be considered the final link between the Kraken protocol parsing library, the Drools rule engine and
 * the {@link at.jku.fim.rubanetra.output.OutputWriterStrategy}.
 * It listens to all appropriate events creates by the protocol decoders and injects them into to Drools knowledge base
 * session ({@link org.kie.api.runtime.KieSession}).
 * Finally, it also reacts to Event removal events, which originate from the rule engine itself and are utilized
 * as indicator to write these events to the pre-defined output stream before they are discarded from memory.
 * <p>
 * It is vital to invoke the {@link #close()} method after usage of the rule engine, because otherwise the
 * remaining events and/or facts which are still in the active knowledge base will _not_ be written to the output stream
 * and this will lead to missing output.
 * <p>
 * This class will provide certain global objects that may be used from the Drools rule files, e.g.
 * a {@link org.slf4j.Logger} via "log" and the used {@link at.jku.fim.rubanetra.output.OutputWriterStrategy}
 * via "outputWriter".
 */
public class DroolsKrakenProtocolHandler implements PcapActivityListener, Closeable {
    private final Logger log = LoggerFactory.getLogger(getClass());
    private final RubanetraSystemConfiguration systemConfig;
    private final Map<ProtocolId, Set<DroolsProtocolPropertyChangeListener>> protocolListener = new HashMap<>();
    private final SessionPseudoClock droolsClock;
    private final OutputWriterStrategy outputWriterStrategy;

    private KieSession kieSession;
    private PcapActivity currentPcapEntry;
    private Instant previousTimestamp, currentTimestamp;
    private RuleRuntimeEventListener droolsRuntimeListener = new RuleRuntimeEventListener() {
        @Override
        public void objectInserted(ObjectInsertedEvent event) {
        }

        @Override
        public void objectUpdated(ObjectUpdatedEvent event) {
        }

        @Override
        public void objectDeleted(ObjectDeletedEvent event) {
                /* every deleted/retracted event/fact will not be used for reasoning any more,
                 * therefore those objects can be written to the output stream and finally discarded from memory */
            if (event.getOldObject() instanceof ReplaceableActivity) {
                ReplaceableActivity oldActivity = (ReplaceableActivity) event.getOldObject();

                if (!oldActivity.isReplaced() && !oldActivity.isExcludedFromOutput()) {
                    outputWriterStrategy.writeActivity(oldActivity);
                }
                logDeletedActivity(oldActivity);
            } else if (event.getOldObject() instanceof Activity) {
                Activity oldActivity = (Activity) event.getOldObject();
                if (!oldActivity.isExcludedFromOutput()) {
                    outputWriterStrategy.writeActivity(oldActivity);
                }
                logDeletedActivity(oldActivity);
            }
        }

        private void logDeletedActivity(Activity oldActivity) {
            if (log.isDebugEnabled()) {
                log.debug("Deleted activity: {} @ {} # {}", oldActivity.getActivityType(), currentTimestamp, oldActivity.getCompoundFrameNumbers());
            }
        }
    };

    /**
     * Instantiates a new handler for a pre-initialized RubanetraSystemConfiguration
     *
     * @param systemConfig the system configuration to use, may not be null and must be initialized
     */
    public DroolsKrakenProtocolHandler(RubanetraSystemConfiguration systemConfig) {
        if (systemConfig == null) {
            log.error("The system configuration must not constitute a null pointer");
            throw new IllegalArgumentException();
        }
        this.systemConfig = systemConfig;
        this.kieSession = systemConfig.getKieSession();
        if (kieSession == null) {
            log.error("The system configuration must contain a valid Kie-Session object, instead a null pointer was passed");
            throw new IllegalArgumentException();
        }
        this.currentPcapEntry = null;
        this.droolsClock = kieSession.getSessionClock();

        outputWriterStrategy = systemConfig.getOutputWriterStrategy();
        if (outputWriterStrategy == null) {
            log.error("The system configuration must contain a valid output writer, instead a null pointer was passed");
            throw new IllegalArgumentException();
        }

        initialize();
    }

    private void initialize() {
        // listen to all appropriate events, however, try to prevent duplicates
        for (KrakenProtocolConfiguration setting : systemConfig.getProtocolSettings()) {
            Map<ProtocolId, KrakenBaseProtocol> boundProtocols = setting.getBoundProtocols();
            for (ProtocolId id : boundProtocols.keySet()) {
                if (!protocolListener.containsKey(id)) {
                    // new unique protocol discovered, therefore add a new listener
                    KrakenBaseProtocol krakenBaseProtocol = boundProtocols.get(id);
                    DroolsProtocolPropertyChangeListener changeListener = new DroolsProtocolPropertyChangeListener(krakenBaseProtocol);
                    krakenBaseProtocol.addPropertyChangeListener(changeListener);
                    protocolListener.put(id, new HashSet<>(Arrays.asList(changeListener)));
                } else if (id instanceof KrakenApplicationProtocolId) {
                    // we want to receive all activities from the application layer, even for redundant application protocols,
                    // e.g. consider UDP->DNS and TCP->DNS.
                    // ENHANCEMENT:
                    // However, this will cause certain duplicate activities iff the application layer handler is bound
                    // to another handler above the application layer (this is currently not the case).
                    // As soon as this constraint is violated, an alternative mechanism has to be implemented.
                    KrakenBaseProtocol krakenBaseProtocol = boundProtocols.get(id);
                    DroolsProtocolPropertyChangeListener changeListener = new DroolsProtocolPropertyChangeListener(krakenBaseProtocol);
                    krakenBaseProtocol.addPropertyChangeListener(changeListener);
                    protocolListener.get(id).add(changeListener);
                }
            }
        }

        /**
         * These variables are accessible from within the rule files
         */
        kieSession.setGlobal("outputWriter", outputWriterStrategy);
        kieSession.setGlobal("log", LoggerFactory.getLogger(DroolsBaseActivity.class));

        kieSession.addEventListener(droolsRuntimeListener);
    }

    private void updateSessionClock() {
        if (previousTimestamp == null) {
            droolsClock.advanceTime(currentTimestamp.getEpochSecond(), TimeUnit.SECONDS);
            droolsClock.advanceTime(currentTimestamp.getNano(), TimeUnit.NANOSECONDS);
            log.debug("Setting initial Drools Pseudo clock time to {}", currentTimestamp);
            log.debug("The current Drools clock time is {}, frame # {}", Instant.ofEpochMilli(droolsClock.getCurrentTime()),
                    currentPcapEntry.getFrameNumber());
        } else if (previousTimestamp.isBefore(currentTimestamp)) {
            Duration duration = Duration.between(previousTimestamp, currentTimestamp);
            log.debug("Advancing Drools Pseudo clock by {}", duration);
            droolsClock.advanceTime(duration.getSeconds(), TimeUnit.SECONDS);
            droolsClock.advanceTime(duration.getNano(), TimeUnit.NANOSECONDS);
            log.debug("The current Drools clock time is {}, reference {}, frame # {}",
                    Instant.ofEpochMilli(droolsClock.getCurrentTime()), currentTimestamp, currentPcapEntry.getFrameNumber());
        }
    }

    @Override
    public void processPcapActivity(PcapActivity pcapActivity) {
        if (pcapActivity == null) {
            log.debug("An invalid pcap activity was passed after {}", currentTimestamp);
            return;
        }
        this.currentPcapEntry = pcapActivity;
        this.previousTimestamp = currentTimestamp;
        this.currentTimestamp = pcapActivity.getPcapTimestamp();
        updateSessionClock();
        kieSession.fireAllRules();
    }

    /**
     * Closes the drools session and deletes, i.e. retracts, all remaining facts and releases any system resources associated
     * with it. If the session is already closed then invoking this method has no effect.
     * This method will clode the used {@link at.jku.fim.rubanetra.output.OutputWriterStrategy} and
     * the used {@link org.kie.api.runtime.KieSession}.
     *
     * @throws java.io.IOException if an I/O error occurs
     */
    @Override
    public void close() throws IOException {
        if (kieSession != null) {
            kieSession.getFactHandles().forEach(kieSession::delete);
            kieSession.dispose();
            systemConfig.getOutputWriterStrategy().closeWriter();
            kieSession = null;
        }
    }

    /**
     * This class listens to all events that carry an {@link at.jku.fim.rubanetra.protocol.activity.Activity}
     * and injects them into the knowledge session
     */
    private class DroolsProtocolPropertyChangeListener implements PropertyChangeListener {
        private final KrakenBaseProtocol source;

        private DroolsProtocolPropertyChangeListener(KrakenBaseProtocol source) {
            this.source = source;
        }

        /**
         * This method gets called when a bound property is changed.
         *
         * @param evt A PropertyChangeEvent object describing the event source
         *            and the property that has changed.
         */
        @Override
        public void propertyChange(PropertyChangeEvent evt) {
            if (evt.getNewValue() instanceof Activity) {
                kieSession.insert(evt.getNewValue());
            } else {
                log.debug("Passed event object is not an Activity: {}", evt.getNewValue());
            }
        }
    }

}
