/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.output.impl;


import at.jku.fim.rubanetra.output.OutputWriterStrategy;
import at.jku.fim.rubanetra.output.mixin.HttpMixIn;
import at.jku.fim.rubanetra.output.mixin.PcapMixIn;
import at.jku.fim.rubanetra.protocol.activity.Activity;
import com.fasterxml.jackson.core.JsonEncoding;
import com.fasterxml.jackson.core.JsonFactory;
import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SerializationFeature;
import org.jnetpcap.packet.PcapPacket;
import org.jnetpcap.protocol.tcpip.Http;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.OutputStream;

/**
 * This class implements a Json specific transformation (using the Jackson library) of the derived Event/Fact/Activity-data.
 * By default, it tries to pretty-print the resulting JSON-output, flushes the output stream after each written
 * activity and does not fail on empty activities, nor will it close the used output-stream.
 * Because certain classes do not provide appropriate inclusion or exclusion annotations, certain MixIn classes
 * will be used as substitute.
 * <p>
 * Additionally the implementation version, vendor and title is provided for external parsers.
 *
 * @see at.jku.fim.rubanetra.output.mixin.HttpMixIn
 * @see at.jku.fim.rubanetra.output.mixin.PcapMixIn
 */
public class JsonOutputWriterStrategy implements OutputWriterStrategy {
    private final Logger log = LoggerFactory.getLogger(getClass());
    private final OutputStream outputStream;
    private final ObjectMapper mapper;

    private JsonGenerator jsonGenerator;

    /**
     * Creates a new instance of this strategy, all passed activities will be written to the provided output-stream.
     *
     * @param outputStream the output-stream to write the received activity-data in JSON-format to
     */
    public JsonOutputWriterStrategy(OutputStream outputStream) {
        if (outputStream == null) {
            log.error("Cannot write to passed outputstream - null pointer");
            throw new NullPointerException();
        }
        this.outputStream = outputStream;

        mapper = new ObjectMapper();
        // to enable standard indentation ("pretty-printing"):
        mapper.enable(SerializationFeature.INDENT_OUTPUT);
        mapper.enable(SerializationFeature.FLUSH_AFTER_WRITE_VALUE);
        // to allow serialization of "empty" POJOs (no properties to serialize)
        // (without this setting, an exception is thrown in those cases)
        mapper.disable(SerializationFeature.FAIL_ON_EMPTY_BEANS);
        // to write java.util.Date, Calendar as number (timestamp):
        mapper.disable(SerializationFeature.WRITE_DATES_AS_TIMESTAMPS);
        mapper.disable(SerializationFeature.CLOSE_CLOSEABLE);

        mapper.addMixInAnnotations(Http.class, HttpMixIn.class);
        mapper.addMixInAnnotations(PcapPacket.class, PcapMixIn.class);

        initialize();
    }

    private void initialize() {
        JsonFactory jsonFactory = mapper.getFactory();
        Package classPackage = getClass().getPackage();
        try {
            jsonGenerator = jsonFactory.createGenerator(outputStream, JsonEncoding.UTF8);
            jsonGenerator.useDefaultPrettyPrinter();
            jsonGenerator.disable(JsonGenerator.Feature.AUTO_CLOSE_TARGET);
            jsonGenerator.enable(JsonGenerator.Feature.AUTO_CLOSE_JSON_CONTENT);
            jsonGenerator.writeStartObject();

            jsonGenerator.writeStringField("implementationTitle", classPackage.getImplementationTitle());
            jsonGenerator.writeStringField("implementationVersion", classPackage.getImplementationVersion());
            jsonGenerator.writeStringField("implementationVendor", classPackage.getImplementationVendor());

            jsonGenerator.writeArrayFieldStart("activities");
        } catch (IOException e) {
            log.error("Unable to create Json generator required for the selected output strategy.", e);
            throw new IllegalStateException();
        }
    }

    @Override
    public void writeActivity(Activity activity) {
        if (activity == null) {
            log.warn("Skipping activity - null pointer");
            return;
        }
        try {
            mapper.writeValue(jsonGenerator, activity);
//            mapper.writeValue(outputStream,activity);
//            System.out.println(mapper.writeValueAsString(activity));
        } catch (IOException e) {
            log.warn("Unable to serialize activity {} to the outputstream, skipping.", activity);
            log.debug("Exception:", e);
        }
    }


    @Override
    public void closeWriter() {
        if (jsonGenerator != null) {
            try {
                jsonGenerator.writeEndArray();
                jsonGenerator.writeEndObject();
                jsonGenerator.flush();
                jsonGenerator.close();
                jsonGenerator = null;
            } catch (IOException e) {
                log.warn("Error occurred while closing the writer", e);
            }
        }
    }
}
