/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.output.impl;

import at.jku.fim.rubanetra.output.OutputWriterStrategy;
import at.jku.fim.rubanetra.output.mixin.HttpMixIn;
import at.jku.fim.rubanetra.output.mixin.PcapMixIn;
import at.jku.fim.rubanetra.protocol.activity.Activity;
import com.fasterxml.jackson.core.JsonEncoding;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SerializationFeature;
import com.fasterxml.jackson.dataformat.xml.XmlFactory;
import com.fasterxml.jackson.dataformat.xml.XmlMapper;
import com.fasterxml.jackson.dataformat.xml.ser.ToXmlGenerator;
import com.fasterxml.jackson.dataformat.xml.util.DefaultXmlPrettyPrinter;
import org.jnetpcap.packet.PcapPacket;
import org.jnetpcap.protocol.tcpip.Http;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.xml.namespace.QName;
import javax.xml.stream.XMLStreamException;
import java.io.IOException;
import java.io.OutputStream;

/**
 * Created by stefan on 9/30/14.
 */
public class XmlOutputWriterStrategy implements OutputWriterStrategy {
    public static final QName QNAME_ACTIVITY_ARRAY = new QName("activityArray");
    public static final QName QNAME_ACTIVITY = new QName("activity");
    private final Logger log = LoggerFactory.getLogger(getClass());
    private final XmlMapper mapper;
    private ToXmlGenerator xmlGenerator;

    public XmlOutputWriterStrategy(OutputStream outputStream) {
        if (outputStream == null) {
            log.error("Cannot write to passed outputstream - null pointer");
            throw new NullPointerException();
        }

        this.mapper = new XmlMapper();
        assignDefaultFeatures(mapper);

        final XmlFactory xmlFactory = this.mapper.getFactory();

        Package classPackage = getClass().getPackage();
        try {
            xmlGenerator = xmlFactory.createGenerator(outputStream, JsonEncoding.UTF8);
            DefaultXmlPrettyPrinter pp = new DefaultXmlPrettyPrinter();
            xmlGenerator.setPrettyPrinter(pp);
            xmlGenerator.enable(ToXmlGenerator.Feature.WRITE_XML_1_1);
            QName rootQName = new QName("root");
            xmlGenerator.initGenerator();
            xmlGenerator.setNextName(rootQName);
            xmlGenerator.writeStartObject();
            final String unknown = "Unknown";
            xmlGenerator.writeStringField("implementationTitle", classPackage.getImplementationTitle() == null ?
                    unknown : classPackage.getImplementationTitle());

            xmlGenerator.writeStringField("implementationVersion", classPackage.getImplementationVersion() == null ?
                    unknown : classPackage.getImplementationVersion());
            xmlGenerator.writeStringField("implementationVendor", classPackage.getImplementationVendor() == null ?
                    unknown : classPackage.getImplementationVendor());

            xmlGenerator.startWrappedValue(QNAME_ACTIVITY_ARRAY, QNAME_ACTIVITY);
        } catch (IOException e) {
            e.printStackTrace();
        }

    }

    public static void assignDefaultFeatures(ObjectMapper mapper) {
        // to enable standard indentation ("pretty-printing"):
        mapper.enable(SerializationFeature.INDENT_OUTPUT);
        mapper.enable(SerializationFeature.FLUSH_AFTER_WRITE_VALUE);
        // to allow serialization of "empty" POJOs (no properties to serialize)
        // (without this setting, an exception is thrown in those cases)
        mapper.disable(SerializationFeature.FAIL_ON_EMPTY_BEANS);
        // to write java.util.Date, Calendar as number (timestamp):
        mapper.disable(SerializationFeature.WRITE_DATES_AS_TIMESTAMPS);
        mapper.disable(SerializationFeature.CLOSE_CLOSEABLE);

        mapper.addMixInAnnotations(Http.class, HttpMixIn.class);
        mapper.addMixInAnnotations(PcapPacket.class, PcapMixIn.class);
    }

    /**
     * Write an {@link at.jku.fim.rubanetra.protocol.activity.Activity} to the provided output-stream
     * using the implementation-specific transformation of the content.
     *
     * @param activity the activity that will be transformed according to the strategy in use and written to the
     *                 provided output-stream
     */
    @Override
    public void writeActivity(Activity activity) {
        if (activity == null) {
            log.warn("Skipping activity - null pointer");
            return;
        }

        try {
            mapper.writeValue(xmlGenerator, activity);
        } catch (IOException e) {
            log.warn("Unable to serialize activity {} to the outputstream, skipping.", activity.getCompoundFrameNumbers());
            log.debug("Exception: {}", e);
        }
    }

    /**
     * Closes the writer but does not release or close the underlying output stream.
     * If the writer is already closed then invoking this
     * method has no effect.
     */
    @Override
    public void closeWriter() throws IOException {
        if (xmlGenerator != null) {
            xmlGenerator.flush();
//            xmlGenerator.writeEndArray();
            try {
                xmlGenerator.getStaxWriter().writeEndDocument();
            } catch (XMLStreamException e) {
                e.printStackTrace();
            }
//            xmlGenerator.close();
            xmlGenerator = null;
        }
    }
}
