/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.pcap;


import at.jku.fim.rubanetra.protocol.activity.AbstractReplaceableActivity;
import org.jnetpcap.packet.PcapPacket;
import org.kie.api.definition.type.Role;
import org.kie.api.definition.type.Role.Type;

import java.time.Duration;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;

/**
 * A PcapActivity may be interpreted as a Network Layer 1 activity type, because all common PCAP-Files contain
 * individual Pcap-entries which are wrapped by this class in order to provide PCAP-specific metadata per entry.
 * This class currently supports the JNetPcap library only (which in turn relies on the native libpcap library),
 * however, if the limited Kraken-Pcap-Parser has to be used, consider converting the Kraken-specific Pcap packet
 * to a {@link org.jnetpcap.packet.JMemoryPacket} in a manner similar to the following (at your own risk):
 * <p>
 * {@code byte[] arr = new byte[getPcapPacket().getPacketData().readableBytes()];}
 * <p>
 * {@code new ChainBuffer(getPcapPacket().getPacketData()).gets(arr);}
 * <p>
 * {@code JMemoryPacket jPacket = new JMemoryPacket(arr);}
 * <p>
 * {@code jPacket.scan(Ethernet.ID);}
 * <p>
 * {@code Ethernet header = jPacket.getHeader(new Ethernet());}
 * <p>
 * If a JNetPcap specific {@link org.jnetpcap.packet.PcapPacket} must be passed to a Kraken parser, the class
 * {@link at.jku.fim.rubanetra.pcap.PcapPacketToKrakenPcapPacketAdapter} may be useful.
 *
 * @see at.jku.fim.rubanetra.pcap.PcapPacketToKrakenPcapPacketAdapter
 */
@Role(Type.EVENT)
public class PcapActivity extends AbstractReplaceableActivity {
    private static final long serialVersionUID = 1884831815993828991L;
    private final long frameNumber;
    private final PcapPacket pcapPacket;
    private final Instant pcapTimestamp;

    /**
     * Create a new PcapActivity using the provided frame number (as seen by the
     * {@link at.jku.fim.rubanetra.pcap.PcapFileHandler}) and the JNetPcap specific PCAP packet.
     *
     * @param frameNumber the frame number of the packet, reconstructed by the {@link at.jku.fim.rubanetra.pcap.PcapFileHandler}
     * @param pcapPacket  the JNetPcap specific PCAP packet representation
     */
    public PcapActivity(long frameNumber, PcapPacket pcapPacket) {
        super(pcapPacket);
        this.frameNumber = frameNumber;
        this.pcapPacket = pcapPacket;

        // the JNetPcap timestamp will be interpreted as a Java 8 specific Instant in time
        this.pcapTimestamp = Instant.ofEpochSecond(pcapPacket.getCaptureHeader().hdr_sec()).
                plus(Duration.of(pcapPacket.getCaptureHeader().hdr_usec(), ChronoUnit.MICROS));

        setStartInstant(pcapTimestamp);
        setEndInstant(pcapTimestamp);
    }

    /**
     * @return The JNetPcap specific representation of this packet capture activity
     */
    public PcapPacket getPcapPacket() {
        return this.pcapPacket;
    }

    /**
     * @return the frame number of this packet capture activity as reconstructed by the used
     * {@link at.jku.fim.rubanetra.pcap.PcapHandler}
     */
    public long getFrameNumber() {
        return this.frameNumber;
    }

    /**
     * @return The Java 8 specific Instant in time, which was built from the JNetPcap specific timestamp information.
     * @see org.jnetpcap.packet.PcapPacket#getCaptureHeader()
     */
    public Instant getPcapTimestamp() {
        return pcapTimestamp;
    }

    @Override
    public Collection<Long> getCompoundFrameNumbers() {
        return Collections.unmodifiableCollection(Arrays.asList(getFrameNumber()));
    }

    @Override
    public String toString() {
        return "PcapActivity{" +
                "frameNumber=" + frameNumber +
                ", pcapPacket=" + pcapPacket +
                ", pcapTimestamp=" + pcapTimestamp +
                '}';
    }
}
