/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.pcap;


import org.jnetpcap.Pcap;
import org.jnetpcap.PcapBpfProgram;
import org.jnetpcap.packet.JPacket;
import org.jnetpcap.packet.JPacketHandler;
import org.jnetpcap.packet.PcapPacket;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;
import java.util.*;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.stream.Stream;

/**
 * Implements a pcap handler for parsing one or multiple files adhering to a supported PCAP-definition (JNetPcap's
 * native Pcap library, libpcap, will be used).
 * This class is subject to future changes, currently it employs a custom way to support multiple files, i.e.:
 * <ul>
 * <li>if multiple files are passed and the sortByPcapTimestamp flag is set to false, all files are treated in
 * the order they have been passed</li>
 * <li>if multiple files are passed and the sortByPcapTimestamp flag is set to true, all files will be opened and
 * the first available pcap entry will be parsed. All timestamps of the timestamps will be compared against each
 * other and a chronological file-order will be established. This mechanism will potentially fail if
 * the files contain overlapping time intervals. A possibly suitable future implementation for this use case
 * could employ a chronologically sorted buffer to alleviate the effects of this issue.</li>
 * <li>if a single file is passed, the sortByPcapTimestamp flag will be ignored</li>
 * </ul>
 * <p>
 * Multiple {@link at.jku.fim.rubanetra.pcap.PcapActivityListener} may be registered with an instance of
 * this class, note however that the events will be dispatched sequentially in the listener registration order.
 * The only exception to this rule is a listener object that is passed directly as parameter to the
 * {@link #readNextPcapEntry(PcapActivityListener)} as this object will be the first one to receive a callback with
 * the appropriate {@link at.jku.fim.rubanetra.pcap.PcapActivity} event.
 *
 * @param <T> A custom implementation of a {@link at.jku.fim.rubanetra.pcap.PcapActivityListener} that
 *            will be called back before all other registered listeners as long as it is passed as argument to the
 *            {@link #readNextPcapEntry(PcapActivityListener)} method.
 */
public class PcapFileHandler<T extends PcapActivityListener> implements PcapHandler<T> {
    private final Logger log = LoggerFactory.getLogger(getClass());
    private final StringBuilder errorBuilder = new StringBuilder();

    private final Queue<File> pcapInputFileQueue;
    private final int numberOfInputFiles;
    private final boolean sortByPcapTimestamp;
    private final List<PcapActivityListener> fileActivityListeners;
    private final JPacketHandler<PcapActivityListener> pcapPacketHandler = this::processNextPacket;
    private final PcapFilter pcapFilter;

    private long currentFrameNumber;
    private PcapActivity currentPcapActivity;
    private Pcap currentPcapHandle;

    /**
     * Instantiates a new pcap file handler using the passed Berkeley Packet Filter pcapFilter and a number of
     * input files. These files will be sorted according to the first encountered timestamp in the pcap entries of each
     * file iff sortByPcapTimestamp is passed as true.
     *
     * @param pcapFilter          the Berkeley Packet Filter to apply to all specified pcapFiles
     * @param sortByPcapTimestamp whether or not to sort the pcapFiles chronologically according to the first
     *                            encountered timestamp of the individual pcap entries
     * @param pcapFiles           the files to parse (must adhere to a PCAP specification supported by JNetPcap/libpcap)
     * @throws java.lang.IllegalArgumentException if no input files are passed
     */
    public PcapFileHandler(PcapFilter pcapFilter, boolean sortByPcapTimestamp, File... pcapFiles) {
        if (pcapFiles == null || pcapFiles.length <= 0) {
            throw new IllegalArgumentException("No input files have been passed to the Pcap File Handler.");
        }
        this.pcapFilter = pcapFilter;
        this.pcapInputFileQueue = new ConcurrentLinkedQueue<>();
        this.sortByPcapTimestamp = sortByPcapTimestamp;
        if (sortByPcapTimestamp && pcapFiles.length > 1) {
            sortInputFilesByPcapTimestamp(pcapFiles);
        } else {
            Collections.addAll(this.pcapInputFileQueue, pcapFiles);
        }
        this.numberOfInputFiles = pcapFiles.length;
        this.fileActivityListeners = new LinkedList<>();
        this.currentFrameNumber = 1;
    }

    /**
     * Instantiates a new pcap file handler using no Berkeley Packet Filter.
     *
     * @param sortByPcapTimestamp whether or not to sort the pcapFiles chronologically according to the first
     *                            encountered timestamp of the individual pcap entries
     * @param pcapFiles           the files to parse (must adhere to a PCAP specification supported by JNetPcap/libpcap)
     * @throws java.lang.IllegalArgumentException if no input files are passed
     * @see #PcapFileHandler(PcapFilter, boolean, java.io.File...)
     */
    public PcapFileHandler(boolean sortByPcapTimestamp, File... pcapFiles) {
        this(null, sortByPcapTimestamp, pcapFiles);
    }

    /**
     * Instantiates a new pcap file handler using no Berkeley Packet Filter and does not sort the input files.
     *
     * @param pcapFiles the files to parse (must adhere to a PCAP specification supported by JNetPcap/libpcap)
     * @throws java.lang.IllegalArgumentException if no input files are passed
     * @see #PcapFileHandler(PcapFilter, boolean, java.io.File...)
     */
    public PcapFileHandler(File... pcapFiles) {
        this(null, false, pcapFiles);
    }

    /**
     * Tries to sort the input files using one {@link at.jku.fim.rubanetra.pcap.PcapFileHandler}
     * per file. This is quite inefficient, therefore:
     * ENHANCEMENT: implement in place sort and/or handle overlapping time intervals
     *
     * @param pcapFiles the input files to sort
     */
    private void sortInputFilesByPcapTimestamp(File[] pcapFiles) {
        HashMap<PcapActivity, File> firstPcapActivities = new HashMap<>();
        for (File f : pcapFiles) {
            PcapFileHandler fileHandler = new PcapFileHandler(pcapFilter, false, f);
            fileHandler.readNextPcapEntry(pcapActivity -> firstPcapActivities.put(pcapActivity, f));
            fileHandler.close();
        }
        Stream<PcapActivity> sortedPcapActivities = firstPcapActivities.keySet().stream()
                .sorted((o1, o2) -> o1.getPcapTimestamp().compareTo(o2.getPcapTimestamp()));
        sortedPcapActivities.forEachOrdered(pcapActivity -> this.pcapInputFileQueue.add(firstPcapActivities.get(pcapActivity)));
    }

    @Override
    public boolean readNextPcapEntry(PcapActivityListener processor) {
        if (this.currentPcapHandle == null && !switchToNextReadablePcapFile()) {
            return false;
        }
        int returnCode = this.currentPcapHandle.dispatch(1, pcapPacketHandler, processor);
        return returnCode > 0 || switchToNextReadablePcapFile() && readNextPcapEntry(processor);

    }

    private void processNextPacket(JPacket packet, PcapActivityListener user) {
        // packet.scan(JRegistry.mapDLTToId(currentPcapHandle.datalink()));
        PcapPacket deepCopy = new PcapPacket(packet);
        this.currentPcapActivity = new PcapActivity(this.currentFrameNumber++, deepCopy);

        for (PcapActivityListener l : fileActivityListeners) {
            l.processPcapActivity(currentPcapActivity);
        }
        if (user != null)
            user.processPcapActivity(currentPcapActivity);
    }

    /**
     * This method opens, if possible, the next available input file and applies the Pcap filter and closes
     * the previous file handle. Because no checks have been performed until now (whether the next file truly is a
     * valid Pcap file) it tries to skip over invalid files until another valid file has been found.
     * An alternative could constitute a generic bail out.
     *
     * @return true, iff the switch to another readable PCAP-file was performed successfully
     */
    private boolean switchToNextReadablePcapFile() {
        boolean switchSuccessful = false;
        while (!switchSuccessful && !this.pcapInputFileQueue.isEmpty()) {
            if (this.currentPcapHandle != null) {
                this.currentPcapHandle.close();
                this.currentPcapHandle = null;
            }

            File currentPcapFile = this.pcapInputFileQueue.remove();
            try {
                this.currentPcapHandle = Pcap.openOffline(currentPcapFile.getCanonicalPath(),
                        errorBuilder);
                applyFilter();
                switchSuccessful = this.currentPcapHandle != null;
            } catch (IOException e) {
                log.warn(String.format("Unable to read file %s, trying next file..., reason: %s%n",
                        currentPcapFile, errorBuilder.toString()));
                errorBuilder.setLength(0);
                switchSuccessful = false;
            }
        }

        return switchSuccessful;
    }

    private void applyFilter() {
        if (this.pcapFilter != null && this.currentPcapHandle != null) {
            PcapBpfProgram bpfProgram = new PcapBpfProgram();
            if (Pcap.OK != this.currentPcapHandle.compile(bpfProgram,
                    pcapFilter.getBpfExpression(), pcapFilter.isOptimize() ? 1 : 0,
                    pcapFilter.getNetmask())) {
                log.warn(this.currentPcapHandle.getErr());
                return;
            }
            if (Pcap.OK != this.currentPcapHandle.setFilter(bpfProgram)) {
                log.warn(this.currentPcapHandle.getErr());
            }
        }
    }

    @Override
    public void close() {
        if (this.currentPcapHandle != null) {
            this.currentPcapHandle.close();
            this.currentPcapHandle = null;
        }
    }

    @Override
    public PcapActivity getCurrentPcapActivity() {
        return currentPcapActivity;
    }

    @Override
    public void addPcapProcessor(T processor) {
        this.fileActivityListeners.add(processor);
    }

    @Override
    public void removePcapProcessor(T processor) {
        this.fileActivityListeners.remove(processor);
    }

    public int getNumberOfInputFiles() {
        return numberOfInputFiles;
    }

    public boolean isSortByPcapTimestamp() {
        return sortByPcapTimestamp;
    }
}
