/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.pcap;


import java.io.Closeable;
import java.io.IOException;

public interface PcapHandler<T extends PcapActivityListener> extends Closeable {
    /**
     * Read the next Pcap entry from the respective input stream.
     * Implementations of this method must ensure that the passed processor parameter is called after all registered
     * PcapProcessors had a chance to handle the Pcap-entry, i.e. after the decoding chain has been traversed.
     * This constraint ensures that the users pcap processor may access subsequently decoded data.
     * The methods signature is still subject to change, i.e. it is planned that the boolean return value will
     * be changed to an integer value indicating a kind of error code.
     *
     * @param processor A Pcap-Entry processor capable of handling the read PcapActivity.
     * @return true, iff the entry has been read and parsed successfully, false otherwise (e.g. end of stream).
     */
    public boolean readNextPcapEntry(T processor);

    /**
     * A convenience method that provides a reference to the currently processed PcapActivity, i.e. it returns the
     * latest successfully decoded Pcap entry.
     *
     * @return the latest successfully decoded PcapActivity
     */
    public PcapActivity getCurrentPcapActivity();

    /**
     * Adds a custom listener or processor to the set of already registered listeners.
     * This processor will receive a notification of all decoded Pcap entries after the previously registered
     * processors have been notified.
     *
     * @param processor a custom listener to decoded pcap entries and related events
     */
    public void addPcapProcessor(T processor);

    /**
     * Removes a custom listener or processor from the set of already registered listeners.
     * This processor will no longer receive a notification of decoded Pcap entries.
     *
     * @param processor a registered listener
     */
    public void removePcapProcessor(T processor);

    /**
     * Closes the underlying input streams (if not closed already) and releases any system resources associated
     * with it. If the streams were already closed then invoking this
     * method has no effect.
     *
     * @throws IOException if an I/O error occurs
     */
    @Override
    public void close() throws IOException;
}
