/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.pcap;


import org.jnetpcap.PcapHeader;
import org.krakenapps.pcap.packet.PacketHeader;
import org.krakenapps.pcap.packet.PcapPacket;
import org.krakenapps.pcap.util.Buffer;
import org.krakenapps.pcap.util.ChainBuffer;

/**
 * This class is an adapter for a given {@link org.jnetpcap.packet.PcapPacket}. It may be used as a drop-in replacement
 * for a Kraken-specific {@link org.krakenapps.pcap.packet.PcapPacket}, i.e. instances of this class may be passed
 * directly to a Kraken decoder pipeline.
 * This class is still considered experimental, due to potentially dangerous long->int conversions and
 * different naming conventions among the utilized libraries.
 */
public class PcapPacketToKrakenPcapPacketAdapter extends PcapPacket {
    private final org.jnetpcap.packet.PcapPacket pcapPacket;

    /**
     * Takes a JNetPcap specific PcapPacket and creates a Kraken header and payload-buffer.
     *
     * @param pcapPacket the JNetPcap packet to adapt into a Kraken packet
     */
    public PcapPacketToKrakenPcapPacketAdapter(org.jnetpcap.packet.PcapPacket pcapPacket) {
        super(createKrakenHeader(pcapPacket), createKrakenPayloadBuffer(pcapPacket));
        this.pcapPacket = pcapPacket;
    }

    /**
     * @return the adapted JNetPcap packet
     */
    public org.jnetpcap.packet.PcapPacket getPcapPacket() {
        return pcapPacket;
    }

    private static PacketHeader createKrakenHeader(org.jnetpcap.packet.PcapPacket pcapPacket) {
        PcapHeader header = pcapPacket.getCaptureHeader();
        int origLen = header.wirelen();
        int inclLen = header.caplen();
        int hdr_sec = checkedInt(header.hdr_sec());
        int hdr_usec = header.hdr_usec();

        return new PacketHeader(hdr_sec, hdr_usec, inclLen, origLen);
    }

    private static Buffer createKrakenPayloadBuffer(org.jnetpcap.packet.PcapPacket pcapPacket) {
        Buffer krakenBuffer = new ChainBuffer();
        byte[] packetArray = pcapPacket.getByteArray(0, pcapPacket.size());
        krakenBuffer.addFirst(packetArray);
        return krakenBuffer;
    }

    private static int checkedInt(long l) {
        if (l < Integer.MIN_VALUE || l > Integer.MAX_VALUE) {
            throw new IllegalArgumentException(String.format("Cannot cast %d to int without data loss", l));
        }
        return (int) l;
    }

}
