/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.protocol;


import at.jku.fim.rubanetra.config.model.ProtocolId;

/**
 * This is the main interface that currently all provided protocol decoders have to implement.
 * The currently employed binding mechanism is up to the implementations, e.g. a {@link at.jku.fim.rubanetra.protocol.KrakenBaseProtocol}
 * based implementation relies on reflection in order to provide multiple {@link #bind(BaseProtocol)} overrides.
 * In order to provide an internal registry mechanism for this framework every protocol decoder must advertise a
 * unique protocol identifier that it is able to parse and decode.
 */
public interface BaseProtocol {
    /**
     * An arbitrary identifier for the protocol(s) this decoder should be able to parse and decode.
     * This identifier must be unique and may not be reused multiple times inside this framework for multiple decoders,
     * even if the same protocol can be decoded by multiple implementations.
     *
     * @return a unique identifier used to represent the decoding capabilities of this decoder
     */
    public abstract ProtocolId getProtocolId();

    /**
     * A protocol decoder should be able to pass its decoded data to another parser. Therefore it should be possible
     * to bind multiple decoders to form a decoding pipeline, similar to the Java I/O stream processing pipeline.
     * One way to implement the network stack would be to statically define which decoder is able to decode which
     * data on a certain layer. However, this approach seems quite error prone and must be adjusted regularly.
     * The currently utilized mechanism relies on the protocol decoder implementation advertisements and the user
     * specifications instead. This way, the protocol parsers can extend their capabilities easily by providing a new
     * implementation of this bind-method and the user may directly specify the decoding pipeline.
     * Usually, protocol decoders bind-methods are called depending on user input, therefore every implementing class
     * should handle errors gracefully.
     *
     * @param baseProtocol the protocol decoder to bind this decoder to, e.g. if base protocol refers to a decoder
     *                     capable of parsing IPv4 data and this class is able to decode Ethernet, and the user
     *                     specified a binding from Ethernet to IPv4, then Ethernet-decoder.bind(IPv4.decoder) will be
     *                     called once. All necessary steps must be taken by the caller and callee to ensure
     *                     successful data exchange. Currently the data exchange should be kept one way only,
     *                     i.e. data will flow from a lower layer to an upper layer only and not vice versa (unless
     *                     the user specified a binding from Ipv4 to Ethernet).
     */
    public abstract void bind(BaseProtocol baseProtocol);
}
