/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.protocol;


import at.jku.fim.rubanetra.config.model.KrakenProtocolConfiguration;
import at.jku.fim.rubanetra.pcap.PcapActivity;
import at.jku.fim.rubanetra.pcap.PcapActivityListener;
import at.jku.fim.rubanetra.pcap.PcapHandler;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

/**
 * An abstract definition of a possible Kraken protocol parser and decoder wrapper.
 * It provides an experimental implementation of the basic {@link at.jku.fim.rubanetra.protocol.BaseProtocol#bind(BaseProtocol)}}
 * specification which relies on reflection to provide a convenient way for subclasses to provide multiple bindings:
 * While it is possible for subclasses to override this mechanism, subclasses should instead define several bind-method
 * signatures directly, i.e. it should advertise possible bindings directly via the available bind-methods.
 * In order to rely on this feature is is mandatory for implementing classes to:
 * <ul>
 * <li>define one or multiple 'bind' methods with differing signatures</li>
 * <li>adhere to the naming convention, i.e. all methods must be called 'bind'</li>
 * <li>should not call 'bind' methods directly (bew aware of calling loops/recursion!),
 * as the input handler usually performs this step using a registry of available decoders</li>
 * </ul>
 * <p>
 * Alternatively, a subclass may disregard this feature and implement a suitable {@link #bind(BaseProtocol)} method
 * on its own.
 */
public abstract class KrakenBaseProtocol implements BaseProtocol {
    public static final String BINDING_METHOD_NAME = "bind";

    protected final Logger log = LoggerFactory.getLogger(getClass());
    protected final PropertyChangeSupport propertyChangeSupport = new PropertyChangeSupport(this);
    protected KrakenProtocolConfiguration protocolSetting;
    private boolean settingIsInitialized = false;
    private PcapHandler<PcapActivityListener> pcapHandler;

    /**
     * Tries to find a suitable bind-method for two concrete implementations of
     * {@link at.jku.fim.rubanetra.protocol.BaseProtocol} using reflection tools.
     * Both classes must share a common naming convention for bind methods, usually 'bind'.
     *
     * @param bindFrom          bind from one concrete protocol decoder ...
     * @param bindTo            ... to another protocol decoder
     * @param bindingMethodName the name that indicates the binding method
     * @return a Method to call for binding the two base protocol decoders together
     * @throws NoSuchMethodException if no such method was discovered or defined
     */
    public static Method findProtocolBindingMethod(BaseProtocol bindFrom, BaseProtocol bindTo,
                                                   String bindingMethodName) throws NoSuchMethodException {
        Class<?> visitable = bindTo.getClass();
        while (BaseProtocol.class.isAssignableFrom(visitable)) {
            Class<?> visitor = bindFrom.getClass(); // getClass();
            if (visitor.isInterface()) {
                // a binding implementation has to be found, but an interface
                // (and all super-interfaces) can not provide an implementation,
                // therefore no appropriate binding method was found
                throw new NoSuchMethodException(String.format(
                        "Unable to find a suitable binding method (from %s to %s)",
                        bindFrom.getClass(), bindTo.getClass()));
            }
            while (BaseProtocol.class.isAssignableFrom(visitor)) {
                try {
                    return visitor.getDeclaredMethod(bindingMethodName, visitable);
                } catch (NoSuchMethodException e) {
                    visitor = visitor.getSuperclass();
                }
            }
            visitable = visitable.getSuperclass();
        }
        throw new NoSuchMethodException(String.format(
                "Unable to find a suitable binding method (from %s to %s)", bindFrom.getClass(),
                bindTo.getClass()));

    }

    /**
     * Initializes this protocol decoder with the information provided by the user.
     * This method will be called once by {@link at.jku.fim.rubanetra.config.model.impl.KrakenProtocolConfigurationBuilderImpl}
     * before the bindings are processed.
     * This method should be called exactly once, multiple calls will result in an exception being thrown.
     *
     * @param krakenProtocolConfigurationImpl the protocol configuration provided by the user
     */
    public void initialize(KrakenProtocolConfiguration krakenProtocolConfigurationImpl) {
        if (settingIsInitialized) {
            throw new IllegalStateException(String.format("The Kraken Protocol %s was already initialized - multiple initializations are prohibited.",
                    getClass().getName()));
        }
        this.protocolSetting = krakenProtocolConfigurationImpl;
        this.pcapHandler = krakenProtocolConfigurationImpl.getPcapHandler();

        this.settingIsInitialized = true;
    }

    /**
     * @return the protocol configuration object obtained through initialization
     */
    public KrakenProtocolConfiguration getProtocolConfiguration() {
        return protocolSetting;
    }

    /**
     * @return the latest dispatched PcapActivity that may or may not be processed by this decoder
     */
    public PcapActivity getCurrentPcapActivity() {
        return pcapHandler.getCurrentPcapActivity();
    }

    /**
     * @return the utilized Pcap input stream
     */
    protected PcapHandler<PcapActivityListener> getPcapHandler() {
        return this.pcapHandler;
    }

    @Override
    public void bind(BaseProtocol baseProtocol) {
        // this implementation relies on reflection
        try {
            Method m = findProtocolBindingMethod(this, baseProtocol, BINDING_METHOD_NAME);
            m.invoke(this, baseProtocol);
        } catch (IllegalAccessException | IllegalArgumentException | InvocationTargetException
                | NoSuchMethodException e) {
            log.warn("Unable to perform protocol decoder binding.", e);
        }
    }

    /**
     * Adds a listener instance to this object, notifications and property names are subject to the individual
     * subclass implementations.
     *
     * @param l the listener to add
     */
    public void addPropertyChangeListener(PropertyChangeListener l) {
        this.propertyChangeSupport.addPropertyChangeListener(l);
    }

    /**
     * Adds a listener instance to this object, notifications and property names are subject to the individual
     * subclass implementations.
     *
     * @param propertyName the name of the property to listen for events
     * @param l            the listener to add
     */
    public void addPropertyChangeListener(String propertyName, PropertyChangeListener l) {
        this.propertyChangeSupport.addPropertyChangeListener(propertyName, l);
    }

    /**
     * Remove a property change listener instance from this objects, i.e. it will no longer receive any notifications
     * from this object.
     *
     * @param l the listener to remove
     */
    public void removePropertyChangeListener(PropertyChangeListener l) {
        this.propertyChangeSupport.removePropertyChangeListener(l);
    }

    /**
     * Remove a property change listener instance from this objects, i.e. it will no longer receive any notifications
     * from this object.
     *
     * @param propertyName the name of the property that was used to register this listener in the first place
     * @param l            the listener to remove
     */
    public void removePropertyChangeListener(String propertyName, PropertyChangeListener l) {
        this.propertyChangeSupport.removePropertyChangeListener(propertyName, l);
    }
}
