/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.protocol.activity;


import at.jku.fim.rubanetra.output.CsvRecordOutputWriter;
import com.fasterxml.jackson.annotation.JsonIgnoreProperties;
import com.fasterxml.jackson.annotation.JsonPropertyOrder;
import org.apache.commons.csv.CSVPrinter;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.io.IOException;
import java.time.Instant;
import java.util.*;

/**
 * A template for a non-replaceable {@link at.jku.fim.rubanetra.protocol.activity.Activity}.
 * It defines an "addFrameNumbers" property that will be used to link two activities regarding the addition
 * of frame numbers during the reasoning process.
 * Additionally an activity extension method is provided, see {@link #extendActivity(Activity)}.
 * If a replaceable activity is required, refer to {@link at.jku.fim.rubanetra.protocol.activity.AbstractReplaceableActivity}.
 */
@JsonIgnoreProperties({"source", "changeSupport", "startTimestamp", "endTimestamp", "excludedFromOutput", "csvSchemaDefinition"})
@JsonPropertyOrder({"activityType", "startInstant", "endInstant", "sourceAddress", "destinationAddress", "description", "compoundFrameNumbers", "optionalFields"})
public abstract class AbstractActivity extends EventObject implements Activity, CsvRecordOutputWriter {
    public static final String PROPERTY_ADD_FRAME_NUMBERS = "addFrameNumbers";
    public static final String PROPERTY_NEW_START_INSTANT = "newStartInstant";
    public static final String PROPERTY_NEW_END_INSTANT = "newEndInstant";

    private final PropertyChangeListener frameNumberPropertyListener = evt -> {
        if (evt.getPropertyName().equals(PROPERTY_ADD_FRAME_NUMBERS) && evt.getNewValue() instanceof Long[]) {
            addFrameNumbers((Long[]) evt.getNewValue());
        }
    };
    private final PropertyChangeListener startInstantListener = evt -> {
        if (evt.getPropertyName().equals(PROPERTY_NEW_START_INSTANT) && evt.getNewValue() instanceof Instant) {
            Instant newStartInstant = (Instant) evt.getNewValue();
            if (getStartInstant() == null || newStartInstant.isBefore(getStartInstant())) {
                setStartInstant(newStartInstant);
            }
        }
    };
    private final PropertyChangeListener endInstantListener = evt -> {
        if (evt.getPropertyName().equals(PROPERTY_NEW_END_INSTANT) && evt.getNewValue() instanceof Instant) {
            Instant newEndInstant = (Instant) evt.getNewValue();
            if (getEndInstant() == null || newEndInstant.isAfter(getEndInstant())) {
                setEndInstant(newEndInstant);
            }
        }
    };
    private static final long serialVersionUID = 3477186587937656979L;
    private final PropertyChangeSupport changeSupport = new PropertyChangeSupport(this);
    private final Collection<Long> compoundFrameNumbers = new TreeSet<>();
    private final Collection<OptionalField> optionalFields = new HashSet<>();
    private String sourceAddress, destinationAddress;
    private Instant startInstant, endInstant;
    private String description;
    private boolean excludedFromOutput;

    /**
     * Constructs a prototypical Activity event.
     *
     * @param source The object on which the Event initially occurred.
     * @throws IllegalArgumentException if source is null.
     */
    public AbstractActivity(Object source) {
        super(source);
    }

    @Override
    public String getActivityType() {
        return getClass().getSimpleName();
    }

    @Override
    public Date getStartTimestamp() {
        return Date.from(startInstant);
    }

    @Override
    public Instant getStartInstant() {
        return startInstant;
    }

    public void setStartInstant(Instant startInstant) {
        Instant oldInstant = this.startInstant;
        this.startInstant = startInstant;
        this.changeSupport.firePropertyChange(PROPERTY_NEW_START_INSTANT, oldInstant, this.startInstant);
    }

    @Override
    public Date getEndTimestamp() {
        return Date.from(endInstant);
    }

    @Override
    public Instant getEndInstant() {
        return endInstant;
    }

    public void setEndInstant(Instant endInstant) {
        Instant oldInstant = this.endInstant;
        this.endInstant = endInstant;
        this.changeSupport.firePropertyChange(PROPERTY_NEW_END_INSTANT, oldInstant, this.endInstant);
    }

    @Override
    public String getSourceAddressAsString() {
        return sourceAddress;
    }

    @Override
    public String getDestinationAddressAsString() {
        return destinationAddress;
    }

    @Override
    public String getDescription() {
        return description;
    }

    public void setDescription(String description) {
        this.description = description;
    }

    /**
     * Extends this activity by act. The following actions are performed:
     * <ul>
     * <li>all frame numbers from act are copied into this set of frame numbers</li>
     * <li>if act is chronologically before this activity, the start Instant of this activity is overwritten</li>
     * <li>if act is chronologically after this activity, the end Instant of this activity is overwritten</li>
     * <li>if act is chronologically during this activity, no instants are modified</li>
     * <li>an ADD_FRAME_NUMBER property listener is registered in act, in order to receive all future updates to the
     * frame number set</li>
     * <li>a NEW_START_INSTANT property listener is registered in act, in order to receive all future updates to the
     * start instant</li>
     * <li>a NEW_END_INSTANT property listener is registered in act, in order to receive all future updates to the
     * end instant</li>
     * </ul>
     * Unintended consequences may be encountered in the following scenario:
     * {@code act1.extendActivity(act2); act2.extendActivity(act1);}
     *
     * @param act the activity to be merged with this activity
     */
    public void extendActivity(Activity act) {
        Collection<Long> frameNumbersToAdd = act.getCompoundFrameNumbers();
        addFrameNumbers(frameNumbersToAdd.toArray(new Long[frameNumbersToAdd.size()]));
        if (act.getStartInstant() != null && (getStartInstant() == null || act.getStartInstant().isBefore(getStartInstant()))) {
            setStartInstant(act.getStartInstant());
        }
        if (act.getEndInstant() != null && (getEndInstant() == null || act.getEndInstant().isAfter(getEndInstant()))) {
            setEndInstant(act.getEndInstant());
        }
        act.addPropertyChangeListener(PROPERTY_ADD_FRAME_NUMBERS, frameNumberPropertyListener);
        act.addPropertyChangeListener(PROPERTY_NEW_START_INSTANT, startInstantListener);
        act.addPropertyChangeListener(PROPERTY_NEW_END_INSTANT, endInstantListener);
    }

    private void addFrameNumbers(Long... frameNumbers) {
        Collection<Long> oldFrames = new TreeSet<>(getCompoundFrameNumbers());
        for (long frameNumber : frameNumbers)
            this.compoundFrameNumbers.add(frameNumber);
        if (oldFrames.size() != getCompoundFrameNumbers().size()) {
            this.changeSupport.firePropertyChange(PROPERTY_ADD_FRAME_NUMBERS, oldFrames, frameNumbers);
        }
    }

    /**
     * Convenience method to add an optional field to the collection of all optional fields.
     *
     * @param field the optional field to add, may not be null
     */
    protected void addOptionalField(OptionalField field) {
        if (field == null) {
            throw new IllegalArgumentException();
        }
        this.optionalFields.add(field);
    }

    @Override
    public Collection<Long> getCompoundFrameNumbers() {
        return Collections.unmodifiableCollection(compoundFrameNumbers);
    }

    @Override
    public Collection<OptionalField> getOptionalFields() {
        return Collections.unmodifiableCollection(optionalFields);
    }

    public void setSourceAddress(String sourceAddress) {
        this.sourceAddress = sourceAddress;
    }

    public void setDestinationAddress(String destinationAddress) {
        this.destinationAddress = destinationAddress;
    }

    public void removeReplacedActivity(ReplaceableActivity act) {
        this.compoundFrameNumbers.removeAll(act.getCompoundFrameNumbers());
    }

    @Override
    public void addPropertyChangeListener(PropertyChangeListener l) {
        this.changeSupport.addPropertyChangeListener(l);
    }

    @Override
    public void addPropertyChangeListener(String propertyName, PropertyChangeListener l) {
        this.changeSupport.addPropertyChangeListener(propertyName, l);
    }

    @Override
    public void removePropertyChangeListener(PropertyChangeListener l) {
        this.changeSupport.removePropertyChangeListener(l);
    }

    @Override
    public void removePropertyChangeListener(String propertyName, PropertyChangeListener l) {
        this.changeSupport.removePropertyChangeListener(propertyName, l);
    }

    @Override
    public boolean isExcludedFromOutput() {
        return excludedFromOutput;
    }

    @Override
    public void setExcludedFromOutput(boolean isExcludedFromOutput) {
        this.excludedFromOutput = isExcludedFromOutput;
    }

    /**
     * Writes a single record to the provided csvPrinter. This record currently contains the following fields:
     * <pre>
     * activityType, startInstant, endInstant, sourceAddress, destinationAddress, description, frameNumbers, optionalFields
     * </pre>
     *
     * @param csvPrinter An initialized csvPrinter, shared among all activities
     * @throws IOException If the csvPrinter reaches an invalid state
     */
    @Override
    public void writeCsvRecord(CSVPrinter csvPrinter) throws IOException {
        csvPrinter.printRecord(getActivityType(), getStartInstant().toString(), getEndInstant().toString(),
                getSourceAddressAsString(), getDestinationAddressAsString(), getDescription(),
                getCompoundFrameNumbers(), getOptionalFields());
    }

    @Override
    public String toString() {
        return "AbstractActivity{" +
                "compoundFrameNumbers=" + compoundFrameNumbers +
                ", optionalFields=" + optionalFields +
                ", sourceAddress='" + sourceAddress + '\'' +
                ", destinationAddress='" + destinationAddress + '\'' +
                ", startInstant=" + startInstant +
                ", endInstant=" + endInstant +
                ", description='" + description + '\'' +
                ", excludedFromOutput=" + excludedFromOutput +
                '}';
    }
}
