/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.protocol.activity;


import java.beans.PropertyChangeListener;
import java.io.Serializable;
import java.time.Instant;
import java.util.Collection;
import java.util.Date;

/**
 * This interface constitutes the basis of all concrete activity implementations, i.e. it defines those fields that are
 * common across all activities and are by default available in the final output unless
 * {@link #setExcludedFromOutput(boolean)} has been called on an object to object basis.
 * However, implementations of this interface cannot be replaced by other activities by simple means, therefore the use
 * of {@link at.jku.fim.rubanetra.protocol.activity.ReplaceableActivity} is recommended instead.
 * Furthermore, the values returned by the getter methods are the only values that will be included in the CVS output
 * while the Json parser will include all fields of all subclasses.
 * Generally it is advisable to establish a meaningful implementation of all defined methods, nevertheless a template
 * is provided by {@link at.jku.fim.rubanetra.protocol.activity.AbstractActivity}.
 */
public interface Activity extends Serializable {
    /**
     * Defines the type of the activity - this value should be unique in the sense that it should be possible to
     * identify the class of the object by using this String exclusively among all activities.
     * This value will be used by a python json parser to setup the equivalent of the Java class/activity structure.
     *
     * @return a unique identifier of the implemented activity
     */
    String getActivityType();

    /**
     * Returns a fixed point in time that fixates the start of this activity's time-line, i.e. it returns a point
     * in time that fulfills any one of the following conditions:
     * <ul>
     * <li>if the activity is defined as an event that occurs at exactly one time without a duration/interval
     * then the returned value corresponds to exactly that time and
     * {@link #getEndInstant()} must return the same reference</li>
     * <li>if the activity spans over a period of time or includes multiple intervals,
     * then the returned value corresponds to the point in time that lies furthest in the past and starts
     * the first interval of the activity</li>
     * </ul>
     * Calling this method multiple times may return different instants, due to the reasoning process that may find
     * a prior starting point in time. Therefore the returned object should generally not be used as an identifier
     * for the activity.
     *
     * @return the point in time that lies furthest in the past and indicates the start of the activity
     */
    Instant getStartInstant();

    /**
     * Returns a fixed point in time that fixates the end of this activity's time-line, i.e. it returns a point
     * in time that fulfills any one of the following conditions:
     * <ul>
     * <li>if the activity is defined as an event that occurs at exactly one time without a duration/interval
     * then the returned value corresponds to exactly that time and
     * {@link #getStartInstant()} must return the same reference</li>
     * <li>if the activity spans over a period of time or includes multiple intervals,
     * then the returned value corresponds to the point in time that lies closest to the current time and ends
     * the last interval of the activity</li>
     * </ul>
     * Calling this method multiple times may return different instants, due to the reasoning process that may find
     * a later starting point in time. Therefore the returned object should generally not be used as an identifier
     * for the activity.
     *
     * @return the point in time that lies closest to the current time indicates the end of the activity
     */
    Instant getEndInstant();

    /**
     * An address that can be interpreted as the source address for this activity.
     * Usually this address should correspond to a human readable value.
     * Because different protocols may be defining different address formats and types, the returned value
     * is highly dependant on the actual implementation and may be empty or null if no source address exists at all.
     *
     * @return a human-readable protocol-dependant source address or null if no such address exists for this activity
     */
    String getSourceAddressAsString();

    /**
     * An address that can be interpreted as the destination address for this activity.
     * Usually this address should correspond to a human readable value.
     * Because different protocols may be defining different address formats and types, the returned value
     * is highly dependant on the actual implementation and may be empty or null if no destination address exists at all.
     *
     * @return a human-readable protocol-dependant destination address or null if no such address exists for this activity
     */
    String getDestinationAddressAsString();

    /**
     * @return an arbitrary String that describes this activity, may be null or empty
     */
    String getDescription();

    /**
     * This method's return value indicates whether or not this activity will be included in the final output.
     * If this method returns false, the activity will be included in the final output.
     *
     * @return true, iff this activity should be excluded from the final output
     */
    boolean isExcludedFromOutput();

    /**
     * Setter for {@link #isExcludedFromOutput()}.
     *
     * @param excludedFromOutput whether or not to exclude this activity from the final output
     */
    void setExcludedFromOutput(boolean excludedFromOutput);

    /**
     * This method returns a sorted collection of all frame numbers in consecutive order and greater than zero, that
     * define this activity. It serves as basis for a manual review of the packets that were used during the reasoning
     * process, i.e. a future use case could consist of using these numbers to extract all frames corresponding to
     * a single activity individually from a potentially large PCAP file.
     *
     * @return a sorted collection of all frame numbers in consecutive order and greater than zero, that
     * define this activity
     */
    Collection<Long> getCompoundFrameNumbers();

    /**
     * This method returns a collection of attribute/value pairs that have to be included in the final output of all
     * output processors including the CVS encoder and the JSON encoder. This serves as an opportunity to include
     * information that would usually not be available due to the fixed nature of CVS columns specification.
     *
     * @return a collection of attribute/value pairs that have to be included in the final output of all
     * output processors including the CVS encoder and the JSON encoder.
     */
    Collection<OptionalField> getOptionalFields();

    /**
     * Adds a property change listener for all properties of this activity.
     *
     * @param l the listener to add
     */
    void addPropertyChangeListener(PropertyChangeListener l);

    /**
     * Adds a property change listener for one exactly defined property of this activity.
     *
     * @param propertyName the name of property to receive notifications for
     * @param l            the listener to add
     */
    void addPropertyChangeListener(String propertyName, PropertyChangeListener l);

    /**
     * Removes the specified listener.
     *
     * @param l the listener to remove
     */
    void removePropertyChangeListener(PropertyChangeListener l);

    /**
     * The specified listener will no longer receive notifications for propertyName.
     *
     * @param propertyName the name of property to stop receiving notifications for
     * @param l            the listener to remove
     */
    void removePropertyChangeListener(String propertyName, PropertyChangeListener l);

    /**
     * This method should only be used within Drools rule-files and serve as timestamp indicator. This is currently
     * required due to the general incompatibility of {@link java.time.Instant} caused by the Drools engine pseudo-clock implementation.
     *
     * @return a timestamp in MILLISECOND precision (backed by an {@link java.time.Instant}), representing
     * the starting point in time for this {@link at.jku.fim.rubanetra.protocol.activity.Activity}.
     * Please note that this method returns an object which may not be persistent, i.e. the start timestamp
     * object may be overwritten by implementations.
     * @see #getStartInstant()
     * @see at.jku.fim.rubanetra.protocol.activity.AbstractActivity#setStartInstant(java.time.Instant)
     * @see at.jku.fim.rubanetra.protocol.activity.AbstractReplaceableActivity#setStartInstant(java.time.Instant)
     */
    Date getStartTimestamp();

    /**
     * This method should only be used within Drools rule-files and serve as timestamp indicator. This is currently
     * required due to the general incompatibility of {@link java.time.Instant} caused by the Drools engine pseudo-clock implementation.
     *
     * @return a timestamp in MILLISECOND precision (backed by an {@link java.time.Instant}), representing
     * the ending point in time for this {@link at.jku.fim.rubanetra.protocol.activity.Activity}.
     * Please note that this method returns an object which may not be persistent, i.e. the end timestamp
     * object may be overwritten by implementations.
     * @see #getEndInstant()
     * @see at.jku.fim.rubanetra.protocol.activity.AbstractActivity#setEndInstant(java.time.Instant)
     * @see at.jku.fim.rubanetra.protocol.activity.AbstractReplaceableActivity#setEndInstant(java.time.Instant)
     */
    Date getEndTimestamp();
}
