/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.protocol.activity.arp;


import at.jku.fim.rubanetra.pcap.PcapActivity;
import at.jku.fim.rubanetra.protocol.activity.AbstractReplaceableActivity;
import org.jnetpcap.packet.PcapPacket;
import org.jnetpcap.protocol.network.Arp;
import org.kie.api.definition.type.Role;
import org.kie.api.definition.type.Role.Type;
import org.krakenapps.pcap.decoder.arp.ArpPacket;
import org.krakenapps.pcap.decoder.ethernet.MacAddress;

import java.net.InetAddress;

/**
 * An encapsulation of an {@link org.krakenapps.pcap.decoder.arp.ArpPacket} including IP information if available.
 * This class also provides the JNetPcap specific Arp packet representation (if available)
 * {@link org.jnetpcap.protocol.network.Arp}.
 */
@Role(Type.EVENT)
public class ArpActivity extends AbstractReplaceableActivity {
    private static final long serialVersionUID = -6124545150681044570L;
    private final PcapActivity pcapActivity;
    private final int hardwareType;
    private final int protocolType;
    private final int hardwareAddressLength;
    private final int protocolAddressLength;
    private final int operation;
    private final MacAddress senderHardwareAddress;
    private final InetAddress senderProtocolAddress;
    private final MacAddress targetHardwareAddress;
    private final InetAddress targetProtocolAddress;
    private final Arp arp;

    /**
     * Default constructor for this class, the source pcap activity will be replaced.
     *
     * @param source on which the ArpActivity occurred or the final PcapActivity that led to the fully decoded packet.
     * @param p      the parsed Arp-packet
     */
    public ArpActivity(PcapActivity source, ArpPacket p) {
        super(source);
        this.pcapActivity = source;
        hardwareType = p.getHardwareType();
        protocolType = p.getProtocolType();
        hardwareAddressLength = p.getHardwareSize();
        protocolAddressLength = p.getProtocolSize();
        operation = p.getOpcode();
        senderHardwareAddress = p.getSenderMac();
        senderProtocolAddress = p.getSenderIp();
        targetHardwareAddress = p.getTargetMac();
        targetProtocolAddress = p.getTargetIp();
        arp = new Arp();
        PcapPacket packet = pcapActivity.getPcapPacket();
        packet.hasHeader(getArp());

        replaceActivity(source);
    }

    /**
     * @return the source on which the ArpActivity occurred or the final PcapActivity that led to the fully
     * decoded packet.
     */
    public PcapActivity getPcapActivity() {
        return pcapActivity;
    }

    /**
     * @see org.krakenapps.pcap.decoder.arp.ArpPacket#getHardwareType()
     */
    public int getHardwareType() {
        return hardwareType;
    }

    /**
     * @see org.krakenapps.pcap.decoder.arp.ArpPacket#getProtocolType()
     */
    public int getProtocolType() {
        return protocolType;
    }

    /**
     * @see org.krakenapps.pcap.decoder.arp.ArpPacket#getHardwareSize()
     */
    public int getHardwareAddressLength() {
        return hardwareAddressLength;
    }

    /**
     * @see org.krakenapps.pcap.decoder.arp.ArpPacket#getProtocolSize()
     */
    public int getProtocolAddressLength() {
        return protocolAddressLength;
    }

    /**
     * @see org.krakenapps.pcap.decoder.arp.ArpPacket#getOpcode()
     */
    public int getOperation() {
        return operation;
    }

    /**
     * @see org.krakenapps.pcap.decoder.arp.ArpPacket#getSenderMac()
     */
    public MacAddress getSenderHardwareAddress() {
        return senderHardwareAddress;
    }

    /**
     * @see org.krakenapps.pcap.decoder.arp.ArpPacket#getSenderIp()
     */
    public InetAddress getSenderProtocolAddress() {
        return senderProtocolAddress;
    }

    /**
     * @see org.krakenapps.pcap.decoder.arp.ArpPacket#getTargetMac()
     */
    public MacAddress getTargetHardwareAddress() {
        return targetHardwareAddress;
    }

    /**
     * @see org.krakenapps.pcap.decoder.arp.ArpPacket#getTargetIp()
     */
    public InetAddress getTargetProtocolAddress() {
        return targetProtocolAddress;
    }

    /**
     * @return the JNetPcap specific Arp packet representation
     */
    public Arp getArp() {
        return arp;
    }

    @Override
    public String getSourceAddressAsString() {
        return String.valueOf(getSenderHardwareAddress());
    }

    @Override
    public String getDestinationAddressAsString() {
        return String.valueOf(getTargetHardwareAddress());
    }

    @Override
    public String toString() {
        return "ArpActivity{" +
                "pcapActivity=" + pcapActivity +
                ", hardwareType=" + hardwareType +
                ", protocolType=" + protocolType +
                ", hardwareAddressLength=" + hardwareAddressLength +
                ", protocolAddressLength=" + protocolAddressLength +
                ", operation=" + operation +
                ", senderHardwareAddress=" + senderHardwareAddress +
                ", senderProtocolAddress=" + senderProtocolAddress +
                ", targetHardwareAddress=" + targetHardwareAddress +
                ", targetProtocolAddress=" + targetProtocolAddress +
                ", arp=" + arp +
                '}';
    }
}
