/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.protocol.activity.dns;


import at.jku.fim.rubanetra.pcap.PcapActivity;
import at.jku.fim.rubanetra.protocol.activity.AbstractReplaceableActivity;
import com.fasterxml.jackson.annotation.JsonIgnore;
import org.kie.api.definition.type.Role;
import org.kie.api.definition.type.Role.Type;
import org.xbill.DNS.*;

import java.util.Arrays;
import java.util.Collections;
import java.util.List;

/**
 * An encapsulation of a decoded {@link org.xbill.DNS.Message}.
 * The default output stream handler {@link at.jku.fim.rubanetra.output.impl.JsonOutputWriterStrategy}
 * will not include the original {@link org.xbill.DNS.Message} object due to space conservation issues. Instead, the
 * available question, answer, authority and additional {@link org.xbill.DNS.Record}s are included as lists and the
 * {@link org.xbill.DNS.Header} as it was encountered.
 */
@Role(Type.EVENT)
public class DnsActivity extends AbstractReplaceableActivity {
    private static final long serialVersionUID = -476178927847593388L;
    private final Message dnsMessage;
    private final List<Record> questionRecords;
    private final List<Record> answerRecords;
    private final List<Record> authorityRecords;
    private final List<Record> additionalRecords;
    private final Header dnsMessageHeader;
    private final boolean isResponse;

    private final PcapActivity pcapActivity;

    /**
     * Constructs a new DnsActivity, while replacing the source pcap activity and wrapping the decoded DNS message
     *
     * @param source     on which the DnsActivity occurred or the final PcapActivity that led to the fully decoded packet.
     * @param dnsMessage the fully decoded DNS content
     */
    public DnsActivity(PcapActivity source, Message dnsMessage) {
        super(source);
        this.dnsMessage = dnsMessage;
        questionRecords = Arrays.asList(dnsMessage.getSectionArray(Section.QUESTION));
        answerRecords = Arrays.asList(dnsMessage.getSectionArray(Section.ANSWER));
        authorityRecords = Arrays.asList(dnsMessage.getSectionArray(Section.AUTHORITY));
        additionalRecords = Arrays.asList(dnsMessage.getSectionArray(Section.ADDITIONAL));
        dnsMessageHeader = dnsMessage.getHeader();
        isResponse = dnsMessage.getHeader().getFlag(Flags.QR);
        this.pcapActivity = source;

        replaceActivity(source);
    }

    /**
     * @return the decoded DNS content as is
     */
    @JsonIgnore
    public Message getDnsMessage() {
        return dnsMessage;
    }

    /**
     * @return an unmodifiable list of the available question records
     */
    public List<Record> getQuestionRecords() {
        return Collections.unmodifiableList(questionRecords);
    }

    /**
     * @return an unmodifiable list of the available answer records
     */
    public List<Record> getAnswerRecords() {
        return Collections.unmodifiableList(answerRecords);
    }

    /**
     * @return an unmodifiable list of the available authority records
     */
    public List<Record> getAuthorityRecords() {
        return Collections.unmodifiableList(authorityRecords);
    }

    /**
     * @return an unmodifiable list of the available additional records
     */
    public List<Record> getAdditionalRecords() {
        return Collections.unmodifiableList(additionalRecords);
    }

    /**
     * @return the DNS message header
     */
    public Header getDnsMessageHeader() {
        return dnsMessageHeader;
    }

    /**
     * @return flag, indicating whether or not this message is a response to a query
     */
    public boolean isResponse() {
        return isResponse;
    }

    /**
     * @return the source on which the DnsActivity occurred or the final PcapActivity that led to the fully
     * decoded packet.
     */
    public PcapActivity getPcapActivity() {
        return pcapActivity;
    }

    @Override
    public String toString() {
        return "DnsActivity{" +
                "dnsMessage=" + dnsMessage +
                ", questionRecords=" + questionRecords +
                ", answerRecords=" + answerRecords +
                ", authorityRecords=" + authorityRecords +
                ", additionalRecords=" + additionalRecords +
                ", dnsMessageHeader=" + dnsMessageHeader +
                ", isResponse=" + isResponse +
                ", pcapActivity=" + pcapActivity +
                '}';
    }
}
