/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.protocol.activity.ethernet;


import at.jku.fim.rubanetra.pcap.PcapActivity;
import at.jku.fim.rubanetra.protocol.activity.AbstractReplaceableActivity;
import org.jnetpcap.packet.PcapPacket;
import org.jnetpcap.protocol.lan.Ethernet;
import org.jnetpcap.protocol.lan.Ethernet.EthernetType;
import org.kie.api.definition.type.Role;
import org.kie.api.definition.type.Role.Type;
import org.krakenapps.pcap.decoder.ethernet.EthernetFrame;
import org.krakenapps.pcap.decoder.ethernet.MacAddress;

/**
 * This class encapsulates an {@link org.jnetpcap.protocol.lan.Ethernet} object, which should provide a vast amount
 * of metadata including the content buffer itself. If this activity is created before it has been processed by
 * subsequent protocol parsers, the content buffer should not be modified in any way, i.e. it should be treated as
 * read-only medium.
 * ENHANCEMENT: encapsulate all metadata fields and prohibit access to the content buffer itself
 */
@Role(Type.EVENT)
public class EthernetActivity extends AbstractReplaceableActivity {
    private static final long serialVersionUID = -8495525878251669060L;
    private final PcapActivity pcapActivity;
    private final MacAddress destinationMacAddress, sourceMacAddress;
    private final int ethernetType;
    private final EthernetType ethernetTypeEnum;
    private final Ethernet ethernet;

    /**
     * Constructs a new ethernet activity, replaces the source pcap activity on which the event occurred.
     *
     * @param source the Pcap activity on which the event occurred. This activity will be excluded from the
     *               output, unless {@link #setExcludedFromOutput(boolean)} is called with 'false'.
     * @param frame  the decoded Ethernet frame (by Kraken)
     */
    public EthernetActivity(PcapActivity source, EthernetFrame frame) {
        super(source);
        this.pcapActivity = source;
        this.destinationMacAddress = frame.getDestination();
        this.sourceMacAddress = frame.getSource();
        this.ethernetType = frame.getType();
        this.ethernetTypeEnum = EthernetType.valueOf(getEthernetType());
        this.ethernet = new Ethernet();
        scanAndFillEthernet();

        replaceActivity(source);
        setExcludedFromOutput(true);
    }

    private void scanAndFillEthernet() {
        PcapPacket pcapPacket = this.pcapActivity.getPcapPacket();
        pcapPacket.hasHeader(ethernet);
    }

    /**
     * @return the Ethernet frame as decoded by JNetPcap
     */
    public Ethernet getEthernet() {
        return this.ethernet;
    }

    /**
     * @return the physical Ethernet address (source)
     */
    public MacAddress getSourceMacAddress() {
        return this.sourceMacAddress;
    }

    /**
     * @return the physical Ethernet address (destination)
     */
    public MacAddress getDestinationMacAddress() {
        return this.destinationMacAddress;
    }

    /**
     * @return The payload type as integer
     * @see #getEthernetTypeEnum()
     */
    public int getEthernetType() {
        return this.ethernetType;
    }

    /**
     * @return the payload type as JNetPcap enumeration object, however, be aware that this enumeration may not be
     * complete
     * @see #getEthernetType()
     */
    public EthernetType getEthernetTypeEnum() {
        return ethernetTypeEnum;
    }

    /**
     * @return the source on which this activity occurred or the final PcapActivity that led to the fully
     * decoded frame.
     */
    public PcapActivity getPcapActivity() {
        return pcapActivity;
    }

    @Override
    public String getSourceAddressAsString() {
        return String.valueOf(getSourceMacAddress());
    }

    @Override
    public String getDestinationAddressAsString() {
        return String.valueOf(getDestinationMacAddress());
    }

    @Override
    public String toString() {
        return "EthernetActivity{" +
                "pcapActivity=" + pcapActivity +
                ", destinationMacAddress=" + destinationMacAddress +
                ", sourceMacAddress=" + sourceMacAddress +
                ", ethernetType=" + ethernetType +
                ", ethernetTypeEnum=" + ethernetTypeEnum +
                ", ethernet=" + ethernet +
                '}';
    }
}
