/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.protocol.activity.ftp;

import at.jku.fim.rubanetra.pcap.PcapActivity;
import at.jku.fim.rubanetra.protocol.activity.AbstractReplaceableActivity;

import java.util.Arrays;

/**
 * TODO: this class represent a partial ftp activity, it must be reassembled by Drools rules based on the tcp socket addresses to represent a FTPSession
 */
public class FtpActivity extends AbstractReplaceableActivity {
    private static final long serialVersionUID = 8828875279114236211L;
    private final PcapActivity pcapActivity;
    private final Type ftpActivityType;

    private String command, reply;
    private byte[] list;

    /**
     * Constructs a prototypical activity, replaces the source Pcap activity and stores the type of the FTP message..
     *
     * @param source The object on which the activity initially occurred.
     * @param type the FTP message type leading to a decoded String, i.e. either COMMAND or REPLY
     * @throws IllegalArgumentException if source is null or an unsupported type was encountered.
     */
    public FtpActivity(PcapActivity source, Type type, String decodedMessage) {
        super(source);
        pcapActivity = source;
        ftpActivityType = type;
        switch (ftpActivityType) {
            case COMMAND:
                this.command = decodedMessage;
                break;
            case REPLY:
                this.reply = decodedMessage;
                break;
            default:
                throw new IllegalArgumentException(String.format("Unsupported FTP message type: %s", type));
        }

        replaceActivity(source);
    }

    /**
     * Constructs a prototypical activity, replaces the source Pcap activity and stores the type of the FTP message..
     *
     * @param source The object on which the activity initially occurred.
     * @param type the FTP message type leading to a decoded byte [], i.e. VIEW_LIST.
     * @throws IllegalArgumentException if source is null or an unsupported type was encountered.
     */
    public FtpActivity(PcapActivity source, Type type, byte [] decodedMessage) {
        super(source);
        pcapActivity = source;
        ftpActivityType = type;
        switch (ftpActivityType) {
            case VIEW_LIST:
                this.list = decodedMessage;
                break;
            default:
                throw new IllegalArgumentException(String.format("Unsupported FTP message type: %s", type));
        }

        replaceActivity(source);
    }

    /**
     * @return the source on which this activity occurred or the final PcapActivity that led to the fully
     * decoded frame.
     */
    public PcapActivity getPcapActivity() {
        return pcapActivity;
    }

    /**
     * @return the FTP message type as enumeration
     */
    public Type getFtpActivityTypeEnum() {
        return ftpActivityType;
    }

    /**
     * @return the command string, if the type of this activity is equal to COMMAND or null otherwise.
     */
    public String getCommand() {
        return command;
    }

    /**
     * @return the decoded reply string, if the type of this activity is equal to REPLY or null otherwise
     */
    public String getReply() {
        return reply;
    }

    /**
     * @return the result of the VIEW_LIST command or null
     */
    public byte[] getList() {
        return list;
    }

    public enum Type {
        COMMAND, REPLY, VIEW_LIST
    }

    @Override
    public String toString() {
        return "FtpActivity{" +
                "pcapActivity=" + pcapActivity +
                ", ftpActivityType=" + ftpActivityType +
                ", command='" + command + '\'' +
                ", reply='" + reply + '\'' +
                ", list=" + Arrays.toString(list) +
                '}';
    }
}
