/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.protocol.activity.http;


import at.jku.fim.rubanetra.protocol.activity.AbstractReplaceableActivity;
import at.jku.fim.rubanetra.protocol.activity.dns.DnsActivity;
import org.kie.api.definition.type.Role;
import org.kie.api.definition.type.Role.Type;

import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

/**
 * A high level aggregation of a {@link at.jku.fim.rubanetra.protocol.activity.http.HttpRequestActivity}
 * and a {@link at.jku.fim.rubanetra.protocol.activity.http.HttpResponseActivity}.
 * Usually instances of this class are created directly by Drools rules, i.e. a request is matched to a response
 * by one or several rules.
 * Additionally, this class defines several optional fields, e.g. a potential preceding Dns-Match to this Http-Activity
 * and multiple {@link at.jku.fim.rubanetra.protocol.activity.http.HttpImageActivity} objects may be
 * aggregated as well.
 */
public class HttpActivity extends AbstractReplaceableActivity {
    private static final long serialVersionUID = 2723756798789149547L;
    private final HttpRequestActivity request;
    private final HttpResponseActivity response;
    private final Set<HttpImageActivity> imageActivities;
    private DnsActivity dnsMatch;

    /**
     * Constructs a new Http-Activity and replaces both, the passed request and the response.
     *
     * @param req  the request that is followed by ...
     * @param resp ... the response for the request
     */
    public HttpActivity(HttpRequestActivity req, HttpResponseActivity resp) {
        super(req);
        this.request = req;
        this.response = resp;
        this.dnsMatch = null;
        this.imageActivities = new HashSet<>();
        replaceActivity(req);
        replaceActivity(resp);
    }

    /**
     * @return the Http response for the {@link #getRequest()}
     */
    public HttpResponseActivity getResponse() {
        return response;
    }

    /**
     * @return the Http request preceding the {@link #getResponse()}
     */
    public HttpRequestActivity getRequest() {
        return request;
    }

    /**
     * The interpretation of this field depends on the actual specification of the rule that filled this variable in the
     * first place.
     * @return a potential DNS-based query leading to the represented Http request/response
     */
    public DnsActivity getDnsMatch() {
        return dnsMatch;
    }

    /**
     * Calling this method results in the following actions:
     * <ul>
     *     <li>if this method was called previously, the old dnsMatch object is will no longer considered to be a
     *     replaced activity and the assigned frames are removed from the frame set</li>
     *     <li>the dnsMatch reference is saved and the DNS activity will be replaced</li>
     * </ul>
     * @param dnsMatch a DNS activity that potentially led to this HTTP activity (speculative)
     */
    public void setDnsMatch(DnsActivity dnsMatch) {
        if (this.dnsMatch != null) {
            super.removeReplacedActivity(this.dnsMatch);
        }
        this.dnsMatch = dnsMatch;
        replaceActivity(dnsMatch);
    }

    /**
     * The interpretation of this field depends on the actual definition of the Drools rules that will fill this set.
     * @return a collection of activities that potentially originated from this HTTP activity.
     */
    public Set<HttpImageActivity> getImageActivities() {
        return Collections.unmodifiableSet(imageActivities);
    }

    /**
     * Adds and replaces a Http image activity which origins may be considered to lie within this HTTP activity.
     * The image activities are currently excluded from output, unless this is overridden manually.
     * @param imgAct a image activity that was caused by this HTTP activity
     */
    public void addImageActivity(HttpImageActivity imgAct) {
        if (this.imageActivities.add(imgAct)) {
            replaceActivity(imgAct);
            imgAct.setExcludedFromOutput(true);
        }
    }

    @Override
    public String getSourceAddressAsString() {
        if(getRequest() != null) {
            return String.valueOf(getRequest().getClientAddress());
        } else {
            return null;
        }
    }

    @Override
    public String getDestinationAddressAsString() {
        if(getRequest() != null) {
            return String.valueOf(getRequest().getServerAddress());
        } else {
            return null;
        }
    }

    @Override
    public String toString() {
        return "HttpActivity{" +
                "request=" + request +
                ", response=" + response +
                ", imageActivities=" + imageActivities +
                ", dnsMatch=" + dnsMatch +
                '}';
    }
}
