/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.protocol.activity.http;


import at.jku.fim.rubanetra.pcap.PcapActivity;
import at.jku.fim.rubanetra.protocol.activity.AbstractReplaceableActivity;
import org.apache.http.NameValuePair;
import org.apache.http.client.utils.URLEncodedUtils;
import org.jnetpcap.packet.PcapPacket;
import org.jnetpcap.protocol.tcpip.Http;
import org.kie.api.definition.type.Role;
import org.kie.api.definition.type.Role.Type;
import org.krakenapps.pcap.decoder.http.HttpMethod;
import org.krakenapps.pcap.decoder.http.HttpRequest;
import org.krakenapps.pcap.decoder.http.HttpVersion;

import java.net.InetSocketAddress;
import java.net.URL;
import java.nio.charset.Charset;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Encapsulates a {@link org.krakenapps.pcap.decoder.http.HttpRequest}, that was decoded by the Kraken library.
 * The goal of this class is to provide a read-only interface to the encountered Http-related information, however,
 * currently the actual content buffer that was used by the decoder can both be read from and written to.
 * Furthermore, a JNetPcap representation is provided in addition to the Kraken-representation.
 */
@Role(Type.EVENT)
public class HttpRequestActivity extends AbstractReplaceableActivity {
    private static final long serialVersionUID = -8446692870521248194L;
    private final HttpVersion httpVersion;
    private final InetSocketAddress serverAddress;
    private final InetSocketAddress clientAddress;
    private final HttpMethod httpMethod;
    private final String httpQueryString;
    private final List<NameValuePair> httpQueryParameters;
    private final Map<String, String> requestHeaderMap;
    private final PcapActivity pcapActivity;
    private final URL url;
    private final Http httpRequest;

    /**
     * Constructs a new request activity and replaces the pcap source activity on which this event occurred.
     * It decodes the query parameters using the default Charset and {@link org.apache.http.client.utils.URLEncodedUtils}.
     *
     * @param source  the Pcap activity on which this activity occurred
     * @param request the decoded Http request
     */
    public HttpRequestActivity(PcapActivity source, HttpRequest request) {
        super(source);
        this.requestHeaderMap = new HashMap<>();
        for (String headerKey : request.getHeaderKeys()) {
            try {
                HttpHeader canonicalHeader = HttpHeader.canonicalizeHeader(headerKey);
                requestHeaderMap.put(canonicalHeader.toString(), request.getHeader(headerKey));
            } catch (IllegalArgumentException e) {
                requestHeaderMap.put(headerKey, request.getHeader(headerKey));
            }
        }
        httpVersion = request.getHttpVersion();
        serverAddress = request.getLocalAddress();
        clientAddress = request.getRemoteAddress();
        httpMethod = request.getMethod();
        httpQueryString = request.getQueryString();
        httpQueryParameters = URLEncodedUtils.parse(httpQueryString, Charset.defaultCharset());
        pcapActivity = source;
        url = request.getURL();

        httpRequest = new Http();
        scanAndFillHttpRequest();
        replaceActivity(source);
    }

    private void scanAndFillHttpRequest() {
        PcapPacket packet = pcapActivity.getPcapPacket();
        packet.hasHeader(getHttpRequest());
    }

    /**
     * @return the Kraken-specific representation of the version field
     */
    public HttpVersion getHttpVersion() {
        return httpVersion;
    }

    /**
     * @return the destination address (Layer 3 and 4)
     */
    public InetSocketAddress getServerAddress() {
        return serverAddress;
    }

    /**
     * @return the source address (Layer 3 and 4)
     */
    public InetSocketAddress getClientAddress() {
        return clientAddress;
    }

    /**
     * @return the Kraken-specific representation of the method field
     */
    public HttpMethod getHttpMethod() {
        return httpMethod;
    }

    /**
     * @return the unmodified query string
     */
    public String getHttpQueryString() {
        return httpQueryString;
    }

    /**
     * @return a collection of the decoded query-parameters (using the default charset)
     */
    public List<NameValuePair> getHttpQueryParameters() {
        return Collections.unmodifiableList(httpQueryParameters);
    }

    /**
     * @param key the header key
     * @return the value for the key, or null if the key was not defined
     */
    public String getRequestHeader(String key) {
        return this.requestHeaderMap.get(key);
    }

    /**
     * @return an unmodifiable collection containing all header keys and values
     */
    public Map<String, String> getRequestHeaderMap() {
        return Collections.unmodifiableMap(requestHeaderMap);
    }

    /**
     * @return the source on which this activity occurred or the final PcapActivity that led to the fully
     * decoded packet.
     */
    public PcapActivity getPcapActivity() {
        return pcapActivity;
    }

    /**
     * @return an URL representation of the destination resource
     */
    public URL getUrl() {
        return url;
    }

    /**
     * @return the JNetPcap specific representation of the request, if available
     */
    public Http getHttpRequest() {
        return httpRequest;
    }

    @Override
    public String getSourceAddressAsString() {
        return String.valueOf(getClientAddress());
    }

    @Override
    public String getDestinationAddressAsString() {
        return String.valueOf(getServerAddress());
    }

    @Override
    public String toString() {
        return "HttpRequestActivity{" +
                "httpVersion=" + httpVersion +
                ", serverAddress=" + serverAddress +
                ", clientAddress=" + clientAddress +
                ", httpMethod=" + httpMethod +
                ", httpQueryString='" + httpQueryString + '\'' +
                ", httpQueryParameters=" + httpQueryParameters +
                ", requestHeaderMap=" + requestHeaderMap +
                ", pcapActivity=" + pcapActivity +
                ", url=" + url +
                ", httpRequest=" + httpRequest +
                '}';
    }
}