/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.protocol.activity.http;


import at.jku.fim.rubanetra.pcap.PcapActivity;
import at.jku.fim.rubanetra.protocol.activity.AbstractReplaceableActivity;
import org.jnetpcap.packet.PcapPacket;
import org.jnetpcap.protocol.tcpip.Http;
import org.kie.api.definition.type.Role;
import org.krakenapps.pcap.decoder.http.HttpResponse;
import org.krakenapps.pcap.decoder.http.HttpVersion;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

/**
 * Encapsulates a {@link org.krakenapps.pcap.decoder.http.HttpResponse}, that was decoded by the Kraken library.
 * The goal of this class is to provide a read-only interface to the encountered Http-related information, however,
 * currently the actual content buffer that was used by the decoder can both be read from and written to.
 * Furthermore, a JNetPcap representation is provided in addition to the Kraken-representation.
 */
@Role(Role.Type.EVENT)
public class HttpResponseActivity extends AbstractReplaceableActivity {
    private static final long serialVersionUID = 8779177553506884946L;
    private final HttpVersion httpVersion;
    private final int responseStatusCode;
    private final String responseStatusLine;
    private final Map<String, String> responseHeaderMap;
    private final PcapActivity pcapActivity;
    private final Http httpResponse;

    /**
     * Constructs a new response activity and replaces the source pcap activity on which this event occurred.
     * Tries to canonicalize the encountered headers using
     * {@link at.jku.fim.rubanetra.protocol.activity.http.HttpHeader}, if it fails
     * the failed header is stored as encountered.
     *
     * @param source   the pcap activity on which this event occurred
     * @param response the Http response, decoded by the Kraken library
     */
    public HttpResponseActivity(PcapActivity source, HttpResponse response) {
        super(source);
        this.responseHeaderMap = new HashMap<>();
        for (String headerKey : response.getHeaderKeys()) {
            try {
                HttpHeader canonicalHeader = HttpHeader.canonicalizeHeader(headerKey);
                responseHeaderMap.put(canonicalHeader.toString(), response.getHeader(headerKey));
            } catch (IllegalArgumentException e) {
                responseHeaderMap.put(headerKey, response.getHeader(headerKey));
            }
        }
        httpVersion = response.getHttpVersion();
        responseStatusCode = response.getStatusCode();
        responseStatusLine = response.getStatusLine();
        pcapActivity = source;

        httpResponse = new Http();
        scanAndFillHttpResponse();
        replaceActivity(source);
    }

    private void scanAndFillHttpResponse() {
        PcapPacket packet = pcapActivity.getPcapPacket();
        packet.hasHeader(getHttpResponse());
    }

    /**
     * @return the Kraken specific representation of the Http version field
     */
    public HttpVersion getHttpVersion() {
        return httpVersion;
    }

    /**
     * @return the HTTP response code as encountered
     */
    public int getResponseStatusCode() {
        return responseStatusCode;
    }

    /**
     * @return Http response status line as encountered
     */
    public String getResponseStatusLine() {
        return responseStatusLine;
    }

    /**
     * @return an unmodifiable collection containing the decoded headers
     */
    public Map<String, String> getResponseHeaderMap() {
        return Collections.unmodifiableMap(responseHeaderMap);
    }

    /**
     * @return the source on which this activity occurred or the final PcapActivity that led to the fully
     * decoded packet.
     */
    public PcapActivity getPcapActivity() {
        return pcapActivity;
    }

    /**
     * @return the JNetPcap specific representation of the Http response
     */
    public Http getHttpResponse() {
        return httpResponse;
    }

    @Override
    public String toString() {
        return "HttpResponseActivity{" +
                "httpVersion=" + httpVersion +
                ", responseStatusCode=" + responseStatusCode +
                ", responseStatusLine='" + responseStatusLine + '\'' +
                ", responseHeaderMap=" + responseHeaderMap +
                ", pcapActivity=" + pcapActivity +
                ", httpResponse=" + httpResponse +
                '}';
    }
}