/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.protocol.activity.icmp;


import at.jku.fim.rubanetra.pcap.PcapActivity;
import at.jku.fim.rubanetra.protocol.activity.AbstractReplaceableActivity;
import org.jnetpcap.packet.PcapPacket;
import org.jnetpcap.protocol.network.Icmp;
import org.jnetpcap.protocol.network.Icmp.IcmpCode;
import org.jnetpcap.protocol.network.Icmp.IcmpType;
import org.krakenapps.pcap.decoder.icmp.IcmpMessage;
import org.krakenapps.pcap.decoder.icmp.IcmpPacket;

import java.net.InetAddress;

/**
 * Encapsulates an {@link org.krakenapps.pcap.decoder.icmp.IcmpPacket}.
 */
public class Icmpv4Activity extends AbstractReplaceableActivity {
    private static final long serialVersionUID = -2557140540329408425L;
    private final int icmpSubType;
    private final IcmpPacket icmpPacket;
    private final String icmpMessage;
    private final PcapActivity pcapActivity;

    private final IcmpType icmpType;
    private final IcmpCode icmpCode;

    private final InetAddress sourceAddress;
    private final InetAddress destinationAddress;
    private final int identifier;
    private final int sequence;
    private final Icmp icmp;

    /**
     * Constructs a new ICMP activity (IPv4 specific) and replaces the source Pcap activity on which this activity occurred
     *
     * @param source     the Pcap activity on which this event occurred
     * @param icmpPacket the Kraken specific representation of an ICMP packet
     */
    public Icmpv4Activity(PcapActivity source, IcmpPacket icmpPacket) {
        super(source);

        icmpSubType = icmpPacket.getCode();
        icmpMessage = IcmpMessage.getMessage(icmpPacket.getType(), icmpSubType);
        sourceAddress = icmpPacket.getSource();
        destinationAddress = icmpPacket.getDestination();
        identifier = icmpPacket.getId();
        sequence = icmpPacket.getSeq();
        this.icmpPacket = icmpPacket;
        pcapActivity = source;
        icmpCode = IcmpCode.valueOf(icmpPacket.getType(), icmpSubType);
        icmpType = IcmpType.valueOf(icmpPacket.getType());

        icmp = new Icmp();

        scanAndFillIcmp();
        replaceActivity(source);
        setExcludedFromOutput(true);
    }

    private void scanAndFillIcmp() {
        PcapPacket packet = pcapActivity.getPcapPacket();
        packet.hasHeader(getIcmp());
    }

    /**
     * @return The Kraken specific representation including content buffer
     */
    public IcmpPacket getIcmpPacket() {
        return icmpPacket;
    }

    /**
     * @return the ICMP message String unaltered
     */
    public String getIcmpMessage() {
        return icmpMessage;
    }

    /**
     * @return the source on which this activity occurred or the final PcapActivity that led to the fully
     * decoded packet.
     */
    public PcapActivity getPcapActivity() {
        return pcapActivity;
    }

    /**
     * @return the JNetPcap specific enumeration for the ICMP code
     */
    public IcmpCode getIcmpCode() {
        return icmpCode;
    }

    /**
     * @return the JNetPcap specific ICMP Type
     */
    public IcmpType getIcmpType() {
        return icmpType;
    }

    /**
     * @return the ICMP Sub-Type or Code as encountered
     */
    public int getIcmpSubType() {
        return icmpSubType;
    }

    /**
     * @return L3/L4 source address
     */
    public InetAddress getSourceAddress() {
        return sourceAddress;
    }

    /**
     * @return L3/L4 destination address
     */
    public InetAddress getDestinationAddress() {
        return destinationAddress;
    }

    /**
     * @return the message identifier as is
     */
    public int getIdentifier() {
        return identifier;
    }

    /**
     * @return the sequence number as is
     */
    public int getSequence() {
        return sequence;
    }

    /**
     * @return the JNetPcap specific ICMP representation if available
     */
    public Icmp getIcmp() {
        return icmp;
    }

    @Override
    public String getSourceAddressAsString() {
        return String.valueOf(getSourceAddress());
    }

    @Override
    public String getDestinationAddressAsString() {
        return String.valueOf(getDestinationAddress());
    }

    @Override
    public String toString() {
        return "Icmpv4Activity{" +
                "icmpSubType=" + icmpSubType +
                ", icmpPacket=" + icmpPacket +
                ", icmpMessage='" + icmpMessage + '\'' +
                ", pcapActivity=" + pcapActivity +
                ", icmpType=" + icmpType +
                ", icmpCode=" + icmpCode +
                ", sourceAddress=" + sourceAddress +
                ", destinationAddress=" + destinationAddress +
                ", identifier=" + identifier +
                ", sequence=" + sequence +
                ", icmp=" + icmp +
                '}';
    }
}
