/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.protocol.activity.icmp;


import at.jku.fim.rubanetra.pcap.PcapActivity;
import at.jku.fim.rubanetra.protocol.activity.AbstractReplaceableActivity;
import org.jnetpcap.packet.PcapPacket;
import org.jnetpcap.protocol.network.Icmp;
import org.krakenapps.pcap.decoder.icmpv6.Icmpv6Message;
import org.krakenapps.pcap.decoder.icmpv6.Icmpv6Packet;

import java.net.InetAddress;

/**
 * IPv6 specific class similar to {@link at.jku.fim.rubanetra.protocol.activity.icmp.Icmpv4Activity},
 * but encapsulates an {@link org.krakenapps.pcap.decoder.icmpv6.Icmpv6Packet} instead.
 */
public class Icmpv6Activity extends AbstractReplaceableActivity {
    private static final long serialVersionUID = 5042112889089462354L;
    private final int icmpType;
    private final int icmpSubType;
    private final Icmpv6Packet icmpPacket;
    private final String icmpMessage;
    private final PcapActivity pcapActivity;
    private final Icmp icmp;
    private final InetAddress destinationAddress;
    private final InetAddress sourceAddress;

    /**
     * Constructs a new ICMPv6 activity, replacing the source Pcap activity and providing a JNetPcap representation of
     * the packet.
     * @param source the event that led to the decoded icmpPacket
     * @param icmpPacket the ICMP message decoded by Kraken
     */
    public Icmpv6Activity(PcapActivity source, Icmpv6Packet icmpPacket) {
        super(source);
        icmpType = icmpPacket.getType();
        icmpSubType = icmpPacket.getCode();
        icmpMessage = Icmpv6Message.getMessage(icmpType);
        sourceAddress = icmpPacket.getSource();
        destinationAddress = icmpPacket.getDestination();
        this.icmpPacket = icmpPacket;
        pcapActivity = source;

        icmp = new Icmp();

        scanAndFillIcmp();
        replaceActivity(source);
    }

    private void scanAndFillIcmp() {
        PcapPacket packet = pcapActivity.getPcapPacket();
        packet.hasHeader(getIcmp());

    }

    /**
     * @return the type of the ICMP message
     */
    public int getIcmpType() {
        return icmpType;
    }

    /**
     * @return the subtype of the ICMP message
     */
    public int getIcmpSubType() {
        return icmpSubType;
    }

    /**
     * @return the related Kraken object, as passed to constructor
     */
    public Icmpv6Packet getIcmpPacket() {
        return icmpPacket;
    }

    /**
     * @return a String representation for various ICMP messages
     * @see org.krakenapps.pcap.decoder.icmpv6.Icmpv6Message
     */
    public String getIcmpMessage() {
        return icmpMessage;
    }

    /**
     * @return the source on which this activity occurred or the final PcapActivity that led to the fully
     * decoded packet.
     */
    public PcapActivity getPcapActivity() {
        return pcapActivity;
    }

    /**
     * @return JNetPcap's representation of an ICMP packet
     */
    public Icmp getIcmp() {
        return icmp;
    }

    /**
     * @return L3/L4 source address
     */
    public InetAddress getSourceAddress() {
        return sourceAddress;
    }

    /**
     * @return L3/L4 destination address
     */
    public InetAddress getDestinationAddress() {
        return destinationAddress;
    }

    @Override
    public String getSourceAddressAsString() {
        return String.valueOf(getSourceAddress());
    }

    @Override
    public String getDestinationAddressAsString() {
        return String.valueOf(getDestinationAddress());
    }

    @Override
    public String toString() {
        return "Icmpv6Activity{" +
                "icmpType=" + icmpType +
                ", icmpSubType=" + icmpSubType +
                ", icmpPacket=" + icmpPacket +
                ", icmpMessage='" + icmpMessage + '\'' +
                ", pcapActivity=" + pcapActivity +
                ", icmp=" + icmp +
                ", destinationAddress=" + destinationAddress +
                ", sourceAddress=" + sourceAddress +
                '}';
    }
}
