/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.protocol.activity.ip;


import at.jku.fim.rubanetra.pcap.PcapActivity;
import at.jku.fim.rubanetra.protocol.activity.AbstractReplaceableActivity;
import org.kie.api.definition.type.Role;
import org.kie.api.definition.type.Role.Type;
import org.krakenapps.pcap.decoder.ip.Ipv4Packet;
import org.krakenapps.pcap.decoder.ipv6.Ipv6Packet;

import java.net.InetAddress;

/**
 * This class represents the abstraction of {@link at.jku.fim.rubanetra.protocol.activity.ip.Ipv4Activity}
 * and {@link at.jku.fim.rubanetra.protocol.activity.ip.Ipv6Activity}. Several common attributes
 * are available through instances of this class, in particular:
 * <ul>
 * <li>the protocol version</li>
 * <li>the protocol type</li>
 * <li>the source L3 address</li>
 * <li>the destination L3 address</li>
 * <li>the source Pcap activity from which this IP packet was derived from</li>
 * </ul>
 */
@Role(Type.EVENT)
public abstract class IpActivity extends AbstractReplaceableActivity {
    private static final long serialVersionUID = 8337666572553550945L;
    private final int version;
    private final int protocol;
    private final InetAddress sourceAddress, destinationAddress;
    private final PcapActivity pcapActivity;

    /**
     * IPv4 constructor to be called by subclasses. The source activity will be replaced and this instance will be
     * excluded from output.
     *
     * @param source     the source event on which the IPv4 packet was found
     * @param ipv4Packet the decoded IPv4 packet (Kraken)
     */
    public IpActivity(PcapActivity source, Ipv4Packet ipv4Packet) {
        super(source);

        this.version = ipv4Packet.getVersion();
        this.protocol = ipv4Packet.getProtocol();
        this.sourceAddress = ipv4Packet.getSourceAddress();
        this.destinationAddress = ipv4Packet.getDestinationAddress();

        this.pcapActivity = source;
        replaceActivity(source);
        setExcludedFromOutput(true);
    }

    /**
     * IPv6 constructor to be called by subclasses. The source activity will be replaced and this instance will be
     * excluded from output.
     *
     * @param source     the source event on which the IPv6 packet was found
     * @param ipv6Packet the decoded IPv6 packet (Kraken)
     */
    public IpActivity(PcapActivity source, Ipv6Packet ipv6Packet) {
        super(source);

        this.version = ipv6Packet.getVersion();
        protocol = ipv6Packet.getNextHeader();
        sourceAddress = ipv6Packet.getSourceAddress();
        destinationAddress = ipv6Packet.getDestinationAddress();
        this.pcapActivity = source;
        replaceActivity(source);
        setExcludedFromOutput(true);
    }

    /**
     * @return the internet protocol version
     */
    public int getVersion() {
        return version;
    }

    /**
     * @return the protocol type (IPv4) or the next header value (IPv6)
     */
    public int getProtocol() {
        return protocol;
    }

    /**
     * @return the IPv4/IPv6 source address
     */
    public InetAddress getSourceAddress() {
        return sourceAddress;
    }

    /**
     * @return the IPv4/IPv6 destination address
     */
    public InetAddress getDestinationAddress() {
        return destinationAddress;
    }

    /**
     * @return the source on which this activity occurred or the final PcapActivity that led to the fully
     * decoded packet.
     */
    public PcapActivity getPcapActivity() {
        return pcapActivity;
    }

    @Override
    public String getSourceAddressAsString() {
        return String.valueOf(getSourceAddress());
    }

    @Override
    public String getDestinationAddressAsString() {
        return String.valueOf(getDestinationAddress());
    }

    @Override
    public String toString() {
        return "IpActivity{" +
                "version=" + version +
                ", protocol=" + protocol +
                ", sourceAddress=" + sourceAddress +
                ", destinationAddress=" + destinationAddress +
                ", pcapActivity=" + pcapActivity +
                '}';
    }
}
