/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.protocol.activity.ip;


import at.jku.fim.rubanetra.pcap.PcapActivity;
import org.jnetpcap.packet.PcapPacket;
import org.jnetpcap.protocol.network.Ip4;
import org.kie.api.definition.type.Role;
import org.kie.api.definition.type.Role.Type;
import org.krakenapps.pcap.decoder.ip.Ipv4Packet;

import java.util.Arrays;

/**
 * Encapsulates all available/decoded IPv4 packet fields and provides a compatible JNetPcap
 * {@link org.jnetpcap.protocol.network.Ip4} object.
 * Instances of this class are excluded from output by default in order to decrease verbosity,
 * {@link #setExcludedFromOutput(boolean)} may be called to change this behaviour.
 */
@Role(Type.EVENT)
public class Ipv4Activity extends IpActivity {
    private static final long serialVersionUID = -7231432195268189132L;
    private final int internetHeaderLength;
    private final int differentiatedServicesCodePoint;
    private final int totalLength;
    private final int identification;
    private final int flags;
    private final int fragmentOffset;
    private final int timeToLive;
    private final int headerChecksum;
    private final byte[] options;

    private final Ip4 ipv4;

    /**
     * Constructs a new IPv4 activity that will not be included in the final output. The source pcap activity will be
     * replaced and the packet fields encapsulated.
     * @param source the Pcap activity from which this packet was decoded (final one, if fragmented).
     * @param ipv4Packet the decoded IPv4 packet (Kraken)
     */
    public Ipv4Activity(PcapActivity source, Ipv4Packet ipv4Packet) {
        super(source, ipv4Packet);

        this.internetHeaderLength = ipv4Packet.getIhl();
        this.differentiatedServicesCodePoint = ipv4Packet.getTos();
        this.totalLength = ipv4Packet.getTotalLength();
        this.identification = ipv4Packet.getId();
        this.flags = ipv4Packet.getFlags();
        this.fragmentOffset = ipv4Packet.getFragmentOffset();
        this.timeToLive = ipv4Packet.getTtl();
        this.headerChecksum = ipv4Packet.getHeaderChecksum();
        this.options = ipv4Packet.getOptions();

        PcapPacket pcap = source.getPcapPacket();

        ipv4 = new Ip4();
        pcap.hasHeader(ipv4);
        replaceActivity(source);
        setExcludedFromOutput(true);
    }

    /**
     * @throws java.lang.IllegalStateException since it is not yet implemented
     */
    public int getExplicitCongestionNotification() {
        throw new IllegalStateException("Not yet implemented");
    }

    /**
     * @return the length of the IPv4 packet header in bytes
     */
    public int getInternetHeaderLength() {
        return internetHeaderLength;
    }

    /**
     * @return the differentiated services code point value
     */
    public int getDifferentiatedServicesCodePoint() {
        return differentiatedServicesCodePoint;
    }

    /**
     * @return the total packet size in bytes
     */
    public int getTotalLength() {
        return totalLength;
    }

    /**
     * @return the identification value
     */
    public int getIdentification() {
        return identification;
    }

    /**
     * @return the flags combined in one integer
     */
    public int getFlags() {
        return flags;
    }

    /**
     * @return the fragment offset value
     */
    public int getFragmentOffset() {
        return fragmentOffset;
    }

    /**
     * @return the TTL value
     */
    public int getTimeToLive() {
        return timeToLive;
    }

    /**
     * @return the checksum
     */
    public int getHeaderChecksum() {
        return headerChecksum;
    }

    /**
     * @return the options as encountered
     */
    public byte[] getOptions() {
        return options;
    }

    /**
     * @return the JNetPcap representation of the IPv4 packet iff available, null otherwise (e.g. if the packet is
     * fragmented)
     */
    public Ip4 getIpv4() {
        return ipv4;
    }

    @Override
    public String toString() {
        return "Ipv4Activity{" +
                "internetHeaderLength=" + internetHeaderLength +
                ", differentiatedServicesCodePoint=" + differentiatedServicesCodePoint +
                ", totalLength=" + totalLength +
                ", identification=" + identification +
                ", flags=" + flags +
                ", fragmentOffset=" + fragmentOffset +
                ", timeToLive=" + timeToLive +
                ", headerChecksum=" + headerChecksum +
                ", options=" + Arrays.toString(options) +
                ", ipv4=" + ipv4 +
                '}';
    }
}
