/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.protocol.activity.ip;


import at.jku.fim.rubanetra.pcap.PcapActivity;
import org.jnetpcap.packet.PcapPacket;
import org.jnetpcap.protocol.network.Ip6;
import org.kie.api.definition.type.Role;
import org.kie.api.definition.type.Role.Type;
import org.krakenapps.pcap.decoder.ipv6.Ipv6Packet;

/**
 * Encapsulates certain IPv6 packet fields, and provides a compatible JNetPcap representation
 * ({@link org.jnetpcap.protocol.network.Ip6}) of the packet.
 * By default, instances of this class will be excluded from the final output due to verbosity,
 * {@link #setExcludedFromOutput(boolean)} may be called to change this behaviour.
 */
@Role(Type.EVENT)
public class Ipv6Activity extends IpActivity {
    private static final long serialVersionUID = 6406883861304202335L;
    private final byte trafficClass;
    private final int flowLabel;
    private final int payloadLength;
    private final byte nextHeader;
    private final int hopLimit;

    private final Ip6 ipv6;
    private final Ipv6Packet ipv6Packet;

    /**
     * Construct a new IPv6 activity and replaces the source Pcap activity on which this activity occurred,
     *
     * @param source     the Pcap activity on which this event occurred.
     * @param ipv6Packet the decoded IPv6 packet (Kraken)
     */
    public Ipv6Activity(PcapActivity source, Ipv6Packet ipv6Packet) {
        super(source, ipv6Packet);
        this.ipv6Packet = ipv6Packet;

        trafficClass = ipv6Packet.getTrafficClass();
        flowLabel = ipv6Packet.getFlowLabel();
        payloadLength = ipv6Packet.getPayloadLength();
        nextHeader = ipv6Packet.getNextHeader();
        hopLimit = ipv6Packet.getHopLimit();

        PcapPacket pcap = source.getPcapPacket();
        ipv6 = new Ip6();
        pcap.hasHeader(ipv6);
        replaceActivity(source);
        setExcludedFromOutput(true);
    }

    /**
     * @return the traffic class value
     */
    public byte getTrafficClass() {
        return trafficClass;
    }

    /**
     * @return the flow label value
     */
    public int getFlowLabel() {
        return flowLabel;
    }

    /**
     * @return the size of the payload in bytes
     */
    public int getPayloadLength() {
        return payloadLength;
    }

    /**
     * @return the next header value, redundant to {@link #getProtocol()}
     */
    public byte getNextHeader() {
        return nextHeader;
    }

    /**
     * @return the Hop limit value
     */
    public int getHopLimit() {
        return hopLimit;
    }

    /**
     * @return the JNetPcap representation of an IPv6 packet, if available. null otherwise
     */
    public Ip6 getIpv6() {
        return ipv6;
    }

    /**
     * @return the original Kraken IPv6 packet representation
     */
    public Ipv6Packet getIpv6Packet() {
        return ipv6Packet;
    }

    @Override
    public String toString() {
        return "Ipv6Activity{" +
                "trafficClass=" + trafficClass +
                ", flowLabel=" + flowLabel +
                ", payloadLength=" + payloadLength +
                ", nextHeader=" + nextHeader +
                ", hopLimit=" + hopLimit +
                ", ipv6=" + ipv6 +
                ", ipv6Packet=" + ipv6Packet +
                '}';
    }
}
