/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.protocol.activity.skype;


import org.jnetpcap.protocol.tcpip.Udp;

/**
 * This class does not represent an activity but provides some useful static helper methods. Particularly it provides
 * quite basic byte-based checks whether or not a certain UDP-Payload could indicate the presence of Skype related
 * traffic.
 * Additionally, the Skype object ID can be extracted as well.
 * Those checks are based on the OpenSkype project, see https://github.com/matthiasbock/OpenSkype
 */
public class SkypeActivityHelper {
    /**
     * Returns the Skype object ID, however, since the object ID could be arbitrary this method
     * does not check whether the payload corresponds to Skype traffic.
     *
     * @param udp the JNetPcap UDP packet representation
     * @return the skype object id
     * @see #hasSkypePayload(org.jnetpcap.protocol.tcpip.Udp)
     */
    public static int objectId(Udp udp) {
        if (udp != null && udp.hasPayload() && udp.getPayloadLength() > 3) {
            return udp.getPacket().getUShort(udp.getPayloadOffset());
        }
        return -1;
    }

    /**
     * Calls {@link #hasSkypeType(org.jnetpcap.protocol.tcpip.Udp, at.jku.fim.rubanetra.protocol.activity.skype.SkypeActivityHelper.SkypeUdpType)}
     * with the PAYLOAD flag internally (convenience method).
     *
     * @param udp the JNetPcap UDP packet representation
     * @return true, if this UDP packet could contain Skype related traffic
     */
    public static boolean hasSkypePayload(Udp udp) {
        return hasSkypeType(udp, SkypeUdpType.PAYLOAD);
    }

    /**
     * Tries to determine whether or not the provided UDP packet contains data that corresponds to some known
     * Skype flags. Note however, that these flags may be invalid at any future point in time.
     *
     * @param udp  the JNetPcap UDP packet representation
     * @param type the type of the data to check against
     * @return true, if udp potentially contains data corresponding to type
     */
    public static boolean hasSkypeType(Udp udp, SkypeUdpType type) {
        if (udp != null && type != null && udp.hasPayload() && udp.getPayloadLength() > 3) {
            int thirdByte = udp.getPacket().getUByte(udp.getPayloadOffset() + 2);
            return (thirdByte & 0x0F) == type.getVal();
        }
        return false;
    }

    /**
     * Several known Skype payload types including their flag value,
     * based on the OpenSkype project, see https://github.com/matthiasbock/OpenSkype
     */
    public enum SkypeUdpType {
        PAYLOAD(0x02), RESEND(0x03), CONFIRM(0x5), ERROR(0x7), AUDIO(0xd), FRAGMENT(0xf);

        private final int val;

        SkypeUdpType(int val) {
            this.val = val;
        }

        public int getVal() {
            return val;
        }
    }
}
