/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.protocol.activity.tcp;


import at.jku.fim.rubanetra.pcap.PcapActivity;
import at.jku.fim.rubanetra.protocol.activity.AbstractReplaceableActivity;
import at.jku.fim.rubanetra.protocol.activity.ip.IpActivity;
import com.fasterxml.jackson.annotation.JsonIgnore;
import org.jnetpcap.packet.PcapPacket;
import org.jnetpcap.packet.format.FormatUtils;
import org.jnetpcap.protocol.tcpip.Tcp;
import org.kie.api.definition.type.Role;
import org.krakenapps.pcap.decoder.tcp.TcpDirection;
import org.krakenapps.pcap.decoder.tcp.TcpPacket;
import org.krakenapps.pcap.decoder.tcp.TcpSession;
import org.krakenapps.pcap.decoder.tcp.TcpState;

import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.util.Arrays;

/**
 * Encapsulates a vast amount of TCP flags/options/values and derived information, however, some fields have not yet
 * been implemented by the Kraken TCP-Decoder and/or the JNetPcap-decoder. In these cases either a value of -1 or null
 * is returned instead, e.g. for relative seq/ack numbers or the client/server state or the traffic direction.
 * Two constructors are provided, one is intended for use by decoded Kraken TCP packets {@link org.krakenapps.pcap.decoder.tcp.TcpPacket}
 * the other one for the JNetPcap equivalent {@link org.jnetpcap.protocol.tcpip.Tcp}.
 * However, independent of the used constructor the JNetPcap representation will always be made available if possible.
 * Instances of this class will not be included in the final output, {@link #setExcludedFromOutput(boolean)} should be
 * called appropriately to override this behaviour.
 */
@Role(Role.Type.EVENT)
public class TcpActivity extends AbstractReplaceableActivity {
    private static final long serialVersionUID = -2422097413536740312L;
    private final int sourcePort, destinationPort;
    private final InetAddress sourceAddress, destinationAddress;
    private final long sequenceNumber;
    private final long acknowledgeNumber;
    private final int relativeSequenceNumber;
    private final int relativeAcknowledgeNumber;
    private final int dataOffset;
    private final int controlBits;
    private final int windowSize, checksum, urgentPointer, tcpLength;
    private final byte[] options, padding;
    private final boolean syn, ack, psh, fin, rst, urg;
    private final TcpDirection direction;
    private final TcpState clientState, serverState;

    private final PcapActivity pcapActivity;
    private final Tcp tcp;
    private final InetSocketAddress sourceSocketAddress, destinationSocketAddress;

    /**
     * Constructs a new Tcp activity using the decoded Kraken data. The source pcap activity on which this TCP-activity
     * occurred will be replaced.
     *
     * @param source     the final pacp activity that led to a successful decode by the Kraken TCP decoder.
     * @param tcpPacket  the decoded TCP packet data
     * @param tcpSession the maintained TCP session information (client & server state)
     */
    public TcpActivity(PcapActivity source, TcpPacket tcpPacket, TcpSession tcpSession) {
        super(source);
        this.sourcePort = tcpPacket.getSourcePort();
        this.destinationPort = tcpPacket.getDestinationPort();
        this.sourceAddress = tcpPacket.getSourceAddress();
        this.sourceSocketAddress = tcpPacket.getSource();
        this.destinationAddress = tcpPacket.getDestinationAddress();
        this.destinationSocketAddress = tcpPacket.getDestination();
        this.sequenceNumber = tcpPacket.getSeq();
        this.acknowledgeNumber = tcpPacket.getAck();
        this.relativeSequenceNumber = tcpPacket.getRelativeSeq();
        this.relativeAcknowledgeNumber = tcpPacket.getRelativeAck();
        this.dataOffset = tcpPacket.getDataOffset();
        this.controlBits = tcpPacket.getFlags();
        this.windowSize = tcpPacket.getWindow();
        this.checksum = tcpPacket.getChecksum();
        this.urgentPointer = tcpPacket.getUrgentPointer();
        this.options = tcpPacket.getOptions();
        this.padding = tcpPacket.getPadding();
        this.tcpLength = tcpPacket.getTotalLength();
        this.direction = tcpPacket.getDirection();
        this.syn = tcpPacket.isSyn();
        this.ack = tcpPacket.isAck();
        this.psh = tcpPacket.isPsh();
        this.fin = tcpPacket.isFin();
        this.rst = tcpPacket.isRst();
        this.urg = tcpPacket.isUrg();
        this.clientState = tcpSession.getClientState();
        this.serverState = tcpSession.getServerState();
        this.pcapActivity = source;

        this.tcp = new Tcp();
        PcapPacket packet = pcapActivity.getPcapPacket();
        packet.hasHeader(tcp);

        replaceActivity(source);
        setExcludedFromOutput(true);
    }

    public TcpActivity(PcapActivity source, Tcp tcp, IpActivity ip) {
        super(source);
        this.sourcePort = tcp.source();
        this.destinationPort = tcp.destination();
        this.sourceAddress = ip.getSourceAddress();
        this.sourceSocketAddress = new InetSocketAddress(sourceAddress, sourcePort);
        this.destinationAddress = ip.getDestinationAddress();
        this.destinationSocketAddress = new InetSocketAddress(destinationAddress, destinationPort);
        this.sequenceNumber = tcp.seq();
        this.acknowledgeNumber = tcp.ack();
        this.relativeSequenceNumber = -1; // TODO: not yet implemented
        this.relativeAcknowledgeNumber = -1; //TODO: not yet implemented
        this.dataOffset = tcp.getOffset();
        this.controlBits = tcp.flags();
        this.windowSize = tcp.window();
        this.checksum = tcp.checksum();
        this.urgentPointer = tcp.urgent();
        this.tcpLength = tcp.getLength();
        this.options = null; // TODO: not yet implemented
        this.padding = null; // TODO: not yet implemented
        this.syn = tcp.flags_SYN();
        this.ack = tcp.flags_ACK();
        this.psh = tcp.flags_PSH();
        this.fin = tcp.flags_FIN();
        this.rst = tcp.flags_RST();
        this.urg = tcp.flags_URG();
        this.direction = null; // TODO: not yet implemented
        this.clientState = null; // TODO: not yet implemented
        this.serverState = null; // TODO: not yet implemented
        this.pcapActivity = source;
        this.tcp = tcp;

        replaceActivity(source);
        setExcludedFromOutput(true);
    }

    /**
     * @return the L4 source address information
     * @see #getSourceSocketAddress()
     */
    public int getSourcePort() {
        return sourcePort;
    }

    /**
     * @return the L4 destination address information
     * @see #getDestinationSocketAddress()
     */
    public int getDestinationPort() {
        return destinationPort;
    }

    /**
     * @return the L3 source address information
     * @see #getSourceSocketAddress()
     */
    public InetAddress getSourceAddress() {
        return sourceAddress;
    }

    /**
     * @return the L3/L4 source address information
     */
    public InetSocketAddress getSourceSocketAddress() {
        return sourceSocketAddress;
    }

    /**
     * @return the L3 destination address information
     * @see #getDestinationSocketAddress()
     */
    public InetAddress getDestinationAddress() {
        return destinationAddress;
    }

    /**
     * @return the L3/L4 destination address information
     */
    public InetSocketAddress getDestinationSocketAddress() {
        return destinationSocketAddress;
    }

    /**
     * @return the decoded sequence number
     */
    public long getSequenceNumber() {
        return sequenceNumber;
    }

    /**
     * @return the decoded ACK number
     */
    public long getAcknowledgeNumber() {
        return acknowledgeNumber;
    }

    /**
     * @return the decoded data offset
     */
    public int getDataOffset() {
        return dataOffset;
    }

    /**
     * @return the control flags
     */
    public int getControlBits() {
        return controlBits;
    }

    /**
     * @return the parsed window size
     */
    public int getWindowSize() {
        return windowSize;
    }

    /**
     * @return the parsed checksum
     */
    public int getChecksum() {
        return checksum;
    }

    /**
     * @return the parsed urgent pointer value
     */
    public int getUrgentPointer() {
        return urgentPointer;
    }

    /**
     * @return the content of the TCP length field
     */
    public int getTcpLength() {
        return tcpLength;
    }

    /**
     * @return the options as array or null (if the JNetPcap constructor was used)
     */
    public byte[] getOptions() {
        return options;
    }

    /**
     * @return the padding bytes or null (if the JNetPcap constructor was used)
     */
    public byte[] getPadding() {
        return padding;
    }

    /**
     * @return true, if the syn flag is set
     */
    public boolean isSyn() {
        return syn;
    }

    /**
     * @return true, if the ack flag is set
     */
    public boolean isAck() {
        return ack;
    }

    /**
     * @return true, if the psh flag is set
     */
    public boolean isPsh() {
        return psh;
    }

    /**
     * @return true, if the fin flag is set
     */
    public boolean isFin() {
        return fin;
    }

    /**
     * @return true, if the rst flag is set
     */
    public boolean isRst() {
        return rst;
    }

    /**
     * @return true, if the urg flag is set
     */
    public boolean isUrg() {
        return urg;
    }

    /**
     * @return the direction of this activity, e.g. client->server. may be null if the JNetPcap constructor was used.
     */
    public TcpDirection getDirection() {
        return direction;
    }

    /**
     * @return the client state. may be null if the JNetPcap constructor was used.
     */
    public TcpState getClientState() {
        return this.clientState;
    }

    /**
     * @return the server state. may be null if the JNetPcap constructor was used.
     */
    public TcpState getServerState() {
        return this.serverState;
    }

    /**
     * @return the source on which this activity occurred or the final PcapActivity that led to the fully
     * decoded packet.
     */
    public PcapActivity getPcapActivity() {
        return pcapActivity;
    }

    /**
     * @return the relative sequence number (calculated by the Kraken decoder, may be -1 if the JNetPcap constructor
     * was used)
     */
    public int getRelativeSequenceNumber() {
        return relativeSequenceNumber;
    }

    /**
     * @return the relative ack number (calculated by the Kraken decoder, may be -1 if the JNetPcap constructor
     * was used)
     */
    public int getRelativeAcknowledgeNumber() {
        return relativeAcknowledgeNumber;
    }

    /**
     * @return the JNetPcap representation of a decoded TCP packet or null, if not available
     */
    @JsonIgnore
    public Tcp getTcp() {
        return tcp;
    }

    /**
     * @return the payload as a String, or null if the payload is not available or empty
     */
    public String payloadString() {
        byte[] payload = tcp.getPayload();
        return payload != null && payload.length > 0 ? new String(payload) : null;
    }

    /**
     * @return a formatted Hex dump of the payload or null if the payload is not available or empty
     * @see org.jnetpcap.packet.format.FormatUtils#hexdump(byte[])
     */
    public String payloadHexFormattedDump() {
        byte[] payload = tcp.getPayload();
        return payload != null && payload.length > 0 ? FormatUtils.hexdump(payload) : null;
    }

    /**
     * @return a Hex dump of the payload or null if the payload is not available or empty
     * @see org.jnetpcap.packet.format.FormatUtils#hexdump(byte[])
     */
    public String payloadHexString() {
        byte[] payload = tcp.getPayload();
        return payload != null && payload.length > 0 ? FormatUtils.asString(payload) : null;
    }

    /**
     * ENHANCEMENT: Crude check based on the source activity should be replaced by a more detailed comparison
     * since the decoders may not be deterministic.
     */
    @Override
    public boolean equals(Object obj) {
        return obj instanceof TcpActivity && ((TcpActivity) obj).getPcapActivity().equals(getPcapActivity());
    }

    @Override
    public String getSourceAddressAsString() {
        return String.valueOf(getSourceAddress());
    }

    @Override
    public String getDestinationAddressAsString() {
        return String.valueOf(getDestinationAddress());
    }

    @Override
    public String toString() {
        return "TcpActivity{" +
                "sourcePort=" + sourcePort +
                ", destinationPort=" + destinationPort +
                ", sourceAddress=" + sourceAddress +
                ", destinationAddress=" + destinationAddress +
                ", sequenceNumber=" + sequenceNumber +
                ", acknowledgeNumber=" + acknowledgeNumber +
                ", relativeSequenceNumber=" + relativeSequenceNumber +
                ", relativeAcknowledgeNumber=" + relativeAcknowledgeNumber +
                ", dataOffset=" + dataOffset +
                ", controlBits=" + controlBits +
                ", windowSize=" + windowSize +
                ", checksum=" + checksum +
                ", urgentPointer=" + urgentPointer +
                ", tcpLength=" + tcpLength +
                ", options=" + Arrays.toString(options) +
                ", padding=" + Arrays.toString(padding) +
                ", syn=" + syn +
                ", ack=" + ack +
                ", psh=" + psh +
                ", fin=" + fin +
                ", rst=" + rst +
                ", urg=" + urg +
                ", direction=" + direction +
                ", clientState=" + clientState +
                ", serverState=" + serverState +
                ", pcapActivity=" + pcapActivity +
                ", tcp=" + tcp +
                ", sourceSocketAddress=" + sourceSocketAddress +
                ", destinationSocketAddress=" + destinationSocketAddress +
                '}';
    }
}
