/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.protocol.activity.tls;


import at.jku.fim.rubanetra.protocol.activity.AbstractReplaceableActivity;
import at.jku.fim.rubanetra.protocol.activity.tcp.TcpActivity;
import org.kie.api.definition.type.Role;

import java.util.Collections;
import java.util.SortedSet;
import java.util.TreeSet;

import static org.kie.api.definition.type.Role.Type;

/**
 * Encapsulates various {@link at.jku.fim.rubanetra.protocol.activity.tcp.TcpActivity} objects
 * that are specific to SSL/TLS streams, e.g. typical "Client Hello", "Server Hello" and "ChangeCipherSpec"
 * messages are stored in this activity - therefore it should be considered a derived fact next to constituting an event.
 * {@link at.jku.fim.rubanetra.protocol.activity.tls.TlsActivityHelper} may be used to decide whether
 * or not a TCP payload may contain SSL/TLS protocol messages.
 */
@Role(Type.EVENT)
public class TlsActivity extends AbstractReplaceableActivity {
    private static final long serialVersionUID = -7664259518283613758L;
    private final TcpActivity clientHello;
    private final TcpActivity serverHello;
    private final SortedSet<Long> clientToServerTraffic, serverToClientTraffic;

    private TcpActivity changeCipherSpec;

    /**
     * Constructs a prototypical TLS Event using a "Client Hello" and a "Server Hello" TCP messages.
     * Both messages are extended instead of replaced due to potential wrongful classifications.
     */
    public TlsActivity(TcpActivity clientHello, TcpActivity serverHello) {
        super(clientHello);

        this.clientHello = clientHello;
        this.serverHello = serverHello;

        serverToClientTraffic = new TreeSet<>();
        clientToServerTraffic = new TreeSet<>();

        extendActivity(clientHello);
        extendActivity(serverHello);
    }

    /**
     * @return the tcp activity resembling the change cipher spec protocol message or null if it has not been set
     */
    public TcpActivity getChangeCipherSpec() {
        return changeCipherSpec;
    }

    /**
     * Designates the passed TCP activity to containing a "Change Cipher Spec" protocol message.
     *
     * @param changeCipherSpec the TCP activity resembling a "Change Cipher Spec" (will be extended)
     */
    public void setChangeCipherSpec(TcpActivity changeCipherSpec) {
        if (this.changeCipherSpec != null) {
            removeReplacedActivity(this.changeCipherSpec);
        }
        this.changeCipherSpec = changeCipherSpec;
        extendActivity(changeCipherSpec);
    }

    /**
     * @return the "Client Hello" protocol message containing TcpActivity
     */
    public TcpActivity getClientHello() {
        return clientHello;
    }

    /**
     * @return the "Server Hello" protocol message containing TcpActivity
     */
    public TcpActivity getServerHello() {
        return serverHello;
    }

    /**
     * @return a sorted set containing individual frame numbers that are part of the TCP session (client -> server)
     * ENHANCEMENT: this is quite inefficient, introduction of ranges?
     */
    public SortedSet<Long> getClientToServerTraffic() {
        return Collections.unmodifiableSortedSet(clientToServerTraffic);
    }

    /**
     * @return a sorted set containing individual frame numbers that are part of the TCP session (server -> client)
     * ENHANCEMENT: this is quite inefficient, introduction of ranges?
     */
    public SortedSet<Long> getServerToClientTraffic() {
        return Collections.unmodifiableSortedSet(serverToClientTraffic);
    }

    /**
     * Assigns and extends a tcp activity to be part of this TLS/SSL stream
     *
     * @param tcp the tcp activity of the same TCP session, client to server direction
     */
    public void addClientToServerTcpActivity(TcpActivity tcp) {
        this.clientToServerTraffic.addAll(tcp.getCompoundFrameNumbers());
        extendActivity(tcp);
    }

    /**
     * Assigns and extends a tcp activity to be part of this TLS/SSL stream
     *
     * @param tcp the tcp activity of the same TCP session, server to client direction
     */
    public void addServerToClientTcpActivity(TcpActivity tcp) {
        this.serverToClientTraffic.addAll(tcp.getCompoundFrameNumbers());
        extendActivity(tcp);
    }

    @Override
    public String getSourceAddressAsString() {
        return getClientHello().getSourceAddressAsString();
    }

    @Override
    public String getDestinationAddressAsString() {
        return getClientHello().getDestinationAddressAsString();
    }

    @Override
    public String toString() {
        return "TlsActivity{" +
                "clientHello=" + clientHello +
                ", serverHello=" + serverHello +
                ", clientToServerTraffic=" + clientToServerTraffic +
                ", serverToClientTraffic=" + serverToClientTraffic +
                ", changeCipherSpec=" + changeCipherSpec +
                '}';
    }
}
