/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.protocol.activity.tls;


import org.jnetpcap.protocol.tcpip.Tcp;

/**
 * Contains various static methods as indicators of TLS/SSL protocol messages in
 * {@link at.jku.fim.rubanetra.protocol.activity.tcp.TcpActivity} payloads.
 * These methods are basically unverified and marginally tested - therefore false positives/negatives may be
 * possible.
 */
public class TlsActivityHelper {
    /**
     * Returns true if the TCP payload contains a "Handshake" control flag.
     *
     * @param tcp the JNetPcap TCP representation containing a payload.
     * @return true if the TCP payload contains a "Handshake" control flag.
     */
    public static boolean isTlsHandshake(Tcp tcp) {
        return TlsContentType.HANDSHAKE.matches(tcp, TlsRecord.CONTENT_TYPE)
                || TlsContentType.SSL_HANDSHAKE.matches(tcp, TlsRecord.CONTENT_TYPE);
    }

    /**
     * Returns true if the TCP payload contains a "Client Hello" control message.
     *
     * @param tcp the JNetPcap TCP representation containing a payload.
     * @return true if the TCP payload contains a "Client Hello" control message.
     */
    public static boolean isClientHello(Tcp tcp) {
        if (TlsContentType.HANDSHAKE.matches(tcp, TlsRecord.CONTENT_TYPE)) {
            return TlsContentType.HANDSHAKE_STATE_CLIENT_HELLO.matches(tcp, TlsRecord.HANDSHAKE_STATE);
        } else if (TlsContentType.SSL_HANDSHAKE.matches(tcp, TlsRecord.CONTENT_TYPE)) {
            return TlsContentType.SSL_HANDSHAKE_STATE_CLIENT_HELLO.matches(tcp, TlsRecord.SSL_HANDSHAKE_STATE);
        }

        return false;
    }

    /**
     * Returns true if the TCP payload contains a "Server Hello" control message.
     *
     * @param tcp the JNetPcap TCP representation containing a payload.
     * @return true if the TCP payload contains a "Server Hello" control message.
     */
    public static boolean isServerHello(Tcp tcp) {
        if (TlsContentType.HANDSHAKE.matches(tcp, TlsRecord.CONTENT_TYPE)) {
            return TlsContentType.HANDSHAKE_STATE_SERVER_HELLO.matches(tcp, TlsRecord.HANDSHAKE_STATE);
        } else if (TlsContentType.SSL_HANDSHAKE.matches(tcp, TlsRecord.CONTENT_TYPE)) {
            return TlsContentType.SSL_HANDSHAKE_STATE_SERVER_HELLO.matches(tcp, TlsRecord.SSL_HANDSHAKE_STATE);
        }

        return false;
    }

    /**
     * Returns true if the TCP payload contains a "ChangeCipherSpec" control message.
     *
     * @param tcp the JNetPcap TCP representation containing a payload.
     * @return true if the TCP payload contains a "ChangeCipherSpec" control message.
     */
    public static boolean isChangeCipherSpec(Tcp tcp) {
        return TlsContentType.CHANGE_CIPHER_SPEC.matches(tcp, TlsRecord.CONTENT_TYPE);
    }

    /**
     * Returns true if the TCP payload contains an "Application" control message.
     *
     * @param tcp the JNetPcap TCP representation containing a payload.
     * @return true if the TCP payload contains an "Application" control message.
     */
    public static boolean isApplication(Tcp tcp) {
        return TlsContentType.APPLICATION.matches(tcp, TlsRecord.CONTENT_TYPE);
    }

    /**
     * Enumeration of possible SSL/TLS records including byte offset into the payload and length values.
     */
    public enum TlsRecord {
        CONTENT_TYPE(0, 1),
        VERSION(1, 2),
        LENGTH(3, 2),

        // for handshake
        HANDSHAKE_STATE(5, 1),
        SSL_HANDSHAKE_STATE(2, 1);

        private final int byteOffset;
        private final int length;

        TlsRecord(int byteOffset, int length) {
            this.byteOffset = byteOffset;
            this.length = length;
        }

    }

    /**
     * Enumeration of possible SSL/TLS content types including flag values.
     */
    public enum TlsContentType {
        CHANGE_CIPHER_SPEC(0x14),
        ALERT(0x15),
        HANDSHAKE(0x16),
        APPLICATION(0x17),

        // special purpose handshake protocol
        HANDSHAKE_STATE_CLIENT_HELLO(0x01),
        HANDSHAKE_STATE_SERVER_HELLO(0x02),

        //SSL v2 exception
        SSL_HANDSHAKE(0x80),

        // special purpose handshake protocol
        SSL_HANDSHAKE_STATE_CLIENT_HELLO(0x01),
        SSL_HANDSHAKE_STATE_SERVER_HELLO(0x02);
        private final int val;

        TlsContentType(int val) {
            this.val = val;
        }

        /**
         * Returns true if the payload of tcp contains this flag in its TlsRecord record.
         *
         * @param tcp    the the JNetPcap TCP representation containing a payload.
         * @param record the TLSRecord of interest
         * @return true if the payload of tcp contains this flag in its TlsRecord record.
         */
        public boolean matches(Tcp tcp, TlsRecord record) {
            if (tcp == null || !tcp.hasPayload()) return false;


            if (tcp.getPayloadLength() >= record.byteOffset + record.length) {
                int off = tcp.getPayloadOffset() + record.byteOffset;
                switch (record.length) {
                    case 1:
                        return tcp.getPacket().getUByte(off) == val;
                    case 2:
                        return tcp.getPacket().getUShort(off) == val;
                    default:
                        return false;
                }
            }

            return false;

        }
    }
}
