/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.protocol.activity.udp;


import at.jku.fim.rubanetra.pcap.PcapActivity;
import at.jku.fim.rubanetra.protocol.activity.AbstractReplaceableActivity;
import com.fasterxml.jackson.annotation.JsonIgnore;
import org.jnetpcap.packet.PcapPacket;
import org.jnetpcap.protocol.tcpip.Udp;
import org.kie.api.definition.type.Role;
import org.kie.api.definition.type.Role.Type;
import org.krakenapps.pcap.decoder.udp.UdpPacket;

import java.net.InetSocketAddress;

/**
 * Encapsulates various UDP packet specific values (decoded by Kraken) and
 * provides a JNetPcap representation of the same UDP data.
 */
@Role(Type.EVENT)
public class UdpActivity extends AbstractReplaceableActivity {
    private static final long serialVersionUID = 351329488071596679L;
    private final int sourcePort;
    private final int destinationPort;
    private final int length;
    private final int checksum;
    private final PcapActivity pcapActivity;
    private final Udp udp;
    private final InetSocketAddress sourceSocketAddress;
    private final InetSocketAddress destinationSocketAddress;

    /**
     * Constructs a new Udp activity, replaces the source pcap activity on which this event occurred and excludes it
     * from the final output due to verbosity.
     *
     * @param source the pcap activity on which this activity occurred
     * @param packet the decoded UDP data (Kraken)
     */
    public UdpActivity(PcapActivity source, UdpPacket packet) {
        super(source);

        sourcePort = packet.getSourcePort();
        destinationPort = packet.getDestinationPort();
        length = packet.getLength();
        checksum = packet.getChecksum();
        sourceSocketAddress = packet.getSource();
        destinationSocketAddress = packet.getDestination();

        pcapActivity = source;

        this.udp = new Udp();

        PcapPacket p = pcapActivity.getPcapPacket();
        p.hasHeader(udp);
        replaceActivity(source);
        setExcludedFromOutput(true);
    }

    /**
     * @return the JNetPcap representation of the decoded UDP data
     */
    @JsonIgnore
    public Udp getUdp() {
        return udp;
    }

    /**
     * @return the L4 source address information
     */
    public int getSourcePort() {
        return sourcePort;
    }

    /**
     * @return the L4 destination address information
     */
    public int getDestinationPort() {
        return destinationPort;
    }

    /**
     * @return the L3/L4 source address
     */
    public InetSocketAddress getSourceSocketAddress() {
        return sourceSocketAddress;
    }

    /**
     * @return the L3/L4 destination address
     */
    public InetSocketAddress getDestinationSocketAddress() {
        return destinationSocketAddress;
    }

    /**
     * @return payload length in bytes
     */
    public int getLength() {
        return length;
    }

    /**
     * @return packet checksum as encountered
     */
    public int getChecksum() {
        return checksum;
    }

    /**
     * @return the source on which this activity occurred or the final PcapActivity that led to the fully
     * decoded packet.
     */
    public PcapActivity getPcapActivity() {
        return pcapActivity;
    }

    @Override
    public String getSourceAddressAsString() {
        return String.valueOf(getSourceSocketAddress());
    }

    @Override
    public String getDestinationAddressAsString() {
        return String.valueOf(getDestinationSocketAddress());
    }

    @Override
    public String toString() {
        return "UdpActivity{" +
                "sourcePort=" + sourcePort +
                ", destinationPort=" + destinationPort +
                ", length=" + length +
                ", checksum=" + checksum +
                ", pcapActivity=" + pcapActivity +
                ", udp=" + udp +
                ", sourceSocketAddress=" + sourceSocketAddress +
                ", destinationSocketAddress=" + destinationSocketAddress +
                '}';
    }
}
