/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.protocol.handler;


import at.jku.fim.rubanetra.config.model.KrakenProtocolConfiguration;
import at.jku.fim.rubanetra.config.model.ProtocolId;
import at.jku.fim.rubanetra.protocol.KrakenBaseProtocol;
import at.jku.fim.rubanetra.protocol.activity.arp.ArpActivity;
import org.krakenapps.pcap.decoder.arp.ArpDecoder;
import org.krakenapps.pcap.decoder.arp.ArpPacket;

/**
 * Represents a link to the Kraken library, i.e. it receives decoded ARP data, wraps it into suitable
 * activities ({@link at.jku.fim.rubanetra.protocol.activity.arp.ArpActivity}) and
 * passes it on to interested objects.
 * The Kraken library expects an Ethernet-Decoder that transfers decoded frames to an
 * {@link org.krakenapps.pcap.decoder.arp.ArpDecoder}, therefore this Ethernet decoder must be bound from
 * a {@link at.jku.fim.rubanetra.protocol.handler.KrakenEthernetProtocolHandler} to this Arp Decoder
 * ({@link #getKrakenArpDecoder()}).
 * This class does not provide a binding method implementation, i.e. it is not possible to bind this handler
 * to another one.
 */
public class KrakenArpHandler extends KrakenBaseProtocol {
    public static final String NEW_ARP_ACTIVITY_PROPERTY_NAME = "newArpActivity";
    public static final ProtocolId krakenArpProtocolId = new ProtocolId() {
        private final String id = "Arp";

        @Override
        public String getProtocolId() {
            return id;
        }

        @Override
        public String getName() {
            return id;
        }

        @Override
        public String toString() {
            return getName();
        }
    };

    private final ArpDecoder krakenArpDecoder;

    /**
     * Constructs a new Arp handler by creating an {@link org.krakenapps.pcap.decoder.arp.ArpDecoder}, a
     * {@link at.jku.fim.rubanetra.protocol.handler.KrakenEthernetProtocolHandler} and binding to it.
     */
    public KrakenArpHandler() {
        super();

        krakenArpDecoder = new ArpDecoder();
        krakenArpDecoder.register(this::fireNewArpActivity);
    }

    @Override
    public ProtocolId getProtocolId() {
        return krakenArpProtocolId;
    }

    @Override
    public void initialize(KrakenProtocolConfiguration setting) {
        super.initialize(setting);
    }

    private void fireNewArpActivity(ArpPacket p) {
        ArpActivity act = new ArpActivity(getCurrentPcapActivity(), p);
        propertyChangeSupport.firePropertyChange(NEW_ARP_ACTIVITY_PROPERTY_NAME, null, act);
    }

    public ArpDecoder getKrakenArpDecoder() {
        return krakenArpDecoder;
    }

}
