/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.protocol.handler;


import at.jku.fim.rubanetra.config.model.KrakenApplicationProtocolId;
import at.jku.fim.rubanetra.config.model.ProtocolId;
import at.jku.fim.rubanetra.protocol.KrakenBaseProtocol;
import at.jku.fim.rubanetra.protocol.activity.dns.DnsActivity;
import org.krakenapps.pcap.Protocol;
import org.krakenapps.pcap.decoder.tcp.TcpProcessor;
import org.krakenapps.pcap.decoder.tcp.TcpSessionKey;
import org.krakenapps.pcap.decoder.udp.UdpPacket;
import org.krakenapps.pcap.decoder.udp.UdpProcessor;
import org.krakenapps.pcap.util.Buffer;
import org.krakenapps.pcap.util.ChainBuffer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.xbill.DNS.Message;

import java.io.IOException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.util.HashMap;
import java.util.Map;

/**
 * This class does not utilize a Kraken decoder directly, instead it represents the decoding processors as
 * an implementation of {@link org.krakenapps.pcap.decoder.udp.UdpProcessor} and
 * {@link org.krakenapps.pcap.decoder.tcp.TcpProcessor}.
 * The org.xbill.dns Java library (see {@link org.xbill.DNS.Message}) is used as primary payload decoder.
 * Several modifications were required to process both, UDP- and TCP-based payloads, therefore this implementation
 * is still considered experimental and may (silently) fail to decode valid DNS messages.
 */
public class KrakenDnsProtocolHandler extends KrakenBaseProtocol implements UdpProcessor, TcpProcessor {
    public static final String NEW_DNS_ACTIVITY_PROPERTY_NAME = "newDnsActivity";
    public static final KrakenApplicationProtocolId krakenDnsProtocolId = new KrakenApplicationProtocolId() {
        private final String protocolId = Protocol.DNS.name();

        @Override
        public Protocol getKrakenApplicationProtocol() {
            return Protocol.DNS;
        }

        @Override
        public String getProtocolId() {
            return protocolId;
        }

        @Override
        public String getName() {
            return "Dns";
        }

        @Override
        public String toString() {
            return getName();
        }
    };
    private final Map<TcpSessionKey, DnsSession> sessionMap = new HashMap<>();
    private final Logger log = LoggerFactory.getLogger(getClass());

    @Override
    public ProtocolId getProtocolId() {
        return krakenDnsProtocolId;
    }

    @Override
    public void process(UdpPacket p) {
        Buffer udpData = p.getData();
        int readableBytes = udpData.readableBytes();
        byte[] dnsData = new byte[readableBytes];
        udpData.gets(dnsData);
        try {
            Message dnsMessage = new Message(dnsData);
            DnsActivity dnsActivity = new DnsActivity(getCurrentPcapActivity(), dnsMessage);
            fireNewDnsActivity(dnsActivity);
        } catch (IOException e) {
            log.debug("Failed to decode UdpPacket {} {}", getCurrentPcapActivity().getFrameNumber(), e);
        }
    }

    private void fireNewDnsActivity(DnsActivity dnsActivity) {
        propertyChangeSupport.firePropertyChange(NEW_DNS_ACTIVITY_PROPERTY_NAME, null, dnsActivity);
    }

    @Override
    public void onReset(TcpSessionKey key) {
        onFinish(key);
    }

    @Override
    public void onEstablish(TcpSessionKey key) {
        // take the valuable IP information and create a new potential DNS session
        InetAddress clientIp = key.getClientIp();
        InetAddress serverIp = key.getServerIp();
        InetSocketAddress clientAddr = new InetSocketAddress(clientIp, key.getClientPort());
        InetSocketAddress serverAddr = new InetSocketAddress(serverIp, key.getServerPort());
        sessionMap.put(key, new DnsSession(clientAddr, serverAddr));
    }

    @Override
    public void onFinish(TcpSessionKey key) {
        sessionMap.remove(key);
    }

    @Override
    public void handleTx(TcpSessionKey session, Buffer data) {
        DnsSession dnsSession = sessionMap.get(session);
        dnsSession.txBuffer.addLast(data);
        if (dnsSession.txMessageSize <= 0) {
            // DNS over TCP is prefixed by a two byte length field (see RFC 1035)
            dnsSession.txMessageSize = dnsSession.txBuffer.getUnsignedShort();
        }
        dnsSession.txBuffer.mark();
        int readableBytes = dnsSession.txBuffer.readableBytes();
        if (readableBytes == dnsSession.txMessageSize) {
            byte[] dnsData = new byte[readableBytes];
            dnsSession.txBuffer.gets(dnsData);
            try {
                Message dnsMessage = new Message(dnsData);
                DnsActivity dnsActivity = new DnsActivity(getCurrentPcapActivity(), dnsMessage);
                fireNewDnsActivity(dnsActivity);
            } catch (IOException e) {
                log.debug("Failed to decode DNS message over tcp {} {}", getCurrentPcapActivity().getFrameNumber(), e);
                dnsSession.txBuffer.reset();
            }
        }
    }

    @Override
    public void handleRx(TcpSessionKey session, Buffer data) {
        DnsSession dnsSession = sessionMap.get(session);
        dnsSession.rxBuffer.addLast(data);
        if (dnsSession.rxMessageSize <= 0) {
            // DNS over TCP is prefixed by a two byte length field (see RFC 1035)
            dnsSession.rxMessageSize = dnsSession.rxBuffer.getUnsignedShort();
        }
        //dnsSession.rxBuffer.skip(2); // skip DNS Message size??
        dnsSession.rxBuffer.mark();
        int readableBytes = dnsSession.rxBuffer.readableBytes();
        if (readableBytes == dnsSession.rxMessageSize) {
            byte[] dnsData = new byte[readableBytes];
            dnsSession.rxBuffer.gets(dnsData);
            try {
                Message dnsMessage = new Message(dnsData);
                DnsActivity dnsActivity = new DnsActivity(getCurrentPcapActivity(), dnsMessage);
                fireNewDnsActivity(dnsActivity);
            } catch (IOException e) {
                log.debug("Failed to decode DNS message over tcp {} {}", getCurrentPcapActivity().getFrameNumber(), e);
                dnsSession.rxBuffer.reset();
            }
        }
    }

    /**
     * Encapsulates RX/TX buffers for DNS messages exchanged between a client and a server
     */
    private class DnsSession {
        private final InetSocketAddress client;
        private final InetSocketAddress server;
        private final ChainBuffer txBuffer;
        private final ChainBuffer rxBuffer;

        private int txMessageSize, rxMessageSize;

        DnsSession(InetSocketAddress client, InetSocketAddress server) {

            this.client = client;
            this.server = server;
            txBuffer = new ChainBuffer();
            rxBuffer = new ChainBuffer();
            this.txMessageSize = this.rxMessageSize = -1;
        }

        public InetSocketAddress getClient() {
            return client;
        }

        public InetSocketAddress getServer() {
            return server;
        }

        public ChainBuffer getTxBuffer() {
            return txBuffer;
        }

        public ChainBuffer getRxBuffer() {
            return rxBuffer;
        }

        public int getTxMessageSize() {
            return txMessageSize;
        }

        public int getRxMessageSize() {
            return rxMessageSize;
        }
    }


}
