/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.protocol.handler;


import at.jku.fim.rubanetra.config.model.KrakenProtocolConfiguration;
import at.jku.fim.rubanetra.config.model.ProtocolId;
import at.jku.fim.rubanetra.pcap.PcapActivity;
import at.jku.fim.rubanetra.pcap.PcapPacketToKrakenPcapPacketAdapter;
import at.jku.fim.rubanetra.protocol.KrakenBaseProtocol;
import at.jku.fim.rubanetra.protocol.activity.ethernet.EthernetActivity;
import org.krakenapps.pcap.decoder.ethernet.EthernetDecoder;
import org.krakenapps.pcap.decoder.ethernet.EthernetFrame;
import org.krakenapps.pcap.decoder.ethernet.EthernetType;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * The Kraken network layer 2 decoder {@link org.krakenapps.pcap.decoder.ethernet.EthernetDecoder} is used by this
 * handler in order to create and deliver {@link at.jku.fim.rubanetra.protocol.activity.ethernet.EthernetActivity}
 * instances. Furthermore, several bindings have been implemented, i.e.:
 * <ul>
 * <li>-> {@link at.jku.fim.rubanetra.protocol.handler.KrakenIpv4Handler}</li>
 * <li>-> {@link at.jku.fim.rubanetra.protocol.handler.KrakenIpv6Handler}</li>
 * <li>-> {@link at.jku.fim.rubanetra.protocol.handler.KrakenArpHandler}</li>
 * </ul>
 */
public class KrakenEthernetProtocolHandler extends KrakenBaseProtocol {
    public static final ProtocolId krakenEthernetProtocolId = new ProtocolId() {
        private final String protocolId = "Ethernet";

        @Override
        public String getProtocolId() {
            return protocolId;
        }

        @Override
        public String getName() {
            return protocolId;
        }

        @Override
        public String toString() {
            return getName();
        }
    };
    public static final String PROPERTY_NEW_ETHERNET_ACTIVITY = "newEthernetActivity";

    private final EthernetDecoder krakenEthernetDecoder;
    private final Logger log = LoggerFactory.getLogger(getClass());

    public KrakenEthernetProtocolHandler() {
        super();
        krakenEthernetDecoder = new EthernetDecoder();
        krakenEthernetDecoder.register(this::fireNewEthernetActivity);
    }

    @Override
    public ProtocolId getProtocolId() {
        return krakenEthernetProtocolId;
    }

    @Override
    public void initialize(KrakenProtocolConfiguration setting) {
        super.initialize(setting);
        getPcapHandler().addPcapProcessor(this::decodePcapPacket);

    }

    private void fireNewEthernetActivity(EthernetFrame frame) {
        EthernetActivity ethernetActivity = new EthernetActivity(getCurrentPcapActivity(), frame);
        propertyChangeSupport.firePropertyChange(PROPERTY_NEW_ETHERNET_ACTIVITY, null, ethernetActivity);
    }

    /**
     * Bind the internal {@link org.krakenapps.pcap.decoder.ethernet.EthernetDecoder} to an IPv4 decoder, i.e.
     * a decoder that shall receive all decoded Ethernet frames.
     *
     * @param ipv4Protocol the IPv4 decoder to bind to
     * @throws java.lang.IllegalArgumentException if the parameter is a null pointer
     */
    public void bind(KrakenIpv4Handler ipv4Protocol) {
        if (ipv4Protocol == null) {
            throw new IllegalArgumentException();
        }
        krakenEthernetDecoder.register(EthernetType.IPV4, ipv4Protocol.getKrakenIpv4Decoder());
    }

    /**
     * Bind the internal {@link org.krakenapps.pcap.decoder.ethernet.EthernetDecoder} to an IPv6 decoder, i.e.
     * a decoder that shall receive all decoded Ethernet frames.
     *
     * @param ipv6Handler the IPv6 decoder to bind to
     * @throws java.lang.IllegalArgumentException if the parameter is a null pointer
     */
    public void bind(KrakenIpv6Handler ipv6Handler) {
        krakenEthernetDecoder.register(EthernetType.IPV6, ipv6Handler.getKrakenIpv6Decoder());
    }

    /**
     * Bind the internal {@link org.krakenapps.pcap.decoder.ethernet.EthernetDecoder} to an ARP decoder, i.e.
     * a decoder that shall receive all decoded Ethernet frames.
     *
     * @param arpHandler the ARP decoder to bind to
     * @throws java.lang.IllegalArgumentException if the parameter is a null pointer
     */
    public void bind(KrakenArpHandler arpHandler) {
        krakenEthernetDecoder.register(EthernetType.ARP, arpHandler.getKrakenArpDecoder());
    }

    /**
     * Takes a Pcap entry and tries to decode the payload as an Ethernet frame using a
     * {@link org.krakenapps.pcap.decoder.ethernet.EthernetDecoder} and a suitable Pcap wrapper
     * {@link at.jku.fim.rubanetra.pcap.PcapPacketToKrakenPcapPacketAdapter}.
     *
     * @param pcapActivity the Pcap entry whose payload is to be decoded as an Ethernet frame
     * @throws java.lang.NullPointerException if pcapActivity is a null pointer
     */
    public void decodePcapPacket(PcapActivity pcapActivity) {
        if (pcapActivity != null && pcapActivity.getPcapPacket() != null) {
            krakenEthernetDecoder.decode(new PcapPacketToKrakenPcapPacketAdapter(pcapActivity.getPcapPacket()));
        } else {
            log.warn("Unable to decode a null pointer Pcap entry");
            throw new NullPointerException();
        }
    }

    /**
     * @return the Ethernet decoder used internally
     */
    public EthernetDecoder getKrakenEthernetDecoder() {
        return krakenEthernetDecoder;
    }
}
