/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.protocol.handler;


import at.jku.fim.rubanetra.config.model.KrakenProtocolConfiguration;
import at.jku.fim.rubanetra.config.model.ProtocolId;
import at.jku.fim.rubanetra.protocol.KrakenBaseProtocol;
import at.jku.fim.rubanetra.protocol.activity.icmp.Icmpv4Activity;
import org.krakenapps.pcap.decoder.icmp.IcmpDecoder;
import org.krakenapps.pcap.decoder.icmp.IcmpPacket;
import org.krakenapps.pcap.decoder.icmp.IcmpProcessor;

/**
 * This ICMP decoder handler relies on {@link org.krakenapps.pcap.decoder.icmp.IcmpDecoder} which is provided by the
 * Kraken library and on {@link at.jku.fim.rubanetra.protocol.handler.KrakenIpv4Handler} in order to
 * receive the IPv4 payload used to transport the ICMP data.
 * No forward bindings are provided by this class.
 */
public class KrakenIcmpv4Handler extends KrakenBaseProtocol {
    public static final String NEW_ICMPV_4_ACTIVITY_PROPERTY_NAME = "newIcmpv4Activity";
    public static final ProtocolId krakenIcmpv4ProtocolId = new ProtocolId() {
        private final String protocolId = "Icmpv4";

        @Override
        public String getProtocolId() {
            return protocolId;
        }

        @Override
        public String getName() {
            return protocolId;
        }

        @Override
        public String toString() {
            return getName();
        }
    };
    private final IcmpProcessor generalIcmpProcessor = new IcmpProcessor() {
        private long lastProcessedFrameNumber = -1;

        @Override
        public void process(IcmpPacket p) {
            // this is a workaround to prevent the redundant call from Kraken's IcmpDecoder of the same event using
            // two different objects
            if (lastProcessedFrameNumber != getCurrentPcapActivity().getFrameNumber()) {
                fireNewIcmpActivity(p);
            }
            lastProcessedFrameNumber = getCurrentPcapActivity().getFrameNumber();
        }
    };

    private final IcmpDecoder krakenIcmpv4Decoder;

    public KrakenIcmpv4Handler() {
        krakenIcmpv4Decoder = new IcmpDecoder();
        krakenIcmpv4Decoder.register(generalIcmpProcessor);
    }

    @Override
    public ProtocolId getProtocolId() {
        return krakenIcmpv4ProtocolId;
    }

    @Override
    public void initialize(KrakenProtocolConfiguration setting) {
        super.initialize(setting);
    }

    protected void fireNewIcmpActivity(IcmpPacket p) {
        Icmpv4Activity icmpActivity = new Icmpv4Activity(getCurrentPcapActivity(), p);
        propertyChangeSupport.firePropertyChange(NEW_ICMPV_4_ACTIVITY_PROPERTY_NAME, null, icmpActivity);
    }

    /**
     * @return the ICMP decoder that is used internally
     */
    public IcmpDecoder getKrakenIcmpv4Decoder() {
        return krakenIcmpv4Decoder;
    }
}
