/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.protocol.handler;


import at.jku.fim.rubanetra.config.model.KrakenProtocolConfiguration;
import at.jku.fim.rubanetra.config.model.ProtocolId;
import at.jku.fim.rubanetra.protocol.KrakenBaseProtocol;
import at.jku.fim.rubanetra.protocol.activity.ip.Ipv4Activity;
import org.krakenapps.pcap.decoder.ethernet.EthernetType;
import org.krakenapps.pcap.decoder.ip.InternetProtocol;
import org.krakenapps.pcap.decoder.ip.IpDecoder;
import org.krakenapps.pcap.decoder.ip.IpProcessor;
import org.krakenapps.pcap.decoder.ip.Ipv4Packet;

/**
 * This is an inefficient handler for Kraken's {@link org.krakenapps.pcap.decoder.ip.IpDecoder} implementation.
 * Since the utilized IP next protocol numbers are not known a priori this class registers all numbers in the
 * range [0x01;0xFF[ during initialization.
 * Forward bindings are provided for:
 * <ul>
 * <li>{@link at.jku.fim.rubanetra.protocol.handler.KrakenUdpHandler}</li>
 * <li>{@link at.jku.fim.rubanetra.protocol.handler.KrakenTcpHandler}</li>
 * <li>{@link at.jku.fim.rubanetra.protocol.handler.KrakenIcmpv4Handler}</li>
 * </ul>
 */
public class KrakenIpv4Handler extends KrakenBaseProtocol {
    public static final String NEW_IPV_4_ACTIVITY_PROPERTY_NAME = "newIpv4Activity";
    public static final ProtocolId krakenIpv4ProtocolId = new ProtocolId() {
        private final String protocolId = "Ipv4";

        @Override
        public String getProtocolId() {
            return protocolId;
        }

        @Override
        public String getName() {
            return protocolId;
        }

        @Override
        public String toString() {
            return getName();
        }
    };
    private final IpProcessor generalIpv4Processor = this::fireNewIpv4Activity;
    private final IpDecoder krakenIpv4Decoder;

    public KrakenIpv4Handler() {
        krakenIpv4Decoder = new IpDecoder();
    }

    @Override
    public ProtocolId getProtocolId() {
        return krakenIpv4ProtocolId;
    }

    @Override
    public void initialize(KrakenProtocolConfiguration setting) {
        super.initialize(setting);

        // the Kraken IPv4 parser always calls back protocol id 0
        krakenIpv4Decoder.register(0, generalIpv4Processor);
    }

    private void fireNewIpv4Activity(Ipv4Packet packet) {
        Ipv4Activity ipActivity = new Ipv4Activity(getCurrentPcapActivity(), packet);
        propertyChangeSupport.firePropertyChange(NEW_IPV_4_ACTIVITY_PROPERTY_NAME, null, ipActivity);
    }

    /**
     * An experimental backward binding, i.e. it registers ethernetProtocol Kraken decoder to this IPv4 decoder
     *
     * @param ethernetProtocol the Ethernet decoder to bind to this IPv4 decoder
     */
    public void bind(KrakenEthernetProtocolHandler ethernetProtocol) {
        ethernetProtocol.getKrakenEthernetDecoder().register(EthernetType.IPV4, krakenIpv4Decoder);
    }

    /**
     * Forward binding to the tcpProtocol internal TCP decoder, i.e. all decoded IPv4 packets are forwarded to the
     * TCP handler.
     *
     * @param tcpProtocol the handler for the decoded IPv4 packets
     */
    public void bind(KrakenTcpHandler tcpProtocol) {
        krakenIpv4Decoder.register(InternetProtocol.TCP, tcpProtocol.getKrakenTcpDecoder());
    }

    /**
     * Forward binding to the udpProtocol internal UDP decoder, i.e. all decoded IPv4 packets are forwarded to this
     * UDP handler.
     *
     * @param udpProtocol the handler for the decoded IPv4 packets
     */
    public void bind(KrakenUdpHandler udpProtocol) {
        krakenIpv4Decoder.register(InternetProtocol.UDP, udpProtocol.getKrakenUdpDecoder());
    }

    /**
     * Bind the internal {@link org.krakenapps.pcap.decoder.ip.IpDecoder} to an ICMPv4 decoder, i.e.
     * a decoder that shall receive all decoded IP packets.
     *
     * @param icmpv4Handler the ICMPv4 decoder to bind to
     * @throws java.lang.IllegalArgumentException if the parameter is a null pointer
     */
    public void bind(KrakenIcmpv4Handler icmpv4Handler) {
        krakenIpv4Decoder.register(InternetProtocol.ICMP, icmpv4Handler.getKrakenIcmpv4Decoder());
    }

    /**
     * @return the internal IPv4 decoder
     */
    public IpDecoder getKrakenIpv4Decoder() {
        return krakenIpv4Decoder;
    }


}
