/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.protocol.handler;


import at.jku.fim.rubanetra.config.model.KrakenProtocolConfiguration;
import at.jku.fim.rubanetra.config.model.ProtocolId;
import at.jku.fim.rubanetra.protocol.KrakenBaseProtocol;
import at.jku.fim.rubanetra.protocol.activity.ip.Ipv6Activity;
import org.krakenapps.pcap.decoder.ethernet.EthernetType;
import org.krakenapps.pcap.decoder.ip.InternetProtocol;
import org.krakenapps.pcap.decoder.ipv6.Ipv6Decoder;
import org.krakenapps.pcap.decoder.ipv6.Ipv6Packet;
import org.krakenapps.pcap.decoder.ipv6.Ipv6Processor;

/**
 * Handles IPv6 packets decoded by {@link org.krakenapps.pcap.decoder.ipv6.Ipv6Decoder} and builds
 * {@link at.jku.fim.rubanetra.protocol.activity.ip.Ipv6Activity} objects during the process.
 * Since the protocol numbers for the next protocol decoder may be arbitrary it registers to all protocol numbers in the
 * range [0x00;0xFF[ using the same processor.
 * Forward bindings are provided for:
 * <ul>
 * <li>{@link at.jku.fim.rubanetra.protocol.handler.KrakenTcpHandler}</li>
 * <li>{@link at.jku.fim.rubanetra.protocol.handler.KrakenUdpHandler}</li>
 * <li>{@link at.jku.fim.rubanetra.protocol.handler.KrakenIcmpv6Handler}</li>
 * </ul>
 */
public class KrakenIpv6Handler extends KrakenBaseProtocol {
    public static final String NEW_IPV_6_ACTIVITY_PROPERTY_NAME = "newIpv6Activity";
    public static final ProtocolId krakenIpv6ProtocolId = new ProtocolId() {
        private final String protocolId = "Ipv6";

        @Override
        public String getProtocolId() {
            return protocolId;
        }

        @Override
        public String getName() {
            return protocolId;
        }

        @Override
        public String toString() {
            return getName();
        }
    };
    private final Ipv6Processor krakenIpv6Processor = this::fireNewIpv6Activity;
    private final Ipv6Decoder krakenIpv6Decoder;

    public KrakenIpv6Handler() {
        krakenIpv6Decoder = new Ipv6Decoder();
    }

    @Override
    public ProtocolId getProtocolId() {
        return krakenIpv6ProtocolId;
    }

    @Override
    public void initialize(KrakenProtocolConfiguration setting) {
        super.initialize(setting);
        for (byte protocolNumber = Byte.MIN_VALUE; protocolNumber < Byte.MAX_VALUE; protocolNumber++) {
            krakenIpv6Decoder.register(protocolNumber, krakenIpv6Processor);
        }
    }

    private void fireNewIpv6Activity(Ipv6Packet packet) {
        Ipv6Activity ipActivity = new Ipv6Activity(getCurrentPcapActivity(), packet);
        propertyChangeSupport.firePropertyChange(NEW_IPV_6_ACTIVITY_PROPERTY_NAME, null, ipActivity);
    }

    /**
     * Experimental backward binding using an Ethernet decoder that is to be bound to this IPv6 decoder.
     *
     * @param ethernetProtocol the Ethernet decoder to bind to this IPv6 decoder
     */
    public void bind(KrakenEthernetProtocolHandler ethernetProtocol) {
        ethernetProtocol.getKrakenEthernetDecoder().register(EthernetType.IPV6, krakenIpv6Decoder);
    }

    /**
     * Forward binding from this IPv6 decoder to a TCP protocol decoder.
     *
     * @param tcpProtocol the TCP decoder to receive the decoded IPv6 packets
     */
    public void bind(KrakenTcpHandler tcpProtocol) {
        krakenIpv6Decoder.register(InternetProtocol.TCP, tcpProtocol.getKrakenTcpDecoder());
    }

    /**
     * Forward binding from this IPv6 decoder to an UDP decoder.
     *
     * @param udpHandler the UDP decoder to receive the decoded IPv6 packets
     */
    public void bind(KrakenUdpHandler udpHandler) {
        krakenIpv6Decoder.register(InternetProtocol.UDP, udpHandler.getKrakenUdpDecoder());
    }

    /**
     * Bind the internal {@link org.krakenapps.pcap.decoder.ipv6.Ipv6Decoder} to an ICMPv6 decoder, i.e.
     * a decoder that shall receive all decoded IP packets.
     *
     * @param icmpv6Handler the ICMPv6 decoder to bind to
     * @throws java.lang.IllegalArgumentException if the parameter is a null pointer
     */
    public void bind(KrakenIcmpv6Handler icmpv6Handler) {
        krakenIpv6Decoder.register(InternetProtocol.ICMPV6, icmpv6Handler.getKrakenImpv6Decoder());
    }

    /**
     * @return Internally used IPv6 decoder
     */
    public Ipv6Decoder getKrakenIpv6Decoder() {
        return krakenIpv6Decoder;
    }
}
