/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.protocol.handler;


import at.jku.fim.rubanetra.config.model.KrakenProtocolConfiguration;
import at.jku.fim.rubanetra.config.model.ProtocolId;
import at.jku.fim.rubanetra.protocol.KrakenBaseProtocol;
import at.jku.fim.rubanetra.protocol.activity.tcp.TcpActivity;
import at.jku.fim.rubanetra.protocol.mapper.KrakenTcpProtocolMapperStrategy;
import org.krakenapps.pcap.decoder.ip.InternetProtocol;
import org.krakenapps.pcap.decoder.tcp.TcpDecoder;
import org.krakenapps.pcap.decoder.tcp.TcpPacket;
import org.krakenapps.pcap.decoder.tcp.TcpSession;

/**
 * This class handles TCP transport layer related events that are dispatched by
 * a {@link org.krakenapps.pcap.decoder.tcp.TcpDecoder}. The following properties
 * may be observed:
 * Due to Kraken's constraint of using a TCP-Next Protocol mapper, which is per se
 * not always desirable for certain use cases, a corresponding Mapper has to be
 * constructed and passed during the creation of the TCP decoder itself. Currently,
 * this is not directly feasible because in theory, the exact strategy is chosen
 * by the user and may deviate from application layer protocol to another application layer
 * protocol. This library constraint is respected by creating the TCP decoder
 * only during the invocation of {@link #initialize(at.jku.fim.rubanetra.config.model.KrakenProtocolConfiguration)}
 * as well as the transport layer to application layer mapping strategy.
 * However, if a restriction based, e.g. port-based, mapping strategy is considered it should be noted that
 * this class tries to dispatch TCP activities regardless of the mapping strategy
 * in use. These activities are exclusively passed to listeners on the individual
 * instance of this class, which may not care about a potential application layer
 * restriction that would in practice discard any non-relevant packets at the
 * transport layer due to the mapping strategy in use and prevent potentially useful analysis.
 * Therefore, {@link at.jku.fim.rubanetra.protocol.activity.tcp.TcpActivity} objects
 * are dispatched to all registered listeners regardless of mapping strategy restrictions in place, and any protocol handler
 * bindings will only receive those decoded packets that pass and conform to the mapping strategy in use (which
 * could be all, none or a subset of all TCP packets).
 * These forward bindings are currently comprised of:
 * <ul>
 * <li>-> {@link at.jku.fim.rubanetra.protocol.handler.KrakenHttpHandler}</li>
 * <li>-> {@link at.jku.fim.rubanetra.protocol.handler.KrakenMsnHandler}</li>
 * <li>-> {@link at.jku.fim.rubanetra.protocol.handler.KrakenDnsProtocolHandler}</li>
 * <li>-> {@link at.jku.fim.rubanetra.protocol.handler.KrakenFtpHandler}</li>
 * <li>-> {@link at.jku.fim.rubanetra.protocol.handler.KrakenNetbiosHandler}</li>
 * <li>-> {@link at.jku.fim.rubanetra.protocol.handler.KrakenPop3Handler}</li>
 * <li>-> {@link at.jku.fim.rubanetra.protocol.handler.KrakenTelnetHandler}</li>
 * <li>-> {@link at.jku.fim.rubanetra.protocol.handler.KrakenSmtpHandler}</li>
 * </ul>
 */
public class KrakenTcpHandler extends KrakenBaseProtocol {
    public static final String NEW_TCP_ACTIVITY_PROPERTY_NAME = "newTcpActivity";
    public static final ProtocolId krakenTcpProtocolId = new ProtocolId() {
        private final String protocolId = "Tcp";

        @Override
        public String getProtocolId() {
            return protocolId;
        }

        @Override
        public String getName() {
            return protocolId;
        }

        @Override
        public String toString() {
            return getName();
        }
    };
    private TcpDecoder krakenTcpDecoder;
    private KrakenTcpProtocolMapperStrategy tcpMappingStrategy;

    @Override
    public void initialize(KrakenProtocolConfiguration setting) {
        super.initialize(setting);
        String mappingStrategyId = setting.getTransportLayerMappingStrategy();
        tcpMappingStrategy = setting.getTransportLayerMappingFactory().createTcpMappingStrategy(mappingStrategyId);
        tcpMappingStrategy.setup(setting);
        krakenTcpDecoder = new TcpDecoder(tcpMappingStrategy);
        krakenTcpDecoder.registerSegmentCallback((session, segment) -> fireNewTcpActivity((TcpPacket) segment, session));
    }

    protected void fireNewTcpActivity(TcpPacket tcpPacket, TcpSession tcpSession) {
        TcpActivity tcpActivity = new TcpActivity(getCurrentPcapActivity(), tcpPacket, tcpSession);
        log.debug("Fire new TcpActivity, frame number: {}", tcpActivity.getPcapActivity().getFrameNumber());
        propertyChangeSupport.firePropertyChange(NEW_TCP_ACTIVITY_PROPERTY_NAME, null, tcpActivity);
    }

    @Override
    public ProtocolId getProtocolId() {
        return krakenTcpProtocolId;
    }

    /**
     * Experimental backward binding from an IPv4 handler to this TCP handler
     *
     * @param protocol the IPv4 handler
     */
    public void bind(KrakenIpv4Handler protocol) {
        protocol.getKrakenIpv4Decoder().register(InternetProtocol.TCP, krakenTcpDecoder);
    }

    /**
     * Experimental backward binding from an IPv6 handler to this TCP handler
     *
     * @param protocol the IPv6 handler
     */
    public void bind(KrakenIpv6Handler protocol) {
        protocol.getKrakenIpv6Decoder().register(InternetProtocol.TCP, krakenTcpDecoder);
    }

    /**
     * Forward binding to a HTTP protocol handler, i.e. all decoded TCP segments that correspond
     * to the transport layer mapping strategy will be forwarded to the provided HTTP handler.
     *
     * @param protocolHandler the HTTP protocol handler
     */
    public void bind(KrakenHttpHandler protocolHandler) {
        tcpMappingStrategy.register(protocolHandler.getKrakenHttpDecoder(), KrakenHttpHandler.krakenHttpProtocolId);
    }

    /**
     * Forward binding to a MSN handler, i.e. all decoded TCP segments that correspond to
     * the transport layer mapping strategy will be forwarded to the provided MSN handler.
     *
     * @param protocolHandler the MSN protocol handler
     */
    public void bind(KrakenMsnHandler protocolHandler) {
        tcpMappingStrategy.register(protocolHandler.getKrakenMsnDecoder(), KrakenMsnHandler.krakenMsnProtocolId);
    }

    /**
     * Forward binding to a DNS handler, i.e. all decoded TCP segments that correspond to
     * the transport layer mapping strategy will be forwarded to the provided DNS handler.
     *
     * @param dnsHandler the DNS protocol handler
     */
    public void bind(KrakenDnsProtocolHandler dnsHandler) {
        tcpMappingStrategy.register(dnsHandler, KrakenDnsProtocolHandler.krakenDnsProtocolId);
    }

    /**
     * Forward binding to a FTP handler, i.e. all decoded TCP segments that correspond to
     * the transport layer mapping strategy will be forwarded to the provided protocol handler.
     *
     * @param ftpHandler the FTP protocol handler
     */
    public void bind(KrakenFtpHandler ftpHandler) {
        tcpMappingStrategy.register(ftpHandler.getKrakenFtpDecoder(), KrakenFtpHandler.krakenFtpProtocolId);
    }

    /**
     * Forward binding to a NetBios handler, i.e. all decoded TCP segments that correspond to
     * the transport layer mapping strategy will be forwarded to the provided protocol handler.
     *
     * @param netbiosHandler the NetBios handler
     */
    public void bind(KrakenNetbiosHandler netbiosHandler) {
        tcpMappingStrategy.register(netbiosHandler.getKrakenNetbiosDecoder(), KrakenNetbiosHandler.krakenNetbiosProtocolId);
    }

    /**
     * Forward binding to a POP3 handler, i.e. all decoded TCP segments that correspond to
     * the transport layer mapping strategy will be forwarded to the provided protocol handler.
     *
     * @param pop3Handler the POP3 handler
     */
    public void bind(KrakenPop3Handler pop3Handler) {
        tcpMappingStrategy.register(pop3Handler.getKrakenPop3Decoder(), KrakenPop3Handler.krakenPop3ProtocolId);
    }

    /**
     * Forward binding to a Telnet handler, i.e. all decoded TCP segments that correspond to
     * the transport layer mapping strategy will be forwarded to the provided protocol handler.
     *
     * @param telnetHandler the Telnet handler
     */
    public void bind(KrakenTelnetHandler telnetHandler) {
        tcpMappingStrategy.register(telnetHandler.getKrakenTelnetDecoder(), KrakenTelnetHandler.krakenTelnetProtocolId);
    }

    /**
     * Forward binding to a SMTP handler, i.e. all decoded TCP segments that correspond to
     * the transport layer mapping strategy will be forwarded to the provided protocol handler.
     *
     * @param smtpHandler the SMTP handler
     */
    public void bind(KrakenSmtpHandler smtpHandler) {
        tcpMappingStrategy.register(smtpHandler.getKrakenSmtpDecoder(), KrakenSmtpHandler.krakenSmtpProtocolId);
    }

    /**
     * @return the internally used TCP decoder
     */
    public TcpDecoder getKrakenTcpDecoder() {
        return krakenTcpDecoder;
    }
}
