/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.protocol.handler;

import at.jku.fim.rubanetra.config.model.KrakenApplicationProtocolId;
import at.jku.fim.rubanetra.config.model.ProtocolId;
import at.jku.fim.rubanetra.protocol.KrakenBaseProtocol;
import at.jku.fim.rubanetra.protocol.activity.telnet.TelnetActivity;
import org.krakenapps.pcap.Protocol;
import org.krakenapps.pcap.decoder.telnet.*;

/**
 * Handles Telnet data decoded by the Kraken library and creates suitable
 * {@link at.jku.fim.rubanetra.protocol.activity.telnet.TelnetActivity}
 * objects that are dispatched to all registered listeners. These activities are
 * currently not reassembled, i.e. the session reassembly should be performed
 * via appropriate Drools rule specifications.
 */
public class KrakenTelnetHandler extends KrakenBaseProtocol {
    public static final KrakenApplicationProtocolId krakenTelnetProtocolId = new KrakenApplicationProtocolId() {
        @Override
        public Protocol getKrakenApplicationProtocol() {
            return Protocol.TELNET;
        }

        @Override
        public String getProtocolId() {
            return Protocol.TELNET.name();
        }

        @Override
        public String getName() {
            return "Telnet";
        }
    };
    private final TelnetDecoder krakenTelnetDecoder;

    public KrakenTelnetHandler() {
        krakenTelnetDecoder = new TelnetDecoder();
        krakenTelnetDecoder.register(new TelnetProcessor() {
            @Override
            public void onClientCommand(TelnetCommand command, TelnetOption option, byte[] data) {
                TelnetActivity act = new TelnetActivity(getCurrentPcapActivity(), TelnetActivity.Type.CLIENT_COMMAND);
                act.setCommand(command);
                act.setOption(option);
                act.setData(data);
                fireNewTelnetActivity(act);
            }

            @Override
            public void onServerCommand(TelnetCommand command, TelnetOption option, byte[] data) {
                TelnetActivity act = new TelnetActivity(getCurrentPcapActivity(), TelnetActivity.Type.SERVER_COMMAND);
                act.setCommand(command);
                act.setOption(option);
                act.setData(data);
                fireNewTelnetActivity(act);
            }

            @Override
            public void onClientAnsiControl(AnsiMode mode, TelnetCommand command, int[] arguments) {
                TelnetActivity act = new TelnetActivity(getCurrentPcapActivity(), TelnetActivity.Type.CLIENT_ANSICONTROL);
                act.setAnsiMode(mode);
                act.setCommand(command);
                act.setArguments(arguments);
                fireNewTelnetActivity(act);
            }

            @Override
            public void onServerAnsiControl(AnsiMode mode, TelnetCommand command, int[] arguments) {
                TelnetActivity act = new TelnetActivity(getCurrentPcapActivity(), TelnetActivity.Type.SERVER_ANSICONTROL);
                act.setAnsiMode(mode);
                act.setCommand(command);
                act.setArguments(arguments);
                fireNewTelnetActivity(act);
            }

            @Override
            public void onClientData(String text) {
                TelnetActivity act = new TelnetActivity(getCurrentPcapActivity(), TelnetActivity.Type.CLIENT_DATA);
                act.setText(text);
                fireNewTelnetActivity(act);
            }

            @Override
            public void onServerData(String text) {
                TelnetActivity act = new TelnetActivity(getCurrentPcapActivity(), TelnetActivity.Type.SERVER_DATA);
                act.setText(text);
                fireNewTelnetActivity(act);
            }

            @Override
            public void onClientTitle(String title) {
                TelnetActivity act = new TelnetActivity(getCurrentPcapActivity(), TelnetActivity.Type.CLIENT_TITLE);
                act.setTitle(title);
                fireNewTelnetActivity(act);
            }

            @Override
            public void onServerTitle(String title) {
                TelnetActivity act = new TelnetActivity(getCurrentPcapActivity(), TelnetActivity.Type.SERVER_TITLE);
                act.setTitle(title);
                fireNewTelnetActivity(act);
            }
        });
    }

    private void fireNewTelnetActivity(TelnetActivity act) {
        propertyChangeSupport.firePropertyChange("newPartialTelnetActivity", null, act);
    }

    @Override
    public ProtocolId getProtocolId() {
        return krakenTelnetProtocolId;
    }

    /**
     * @return the internally used Telnet decoder
     */
    public TelnetDecoder getKrakenTelnetDecoder() {
        return krakenTelnetDecoder;
    }
}
