/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.protocol.handler;


import at.jku.fim.rubanetra.config.model.KrakenProtocolConfiguration;
import at.jku.fim.rubanetra.config.model.ProtocolId;
import at.jku.fim.rubanetra.protocol.KrakenBaseProtocol;
import at.jku.fim.rubanetra.protocol.activity.udp.UdpActivity;
import at.jku.fim.rubanetra.protocol.mapper.KrakenTransportLayerMappingFactory;
import at.jku.fim.rubanetra.protocol.mapper.KrakenUdpProtocolMapperStrategy;
import org.krakenapps.pcap.Protocol;
import org.krakenapps.pcap.decoder.ip.InternetProtocol;
import org.krakenapps.pcap.decoder.udp.UdpDecoder;
import org.krakenapps.pcap.decoder.udp.UdpPacket;

/**
 * This class handles UDP transport layer related events that are dispatched by
 * a {@link org.krakenapps.pcap.decoder.udp.UdpDecoder} similar to the
 * {@link at.jku.fim.rubanetra.protocol.handler.KrakenTcpHandler} implementation.
 * Due to Kraken's constraint of using a UDP-Next Protocol mapper, which is per se
 * not always desirable for certain use cases, a corresponding Mapper has to be
 * constructed and passed during the creation of the UDP decoder itself. Currently,
 * this is not directly feasible because in theory, the exact strategy is chosen
 * by the user and may deviate from application layer protocol to another application layer
 * protocol. This library constraint is respected by creating the UDP decoder
 * only during the invocation of {@link #initialize(at.jku.fim.rubanetra.config.model.KrakenProtocolConfiguration)}
 * as well as the transport layer to application layer mapping strategy.
 * However, if a restriction based, e.g. port-based, mapping strategy is considered it should be noted that
 * this class tries to dispatch UDP activities regardless of the mapping strategy
 * in use. These activities are exclusively passed to listeners on the individual
 * instance of this class, which may not care about a potential application layer
 * restriction that would in practice discard any non-relevant packets at the
 * transport layer due to the mapping strategy in use and prevent potentially useful analysis.
 * Therefore, {@link at.jku.fim.rubanetra.protocol.activity.udp.UdpActivity} objects
 * are dispatched to all registered listeners regardless of mapping strategy restrictions in place, and any protocol handler
 * bindings will only receive those decoded packets that pass and conform to the mapping strategy in use (which
 * could be all, none or a subset of all UDP packets).
 * These forward bindings are currently comprised of:
 * <ul>
 * <li>-> {@link at.jku.fim.rubanetra.protocol.handler.KrakenDnsProtocolHandler}</li>
 * <li>-> {@link at.jku.fim.rubanetra.protocol.handler.KrakenDhcpHandler}</li>
 * <li>-> {@link at.jku.fim.rubanetra.protocol.handler.KrakenNetbiosHandler}</li>
 * <li>-> {@link at.jku.fim.rubanetra.protocol.handler.KrakenSnmpHandler}</li>
 * </ul>
 */
public class KrakenUdpHandler extends KrakenBaseProtocol {
    private UdpDecoder krakenUdpDecoder;
    private KrakenUdpProtocolMapperStrategy mappingStrategy;

    @Override
    public void initialize(KrakenProtocolConfiguration setting) {
        super.initialize(setting);
        String mappingStrategyId = setting.getTransportLayerMappingStrategy();
        mappingStrategy = setting.getTransportLayerMappingFactory().createUdpMappingStrategy(mappingStrategyId);
        mappingStrategy.setup(setting);
        krakenUdpDecoder = new UdpDecoder(mappingStrategy);
        krakenUdpDecoder.registerUdpProcessor(this::fireNewUdpActivity);
    }

    @Override
    public ProtocolId getProtocolId() {
        return krakenUdpProtocolId;
    }

    public static final ProtocolId krakenUdpProtocolId = new ProtocolId() {
        private final String protocolId = "Udp";

        @Override
        public String getProtocolId() {
            return protocolId;
        }

        @Override
        public String getName() {
            return protocolId;
        }

        @Override
        public String toString() {
            return getName();
        }
    };

    protected void fireNewUdpActivity(UdpPacket udpPacket) {
        UdpActivity udpActivity = new UdpActivity(getCurrentPcapActivity(), udpPacket);
        propertyChangeSupport.firePropertyChange("newUdpActivity", null, udpActivity);
    }

    /**
     * Experimental backward binding from an IPv4 protocol decoder to this UDP decoder
     *
     * @param protocol the IPv4 handler
     */
    public void bind(KrakenIpv4Handler protocol) {
        protocol.getKrakenIpv4Decoder().register(InternetProtocol.UDP, krakenUdpDecoder);
    }

    /**
     * Experimental backward binding from an IPv6 protocol decoder to this UDP decoder
     *
     * @param protocol the IPv6 handler
     */
    public void bind(KrakenIpv6Handler protocol) {
        protocol.getKrakenIpv6Decoder().register(InternetProtocol.UDP, krakenUdpDecoder);
    }

    /**
     * Forward binding from this UDP decoder to the provided DNS protocol handler, i.e.
     * any decoded UDP segments will be forwarded to the passed protocol decoder.
     *
     * @param dnsHandler the DNS protocol handler
     */
    public void bind(KrakenDnsProtocolHandler dnsHandler) {
        mappingStrategy.register(dnsHandler, KrakenDnsProtocolHandler.krakenDnsProtocolId);
    }

    /**
     * Forward binding from this UDP decoder to the provided DHCP protocol handler, i.e.
     * any decoded UDP segments will be forwarded to the passed protocol decoder.
     *
     * @param dhcpHandler the DHCP protocol handler
     */
    public void bind(KrakenDhcpHandler dhcpHandler) {
        mappingStrategy.register(dhcpHandler.getKrakenDhcpDecoder(), KrakenDhcpHandler.krakenDhcpProtocolId);
    }

    /**
     * Forward binding from this UDP decoder to the provided NetBios protocol handler, i.e.
     * any decoded UDP segments will be forwarded to the passed protocol decoder.
     *
     * @param netbiosHandler the NetBios handler
     */
    public void bind(KrakenNetbiosHandler netbiosHandler) {
        mappingStrategy.register(netbiosHandler.getKrakenNetbiosDecoder(), KrakenNetbiosHandler.krakenNetbiosProtocolId);
    }

    /**
     * Forward binding from this UDP decoder to the provided SNMP protocol handler, i.e.
     * any decoded UDP segments will be forwarded to the passed protocol decoder.
     *
     * @param snmpHandler the SNMP protocol handler
     */
    public void bind(KrakenSnmpHandler snmpHandler) {
        mappingStrategy.register(snmpHandler.getKrakenSnmpDecoder(), KrakenSnmpHandler.krakenSnmpProtocolId);
    }

    /**
     * @return the internally used UDP protocol
     */
    public UdpDecoder getKrakenUdpDecoder() {
        return krakenUdpDecoder;
    }


}
