/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.protocol.mapper;


import at.jku.fim.rubanetra.protocol.mapper.impl.KrakenTcpDestinationPortProtocolMapper;
import at.jku.fim.rubanetra.protocol.mapper.impl.KrakenUdpDestinationPortProtocolMapper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.HashMap;
import java.util.Map;

/**
 * Constructs suitable {@link at.jku.fim.rubanetra.protocol.mapper.TransportLayerMappingStrategy}
 * implementations given a specific strategy identifier string.
 * Handles both, the UDP and TCP strategies, i.e.: {@link at.jku.fim.rubanetra.protocol.mapper.KrakenTcpProtocolMapperStrategy}
 * and {@link at.jku.fim.rubanetra.protocol.mapper.KrakenUdpProtocolMapperStrategy}.
 * Custom strategies may be registered during runtime.
 * The default factory {@link #createDefaultKrakenTransportLayerMappingFactory()} only provides a
 * destination port strategy for TCP/UDP.
 */
public class KrakenTransportLayerMappingFactory {
    public static final String DESTINATION_PORT_STRATEGY = "DESTINATION_PORT";
    private final Map<String, Class<? extends KrakenTcpProtocolMapperStrategy>> tcpStrategyMap;
    private final Map<String, Class<? extends KrakenUdpProtocolMapperStrategy>> udpStrategyMap;
    private final Logger log = LoggerFactory.getLogger(getClass());

    public KrakenTransportLayerMappingFactory() {
        tcpStrategyMap = new HashMap<>();
        udpStrategyMap = new HashMap<>();
    }

    /**
     * Creates a default transport layer mapping factory that consists of only one strategy: a Destination port to
     * application layer protocol strategy, i.e. {@link at.jku.fim.rubanetra.protocol.mapper.impl.KrakenTcpDestinationPortProtocolMapper}
     * and {@link at.jku.fim.rubanetra.protocol.mapper.impl.KrakenUdpDestinationPortProtocolMapper}.
     *
     * @return the default transport layer mapping factory
     */
    public static KrakenTransportLayerMappingFactory createDefaultKrakenTransportLayerMappingFactory() {
        KrakenTransportLayerMappingFactory factory = new KrakenTransportLayerMappingFactory();
        factory.registerTcpStrategy(DESTINATION_PORT_STRATEGY, KrakenTcpDestinationPortProtocolMapper.class);
        factory.registerUdpStrategy(DESTINATION_PORT_STRATEGY, KrakenUdpDestinationPortProtocolMapper.class);
        return factory;
    }

    /**
     * Registers a new TCP mapping strategy, the provided class must define a default constructor.
     *
     * @param tcpStrategyId    the unique strategy identifier
     * @param tcpStrategyClass the class to map to the identifier
     */
    public void registerTcpStrategy(String tcpStrategyId, Class<? extends KrakenTcpProtocolMapperStrategy> tcpStrategyClass) {
        if (tcpStrategyMap.containsKey(tcpStrategyId)) {
            log.error("{} has already been registered as TCP mapping strategy.", tcpStrategyId);
            throw new IllegalArgumentException();
        }
        tcpStrategyMap.put(tcpStrategyId, tcpStrategyClass);
    }

    /**
     * Registers a new UDP mapping strategy, the provided class must define a default constructor
     *
     * @param udpStrategyId    the unique strategy identifier
     * @param udpStrategyClass the class to map to the identifier
     */
    public void registerUdpStrategy(String udpStrategyId, Class<? extends KrakenUdpProtocolMapperStrategy> udpStrategyClass) {
        if (udpStrategyMap.containsKey(udpStrategyId)) {
            log.error("{} has already been registered as UDP mapping strategy.", udpStrategyId);
            throw new IllegalArgumentException();
        }
        udpStrategyMap.put(udpStrategyId, udpStrategyClass);
    }

    /**
     * Tries to create a new mapping strategy using the passed identifier to determine the mapping class that will be
     * instantiated using the default constructor.
     *
     * @param mappingStrategy the identifier that was used to register the TCP mapping strategy
     * @return a new instance of a TCP mapping strategy
     */
    public KrakenTcpProtocolMapperStrategy createTcpMappingStrategy(String mappingStrategy) {
        Class<? extends KrakenTcpProtocolMapperStrategy> tcpProtocolMapperClass = tcpStrategyMap.get(mappingStrategy);
        if (tcpProtocolMapperClass == null) {
            throw new IllegalArgumentException(String.format("Mapping strategy %s was not registered to the TransportLayerMappingFactory", mappingStrategy));
        }

        try {
            return tcpProtocolMapperClass.newInstance();
        } catch (IllegalAccessException | InstantiationException e) {
            throw new IllegalArgumentException(String.format("A default constructor for mapping strategy %s and class %s must be provided.", mappingStrategy, tcpProtocolMapperClass.getName()));
        }
    }

    /**
     * Tries to create a new mapping strategy using the passed identifier to determine the mapping class that will be
     * instantiated using the default constructor.
     *
     * @param mappingStrategy the identifier that was used to register the UDP mapping strategy
     * @return a new instance of a UDP mapping strategy
     */
    public KrakenUdpProtocolMapperStrategy createUdpMappingStrategy(String mappingStrategy) {
        Class<? extends KrakenUdpProtocolMapperStrategy> udpProtocolMapperClass = udpStrategyMap.get(mappingStrategy);
        if (udpProtocolMapperClass == null) {
            throw new IllegalArgumentException(String.format("Mapping strategy %s was not registered to the TransportLayerMappingFactory", mappingStrategy));
        }

        try {
            return udpProtocolMapperClass.newInstance();
        } catch (IllegalAccessException | InstantiationException e) {
            throw new IllegalArgumentException(String.format("Please provide a default constructor for mapping strategy %s and class %s.", mappingStrategy, udpProtocolMapperClass.getName()));
        }
    }

}
