/**
 * This file is part of Rubanetra.
 * Copyright (C) 2013,2014 Stefan Swerk (stefan_rubanetra@swerk.priv.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package at.jku.fim.rubanetra.protocol.mapper.impl;


import at.jku.fim.rubanetra.config.model.KrakenApplicationProtocolId;
import at.jku.fim.rubanetra.config.model.KrakenProtocolConfiguration;
import at.jku.fim.rubanetra.config.model.PortSpecification;
import at.jku.fim.rubanetra.protocol.mapper.KrakenTcpProtocolMapperStrategy;
import org.krakenapps.pcap.Protocol;
import org.krakenapps.pcap.decoder.tcp.TcpProcessor;
import org.krakenapps.pcap.decoder.tcp.TcpSegment;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.net.InetSocketAddress;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

/**
 * Maps a TCP based {@link at.jku.fim.rubanetra.config.model.PortSpecification} to a single application
 * layer protocol. Multiple TCP processors may be registered for the same application layer protocol.
 */
public class KrakenTcpDestinationPortProtocolMapper implements KrakenTcpProtocolMapperStrategy {
    private final Logger log = LoggerFactory.getLogger(getClass());
    private final Set<TcpProcessor> tcpProcessorSet;
    private PortSpecification portSpec;
    private KrakenApplicationProtocolId applicationProtocolId;

    public KrakenTcpDestinationPortProtocolMapper() {
        this.tcpProcessorSet = new HashSet<>();
    }

    @Override
    public Protocol map(TcpSegment segment) {
        // at this point we have to identify the protocol, however,
        // this is not feasible due to the invalid assumption that there is a single Protocol per port.
        // Therefore whenever the destination port is checked at this point not only one but several protocols
        // may be identified. There are at least two solutions to this issue: Either the raw buffer is parsed according
        // to every application protocol setting or the packet is decoded multiple times by different Protocol
        // Mappers.
        if (this.applicationProtocolId != null && portSpec.contains(segment.getDestinationPort())) {
            return this.applicationProtocolId.getKrakenApplicationProtocol();
        } else {
            log.debug("returning null pointer as KrakenProtocol for application layer protocol {}", applicationProtocolId);
            return null;
        }
    }

    @Override
    public Collection<TcpProcessor> getTcpProcessors(Protocol protocol) {
        if (this.applicationProtocolId == null) {
            if (protocol == null) {
                // this means probably that there is no official Application Layer Protocol as successor, therefore we return the set
                return Collections.unmodifiableCollection(tcpProcessorSet);
            } else {
                // protocols do not match
                return Collections.emptySet();
            }
        }
        return protocol == this.applicationProtocolId.getKrakenApplicationProtocol() ?
                Collections.unmodifiableCollection(tcpProcessorSet) : Collections.unmodifiableCollection(new HashSet<>());
    }

    @Override
    public void register(Protocol protocol, TcpProcessor processor) {
        if (this.applicationProtocolId == null || protocol != this.applicationProtocolId.getKrakenApplicationProtocol()) {
            throw new IllegalArgumentException();
        }
        this.tcpProcessorSet.add(processor);
    }

    @Override
    public void unregister(Protocol protocol, TcpProcessor processor) {
        this.tcpProcessorSet.remove(processor);
    }

    @Override
    public void register(InetSocketAddress server, Protocol protocol) {
        throw new IllegalStateException("Not yet implemented");
    }

    @Override
    public void unregister(InetSocketAddress server) {
        throw new IllegalStateException("Not yet implemented");
    }

    @Override
    @Deprecated
    public TcpProcessor getTcpProcessor(Protocol protocol) {
        throw new IllegalStateException("Not yet implemented");
    }

    @Override
    @Deprecated
    public void unregister(Protocol protocol) {
        throw new IllegalStateException("Not yet implemented");
    }

    @Override
    public void setup(KrakenProtocolConfiguration protocolSetting) {
        this.portSpec = protocolSetting.getPortSpecification();
    }

    @Override
    public void register(TcpProcessor processor, KrakenApplicationProtocolId protocolId) {
        if (this.applicationProtocolId != null) {
            throw new IllegalStateException("This destination port protocol mapper only supports a single Application Layer Protocol per instance");
        }
        this.tcpProcessorSet.add(processor);
        this.applicationProtocolId = protocolId;
    }

    @Override
    public Set<TcpProcessor> getTcpProcessors() {
        return Collections.unmodifiableSet(tcpProcessorSet);
    }
}
